/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CORBAServer
// @module  CORBAServer | This module contains the implementation of the
//          CORBAServer class.  This class is responsible for hiding low-level
//          CORBA functionality and wrapping it up into an easy to use class.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CORBA/CORBAServer.h>
using namespace CORBA;

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates an instance of a CORBAServer object.  This
//          particular constructor creates its own ORB and BOA references.
// @parm    The number of command-line arguments
// @parm    The array of command-line arguments
//
// @xref    <c CORBAGeneral>
//
// @ex      How to create a CORBAServer object using this constructor: |
//          int main( int argc, char** argv )
//          {
//              // Create CORBAServer object
//              CORBAServer* pServer = new CORBAServer( argc, argv );
//
//              ...
//
//              delete pServer;
//
//              return 1;
//          }
CORBAServer::CORBAServer( int argc, char** argv )
        :CORBAGeneral( argc, argv )
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates an instance of a CORBAServer object.  This
//          particular constructor takes references to an ORB or BOA.
// @parm    The current ORB object.
// @parm    The current BOA object.
//
// @xref    <c CORBAGeneral>
//
// @ex      How to create a CORBAServer object using this constructor: |
//          int main( int argc, char** argv )
//          {
//              // Initialise ORB and BOA
//              CORBA::ORB_ptr orb = CORBA::ORB_Init( argc, argv, "omniORB2" );
//              CORBA::BOA_ptr boa = orb->BOA_Init( argc, argv, "omniORB2_BOA" );
//
//              // Create CORBAServer object
//              CORBAServer* pServer = new CORBAServer( orb, boa );
//
//              ...
//
//              delete pServer;
//
//              return 1;
//          }

CORBAServer::CORBAServer( ORB_ptr orb, BOA_ptr boa )
        :CORBAGeneral( orb, boa )
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor.

CORBAServer::~CORBAServer()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is responsible for registering an object with the
//          NameService and ensuring that the ORB knows that the object is
//          ready for use.
// @parm    The object to be registered.
// @parm    The type of the object being registered.
// @parm    The name to register the object under.
//
// @xref    <mf CORBAServer::DeregisterObject>
//
// @ex      Typical example of registering an object: |
//          int main( int argc, char** argv )
//          {
//              // Create CORBAServer object
//              CORBAServer* pServer = new CORBAServer( argc, argv );
//
//              // Create a new Factory
//              FactoryImpl* factory = new FactoryImpl( "ACME", pServer->GetORB() );
//
//              // Register factory
//              pServer->RegisterObject( factory, Factory::CLASS_NAME, "ACME" );
//
//              ...
//
//              // Deregister factory
//              pServer->DeregisterObject( Factory::CLASS_NAME, "ACME" );
//
//              // Tidy up memory
//              factory->_dispose();
//              delete pServer;
//
//              return 1;
//          }

void CORBAServer::RegisterObject( Object_ptr obj, const char* cls,
                                  const char* name )
{
    // No errors - yet
    const char* where = "CORBAServer::RegisterObject";
    int status = CYR_UNKNOWN_ERROR;
    const char* error = NULL;

    try
    {
        // Inform ORB that object is ready to be accessed
        ReadyObject( obj );
        CORBA::Object_var obj_var=CORBA::Object::_duplicate(obj);
    }
    catch( CyrCORBAException )
    {
        // If we catch an exception here then continue as we have an anonymous
        // class which has already been declared ready.
    }

    // Try to resolve naming context.

    CosNaming::NamingContext_var context = GetNamingService();

    // Register object with given name
    CosNaming::Name cosName;
    cosName.length( 1 );
    CosNaming::NameComponent cosNameComponent;
    cosNameComponent.id   = CORBA::string_dup(name);
    cosNameComponent.kind = CORBA::string_dup(cls);
    cosName[0]=cosNameComponent;

    try
    {
        // Try to bind object in NameService
        context->bind( cosName, obj );
    }
    catch( CosNaming::NamingContext::AlreadyBound )
    {
        // We might fail because the object is already bound.  If so rebind the
        // object and continue as normal as this is not strictly an error.
        context->rebind( cosName, obj );
    }
    catch( CORBA::Exception )
    {
        // Any other CORBA exception is trapped here.  The most likely cause of
        // this problem will be due to the NameService being unavailable.
        error = "Could not bind object.  Is NameService active?";
        status = CYR_CORBA_OBJECT_NOT_FOUND;
    }
    catch( ... )
    {
        // Catch any other exceptions and complain bitterly.
        error = "Unknown exception";
    }

    // If we have an error then throw an exception
    if ( error != NULL )
        throw CyrCORBAException( error, where, status );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is responsible for removing an object reference from
//          the NameService.  If this is not done then errors can occur when
//          reconnecting to that object, as NameService still believes it
//          exists when in fact it could have been destroyed.
// @parm    The type of the object to deregister.
// @parm    The name of the object to deregister.
//
// @xref    <mf CORBAServer::RegisterObject>
void CORBAServer::DeregisterObject( const char* cls, const char* name )
{
    // No errors - yet
    const char* where = "CORBAServer::DeregisterObject";
    int status = CYR_UNKNOWN_ERROR;
    const char* error = NULL;

    // Try to resolve naming context.
    CosNaming::NamingContext_var context = GetNamingService();

    // Register object with given name
    CosNaming::Name cosName;
    cosName.length( 1 );
    cosName[0].id   = name;
    cosName[0].kind = cls;

    try
    {
        // Remove name from NameService
        context->unbind( cosName );
    }
    catch( CORBA::Exception )
    {
        // If we catch a CORBA exception here then it is likely that it we
        // are trying to unbind an obect that hasn't been bound.
        error = "Could not unbind object.  Is object registered?";
        status = CYR_CORBA_OBJECT_NOT_FOUND;
    }
    catch( ... )
    {
        // Catch any other exceptions and complain bitterly.
        error = "Unknown exception";
    }

    // If we have an error then throw an exception
    if ( error != NULL )
        throw CyrCORBAException( error, where, status );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method informs the BOA that an object is ready for use.  If
//          this code is not called then the ORB will not know about the object
// @parm    The object that is ready for use.
//
// @ex      Here is how to use ReadyObject to create a Factory that is not
//          registered with the NameService: |
//          ...
//
//          // Create a new Factory
//          FactoryImpl* factory = new FactoryImpl( "ACME", pServer->GetORB() );
//
//          // Ready factory
//          pServer->ReadyObject( factory );
//
//          ...
void CORBAServer::ReadyObject( Object_ptr obj )
{
    // Error information
    const char* where = "CORBAServer::ReadyObject";
    const char* error = NULL;

    try
    {
        // Inform ORB that object is ready to be accessed
        m_boa->obj_is_ready( obj );
    }
    catch( ... )
    {
        // Catch all exceptions
        error = "Could not ready object";
    }

    // If an error has occurred then throw an exception
    if ( error != NULL )
        throw CyrCORBAException( error, where, CYR_CORBA_OBJECT_NOT_READY );
}






