/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __MUTEX__
#define __MUTEX__

//
//  class CTMutex : Should be used to create and control a mutex.
//  -------------- - constructor : Construct a new mutex.
//                 - destructor : Destroy the mutex.
//                 - lock : lock the mutex. Wait if the mutex is locked by
//                   another thread.
//                 - unlock : unlock the mutex.
//                 - trylock : try to lock the mutex. Return TRUE if the mutex
//                   is locked by the calling thread and FALSE if the mutex is
//                   locked by another thread.
//  class CTSemaphore : Should be used to create and control a semaphore.
//  ------------------ - constructor : Construct a new semaphore.
//                     - destructor : Destroy the semaphore.
//                     - lock : decrease the semaphore by one. Wait if the
//                       counter is zero.
//                     - unlock : increase the semaphore.
//                     - trylock : try to decrease the semaphore. Return TRUE
//                       if the semaphore is decreased and FALSE if the counter
//                       is zero.
//


//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __EXCEPTION__
#  include <exception.hpp>
#endif  // __EXCEPTION__

#include <stddef.h>

#ifndef __MTHREAD__
#  include <mthread.hpp>
#endif  // __MTHREAD__

#ifndef __MTSAFE__
#  include <mtsafe.hpp>
#endif  // __MTSAFE__


//===================================================================< Macros >
#define DLLSPEC_MUTEX   DLLSPEC (CYRTOOLS)


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Class CTSynchronizationException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep synchronization exception.
class DLLSPEC_MUTEX CTSynchronizationException : public CTThreadException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTSynchronizationException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTThreadException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTSynchronizationException& exception)
    {
        CTThreadException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTSynchronizationException

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// End of Class CTSynchronizationException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////// Class CTMutexException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep mutex exception.
class DLLSPEC_MUTEX CTMutexException : public CTSynchronizationException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTMutexException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTSynchronizationException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTMutexException& exception)
    {
        CTSynchronizationException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTMutexException

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////// End of Class CTMutexException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Class CTSemaphoreException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep semaphore exception.
class DLLSPEC_MUTEX CTSemaphoreException : public CTSynchronizationException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTSemaphoreException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTSynchronizationException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTSemaphoreException& exception)
    {
        CTSynchronizationException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTSemaphoreException

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// End of Class CTSemaphoreException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////// Class CTSyncObject //
///////////////////////////////////////////////////////////////////////////////
// Class used to create and control a synchronization object.
class DLLSPEC_MUTEX CTSyncObject : virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTSyncObject ()
: pSyncObjectInfo_ (NULL)
    { }

    ~CTSyncObject ()
    { }


public:             // Get methodes.
    void*           getSyncObjectInfo (void)
    {
        return pSyncObjectInfo_ ;
    } ;

public:             // Control methodes.
    virtual void    lock (void) = 0 ;
    virtual bool    trylock (void) = 0 ;
    virtual void    unlock (void) = 0 ;

    //-----------------------------------------------------------------------------
protected:          // Members.
    void*           pSyncObjectInfo_ ;

    //-----------------------------------------------------------------------------
}

; // CTSyncObject

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////// End of Class CTSyncObject //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////// Class CTMutex //
///////////////////////////////////////////////////////////////////////////////
// Class used to create and control a mutex.
class DLLSPEC_MUTEX CTMutex : public CTSyncObject
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTMutex () ;
    ~CTMutex () ;

public:             // Get methodes.
    void*           getMutexInfo (void)
    {
        return pSyncObjectInfo_ ;
    } ;

public:             // Control methodes.
    void            lock (void) ;
    bool            trylock (void) ;
    void            unlock (void) ;

    //-----------------------------------------------------------------------------
}

; // CTMutex

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////// End of Class CTMutex //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Class CTSemaphore //
///////////////////////////////////////////////////////////////////////////////
// Class used to create and control a semaphore.
class DLLSPEC_MUTEX CTSemaphore : public CTSyncObject
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTSemaphore (long lInitialCount = 0L, long lMaximumCount = 0L) ;
    ~CTSemaphore () ;

public:             // Get methodes.
    void*           getSemaphoreInfo (void)
    {
        return pSyncObjectInfo_ ;
    } ;

public:             // Control methodes.
    void            lock (void) ;
    bool            trylock (void) ;
    void            unlock (void) ;

    //-----------------------------------------------------------------------------
}

; // CTSemaphore

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Class CTSemaphore //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Template Class CTLockGuard //
///////////////////////////////////////////////////////////////////////////////
// Template class used to lock a synchronization object at construction time,
// and release it at destruction time. This simplify synchronization object
// acquisition and release, and it safe the presence of exceptions.
template <class T> class CTLockGuard
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTLockGuard (T& t)
: t_ (t)
    {
        t_.lock () ;
    } ;
    ~CTLockGuard ()
    {
        t_.unlock () ;
    } ;

    //-----------------------------------------------------------------------------
private:            // Operators.
    CTLockGuard&    operator= (const CTLockGuard& lockGuard) ;

private:            // Members.
    T&              t_ ;

    //-----------------------------------------------------------------------------
}

; // CTLockGuard

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// End of Template Class CTLockGuard //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////// Template Class CTTryLockGuard //
///////////////////////////////////////////////////////////////////////////////
// Template class used to try to lock a synchronization object at construction
// time, and release it (if locked) at destruction time. This simplify
// synchronization object acquisition and release, and it safe the presence of
// exceptions.
template <class T> class CTTryLockGuard
{
    //-----------------------------------------------------------------------------
public:                 // Constructor(s) and destructor.
    CTTryLockGuard (T& t, bool& bLocked)
: t_ (t), bLocked_ (bLocked)
    {
        bLocked_ = t_.trylock () ;
    } ;
    ~CTTryLockGuard ()
    {
        if (bLocked_) t_.unlock () ;
    } ;

    //-----------------------------------------------------------------------------
private:                // Operators.
    CTTryLockGuard&     operator= (const CTTryLockGuard& tryLockGuard) ;

private:                // Members.
    T&                  t_ ;
    bool&               bLocked_ ;

    //-----------------------------------------------------------------------------
}

; // CTTryLockGuard

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// End of Template Class CTTryLockGuard //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////// Template Class CTUnlockGuard //
///////////////////////////////////////////////////////////////////////////////
// Template class used to unlock a synchronization object at construction time,
// and lock it at destruction time. This simplify synchronization object
// acquisition and release, and it safe the presence of exceptions.
template <class T> class CTUnlockGuard
{
    //-----------------------------------------------------------------------------
public:                 // Constructor(s) and destructor.
    CTUnlockGuard (T& t)
: t_ (t)
    {
        t_.unlock () ;
    } ;
    ~CTUnlockGuard ()
    {
        t_.lock () ;
    } ;

    //-----------------------------------------------------------------------------
private:                // Operators.
    CTUnlockGuard&      operator= (const CTUnlockGuard& unlockGuard) ;

private:                // Members.
    T&                  t_ ;

    //-----------------------------------------------------------------------------
}

; // CTUnlockGuard

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// End of Template Class CTUnlockGuard //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// Template Class CTMonitor //
///////////////////////////////////////////////////////////////////////////////
// Template class used to synchronize access to a shared resource.
template <class T> class CTMonitor
{
    //-----------------------------------------------------------------------------
public:             // Types.
    typedef CTLockGuard<T>      LockGuard ;
    typedef CTTryLockGuard<T>   TryLockGuard ;
    typedef CTUnlockGuard<T>    UnlockGuard ;

public:             // Constructor(s) and destructor.
    CTMonitor ()
    { }
    ;
    ~CTMonitor ()
    { }
    ;

public:             // Get methodes.
    T&              monitor ()
    {
        return t_ ;
    } ;

    //-----------------------------------------------------------------------------
private:            // Members.
    T               t_ ;

    //-----------------------------------------------------------------------------
}

; // CTMonitor

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////// End of Template Class CTMonitor //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////// Class CTConditionException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep condition exception.
class DLLSPEC_MUTEX CTConditionException : public CTThreadException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CTConditionException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTThreadException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CTConditionException& exception)
    {
        CTThreadException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CTConditionException

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// End of Class CTConditionException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Class CTCondition //
///////////////////////////////////////////////////////////////////////////////
// Class used to synchronize execution on a condition.
#ifdef  HPUX
#  ifdef    wait
#  undef    wait
#  endif    // wait
#endif  // HPUX

class DLLSPEC_MUTEX CTCondition : virtual public CTMTSafe
{
    //-----------------------------------------------------------------------------
public:             // Constructor(s) and destructor.
    CTCondition (CTMutex& mutex) ;
    ~CTCondition () ;

public:             // Control methodes.
    void            wait (void) ;
    void            signal (void) ;
    void            broadcast (void) ;

    //-----------------------------------------------------------------------------
private:            // Members.
    void*           pConditionInfo_ ;

    //-----------------------------------------------------------------------------
}

; // CTCondition

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Class CTCondition //
///////////////////////////////////////////////////////////////////////////////

#endif  // __MUTEX__
