/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     VDK_MODULE
// @module  Cyrano Provider VDK | This module is the implementation of a VDK
//          module.  These modules are responsible for creating and
//          initialising Providers.
//
//          Typically VDKs will be loaded by the Cyrano Daemon applciation
//          and utilise the Factory within that application.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Repository/CyrStProv.h>
#include <Factory/RegistryProvider.h>
#include <Daemon/CyrVdk.h>
#include <Logging/CECLog.h>
#include <CECInetInfo.h>
///////////////////////////////////////////////////////////////////////////////
// Global variables
// @globalv CyrStProvider | pProvSt | Pointer to Provider of OpenSTARep data.
CyrStProvider* pProvSt;
RegistryProvider* pProvReg;
extern CECLog* v2logger = NULL;

///////////////////////////////////////////////////////////////////////////////
// @func   This function initialises the Provider objects within the DLL.  It
//         is also responsible for any other setup stuff.
// @rdesc  A standard Cyrano error code.
// @parm   The name that all Providers within this DLL will register themselves
//         with.
// @parm   A reference to CORBA helper functionality.
VDK_EXP CyrVdkInitialise( const char* factory, CORBAServer* pServer )
{
    // Create a new collection of Providers
    string name = factory;
    name += "_CyrStProvider_001";
    pProvSt = new CyrStProvider( pServer, name.c_str(), factory );

    name = factory;
    name += "_RegistryProvider_001";
    pProvReg = new RegistryProvider( pServer, name.c_str(), factory );

    // Create a new logging object...

    // Get the host name
    CECInetInfo inet;
    string hostname=inet.GetIPAddressAsString();

    // From that determine the Logfile name
    string logName = hostname;
    logName += "_";
    logName += CYR_VDK002;

    // And default the tracing to 1000
    int trace = 1000;
    v2logger = new CECLog( pServer, logName.c_str(), trace );

    // And start tracing
    std::string msg="CyrVdk002 Initialising, Log Name: " + logName ;
    v2logger->AddTrace( CYR_VDK002, 10, msg.c_str() );

    return CYR_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
// @func  This function is responisble for destroying any Provider objects
//        created in the <f CyrVdkInitialise> function.  It also performs any
//        other required tidying up.
// @rdesc A standard Cyrano error code.
VDK_EXP CyrVdkUninitialise()
{
    // And add some tracing
    v2logger->AddTrace( CYR_VDK002, 100, "Shutting down the providers." );

    // Dispose of any providers
    if ( !CORBA::is_nil( pProvSt->_this() ))
    {
        pProvSt->Shutdown();
        pProvSt->_remove_ref();
    }

    // And add some tracing
    v2logger->AddTrace( CYR_VDK002, 100, "Releasing pointers to Corba." );

    if ( !CORBA::is_nil( pProvReg->_this() ))
    {
        pProvReg->Shutdown();
        pProvReg->_remove_ref();
    }

    // And add some tracing
    v2logger->AddTrace( CYR_VDK002, 10, "Closing Log File." );

    // Get rid of logging object
    if ( v2logger != NULL )
    {
        delete v2logger;
        v2logger = NULL;
    }

    return CYR_SUCCESS;
}
