/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Provider
// @module  Provider | This module contains the implementation of the ProviderImpl
//          class.  ProviderImpl is the base class for all Provider objects and
//          contains all generic Provider functionality. It also contains
//          implementation for any helper classes that is required to sort,
//          merge and handle the leasing of objects.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Factory/Provider_i.h>
using namespace CORBA;

///////////////////////////////////////////////////////////////////////////////
// The following are a collection of helper functions that will assist with the
// LEASED_OBJECT management.

void SubtractPeriod( PLEASED_OBJECT& obj, void* data )
{
    (obj->period)--;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor is responsible for starting the CORBA client and
//          server services for this object as well as registering the object
//          with the parent Factory and the NameService.
// @parm    Pointer to CORBA server helper functionality.
// @parm    The name of the Provider.
// @parm    The name of the parent Factory.
ProviderImpl::ProviderImpl( CORBAServer* pServer, const char* name, const char* owner )
{
    // Prepare client and server connections
    m_pServer = pServer;

    // Register Provider object with NameService
    m_name = name;

    // Retrieve a connection to the Factory
    CORBA::Object_ptr obj =
        m_pServer->FindObjectWithName( CyrFactory::CLASS_NAME, owner );
    m_factory = CyrFactory::_narrow(obj);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is responsible for destroying any allocated memory
//          and shutting down the CORBA services.
void ProviderImpl::Shutdown()
{
    // Deregister all classes from factory
    DeregisterAll( m_name.c_str() );

    // Deregister Provider
    m_pServer->UnNameObject( Provider::CLASS_NAME, m_name.c_str() );

    // Destroy all created objects that belong to this Provider
    EmptyList( 0, m_lstObj.size(), false );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method deletes entries from the object list.  It disposes of
//          the object that is pointed to and also deletes the
//          <t LEASED_OBJECT> within the list.  Finally the entries are
//          removed from the list.  Note that this might invalidate any
//          iterators currently pointing to elements within the designated
//          section.
// @parm    The first element in the list.
// @parm    The element just after the end of the list.
// @parm    Determines whether entries should be destroyed before their lease
//          is up.
void ProviderImpl::EmptyList( long start, long stop, bool reap )
{
    // Loop through all entries and delete the dead ones.
    for ( long i = stop - 1; i >= start; i-- )
    {
        if (( m_lstObj[i]->period <= 0 ) || ( reap == false ))
        {
		    m_lstObj[i]->obj->_remove_ref(); // Dispose of the CORBA object
		    //delete m_lstObj[i];               // Delete LEASED_OBJECT structure
		    m_lstObj[i] = NULL;
        }
    }

    for(ObjectList::iterator j=m_lstObj.begin();j!=m_lstObj.end();j++)
    {
        if(*j==NULL)
            m_lstObj.erase(j);
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method registers a single class with the parent Factory.
// @parm    The name of the class to register.
// @parm    The name of the Provider (as registered with the NameService).
// @parm    The type of the class.
void ProviderImpl::Register( const char* cls, const char* provider, long type )
{
    // Register class with main factory
    m_factory->Register( cls, provider, type );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method deregisters a single class from the parent Factory.
// @parm    The name of the class to deregister.
// @parm    The name of the Provider (as registered with the NameService).
void ProviderImpl::Deregister( const char* cls, const char* provider )
{
    // Dergister class from main factory
    m_factory->Deregister( cls, provider );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method deregisters all classes from the parent Factory.
// @parm    The name of the Provider (as registered with the NameService).
void ProviderImpl::DeregisterAll( const char* provider )
{
    // Deregister all classes from main factory
    m_factory->DeregisterAll( provider );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is used to store a CORBA object in the Provider's
//          store. If these objects are to be destroyed later then this should
//          be done.
// @parm    The CORBA object pointer.
// @parm    The type of the class being registered.
// @parm    The name of the object as registered with the NameService.
// @parm    The length of time in minutes that the object will exist for.
void ProviderImpl::Store( PortableServer::RefCountServantBase* ref, const char* cls,
                          const char* name, long period )
{
    // Create a new object reference structure
    PLEASED_OBJECT p = new LEASED_OBJECT;
    p->obj = ref;
    p->cls = cls;
    p->name = name;
    p->period = period;

    // Add this object to the list
    m_lstObj.push_back( p );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method destroys any objects that are out of date.  All stored
//          objects are in time order, with the shorter time objects appearing
//          at the start of the list.  Every ten minutes the dead object at
//          the beginning of the list are thrown out.
void ProviderImpl::Reaper()
{
    // If there are no elements then it is pointless continuing with this
    // method and we might as well return now.
    if ( m_lstObj.empty() )
        return;

    // Subtract a single period from all objects and reap any dead objects
    for(ObjectList::iterator j=m_lstObj.begin();j!=m_lstObj.end();j++)
    {
        SubtractPeriod(*j,NULL);
    }
    EmptyList( 0, m_lstObj.size(), true );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method disposes of an object within the Provider's list.  This
//          ignores any leasing periods and therefore should be used with
//          caution.
// @parm    The class of the object to update.
// @parm    The name of the object to update.
void ProviderImpl::Dispose( const char* cls, const char* name )
{
    // Locate correct object
    size_t index = Find( 0, m_lstObj.size(), cls, name );
    if ( index == m_lstObj.size() )
        return;

    // Dispose of object.
    EmptyList( index, index + 1, false );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method increases the length of time an object has before its
//          lease runs out.  Given the class and the name of the object this
//          method locates the object, updates the lease and then re-inserts
//          the element in the correct location in the object list.
//          Re-insertion of the object is done to avoid having to re-sort the
//          entire list.
// @parm    The class of the object to update.
// @parm    The name of the object to update.
// @parm    The number of periods to extend lease by.
void ProviderImpl::Borrow( const char* cls, const char* name, long period )
{
    // Locate correct object
    size_t index = Find( 0, m_lstObj.size(), cls, name );
    if ( index == m_lstObj.size() )
        return;

    // Update lease period
    m_lstObj[index]->period += period;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method locates a specific entry in the object list.
// @rdesc   A pointer to the located element.
// @parm    The start of the specified range.
// @parm    The end of the specifid range.
// @parm    The class of the object to update.
// @parm    The name of the object to update.
size_t ProviderImpl::Find( size_t start, size_t stop, const char* cls,
                           const char* name )
{
    // Iterate through list and locate entry we require and increase
    // lease period.
    for ( size_t i = start; i < stop; i++ )
    {
        // Do we have right entry?
        if (( m_lstObj[i]->name == name ) && ( m_lstObj[i]->cls == cls ))
            return i;
    }

    return stop;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc  This method allows simplistic logging functionality for events that
//         occur within a Provider.  The event string is passed on to the
//         parent Factory where it will be handled appropriately.
// @parm   The level of the event string.
// @parm   The event string to log.
void ProviderImpl::LogEvent( long level, const char* event )
{
    // Simply pass event string on to Factory for logging
    m_factory->LogEvent( level, event );
}

