/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     Logging
// @module  Cyrano Logging | This module contains an implementation of a
//          Cyrano logging class.  This particular implementation saves the
//          logging information to a file in the same directory as the
//          currently running application.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <definition.h>
#include <CyrHelper.inl>
#include <Logging/CyrLogFile.h>
#include <Factory/Factory.h>
#include <Factory/ParamList_i.h>

using std::fstream;
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor creates a log file in the same directory as the
//          currently running application.  All other initialisation is
//          performed by the base constructor.
// @parm    Pointer to a CORBA helper class.
// @parm    The default logging level.
CyrLogFile::CyrLogFile( CORBAServer* pServer, long level )
{
    // Determine current node name from host address
    CECInetInfo host;
    m_node = host.GetHostName();

    // Get current executable name
    m_exec = CyrGetModuleName();
    if ( m_exec == "" )
        m_exec = "Unknown";

    // Get current user name
    m_user = CyrGetCurrentUser();

    // Assign remaining member variables
    m_pServer = pServer;
    m_defaultLevel = level;
    m_iterator = m_traceList.begin();

    // Calculate log file name
    m_filename = CyrGetLogFileName();

    // Open output file
    m_file = new fstream( m_filename.c_str(), std::ios::binary|std::ios::app|std::ios::out  );

    // write out the file header:
    // Build trace string
    string str( "" );

    // Set Title
    str += "   CYRANO Logfile ";
    str += SEPARATOR;

    // Add user, node and executor
    str += m_node;
    str += SEPARATOR;
    str += m_user;
    str += SEPARATOR;
    str += m_exec;
    str += "\r\n\n";
    m_file->write( str.c_str(), str.length() );

    // Flush file contents to disk - otherwise nothing may be written
    // if application crashes.
    m_file->flush();

    // Log event
    AddTrace( CYR_LOG, 1, "New log started" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Before the list is emptied, this destructor writes any remaining
//          entries to disk.
CyrLogFile::~CyrLogFile()
{
    AddTrace( CYR_LOG, 1, "Log stopped" );

    // Loop through all remaining entries and write them to disk
    if ( m_file->is_open() )
    {
        for ( TraceList::iterator i = m_traceList.begin(); i != m_traceList.end(); i++ )
        {
            string entry = *i;
            m_file->write( entry.c_str(), entry.length() );
            m_file->write( "GHx \r\n", 6 );
        }
    }

    // Close file and flush contents to disk
    delete m_file;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method sets the current logging level.  If the level cannot
//          be found then loggin is switched off.
// @parm    The name of the new loggin level.
void CyrLogFile::SetLevel( const char* category, long level )
{
    // Set level in array.  As a side effect the category is added if it
    // does not exist.
    string cat(category);
    m_catLevels[ cat ] = level;

    char buf[20];
    sprintf( buf, "%d", level );
    string event( "Logging level of " );
    event += category;
    event += " has been set to ";
    event += buf;
    AddTrace( CYR_LOG, 2, event.c_str() );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the current logging level.
// @rdesc   The current logginglevel.
long CyrLogFile::GetLevel( const char* category )
{
    // Get level from array
    return m_catLevels[string(category)];
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a single trace entry to the list of trace entries.
//          All generic information is determined in the constructor, so the
//          missing entries are the level, the location and the actual entry.
//          If the specified logging level is below the current trace level the
//          entry is ignored.  Otherwise the entry is concatenated into a
//          single string, each field separated by a single '\|' character.
// @parm    The category of the entry
// @parm    The level the entry should be added at.
// @parm    The actual log entry.
void CyrLogFile::AddTrace( const char* category, long level, const char* msg )
{
    // If category does not exist within array then add it at default level
    if ( m_catLevels.find( category )==m_catLevels.end())
        SetLevel( category, m_defaultLevel );

    // Determine if log entry should be added
    int logLevel = m_catLevels[ string(category) ];
    if ( logLevel < level )
        return;

    // Build trace string
    string str( "" );
    char buf[20];

    // Add time in seconds
    CTTime Tm;
    Tm.now();
    sprintf( buf, "%d", Tm.getSeconds() );
    str += buf;
    str += SEPARATOR;

    // Add user, node and executor
    //str += m_node; str += SEPARATOR;
    //str += m_user; str += SEPARATOR;
    //str += m_exec; str += SEPARATOR;

    // Add category, level and event
    str += category;
    str += SEPARATOR;
    sprintf( buf, "%d", level );
    str += buf;
    str += SEPARATOR;
    str += msg;
    str += SEPARATOR;

    // Add string to trace elements
    m_traceList.push_back( str );


    // If current number of entries exceeds total then flush
    FlushEntries();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a list of strings representing the categories
//          of tracing currently understood by this logging object.
// @rdesc   The list of strings.
TraceCat* CyrLogFile::GetTraceCategories()
{
    size_t count = m_catLevels.size();

    // Create sequence of trace levels for transmission
    TraceCat_var traceSeq = new TraceCat();
    traceSeq->length( count );

    // Loop through all trace levels and add them to sequence
    TraceLevelMap::iterator i;
    count = 0;
    for ( i = m_catLevels.begin(); i != m_catLevels.end(); i++ )
    {
        traceSeq[count] = CORBA::string_dup( i->first.c_str() );
        count++;
    }

    // Return reference and inform client that it has responsiblity for
    // deallocating memory.  Code copied from Henning and Vinoski pg 423.
    return traceSeq._retn();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the first trace entry in the current list.
// @rdesc   The first trace entry.
char* CyrLogFile::GetFirstTrace()
{
    m_iterator = m_traceList.begin();
    return GetNextTrace();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Returns the next trace entry in the current list.
// @rdesc   The next trace entry or an empty string if at end of list.
char* CyrLogFile::GetNextTrace()
{
    if ( m_iterator !=  m_traceList.end() )
    {
        return CORBA::string_dup((*m_iterator++).c_str());
    }
    else
        return CORBA::string_dup( "" );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method overrides the base method for flushing entries and
//          ensures that they are written to disk before being discarded.
void CyrLogFile::FlushEntries()
{
    // Write last entry to output file
    if ( m_file->is_open() )
    {
        string entry = m_traceList.back();
        entry += "\r\n";
        m_file->write( entry.c_str(), entry.length() );

        // Flush file contents to disk - otherwise nothing may be written
        // if application crashes.
        m_file->flush();
    }

    // The next section simply reduces the size of the list
    // i.e. each entry is written individually but the list
    // is only shortened once we reach MIN_FLUSH

    // If we have less than a hundred entries do nothing.
    if ( m_traceList.size() < MIN_FLUSH )
        return;

    // Remove first 50 entries from trace list
    for ( size_t i = 0; i < NUM_FLUSH; i++ )
        m_traceList.pop_front( );

    m_iterator = m_traceList.begin();
}

////////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the full path name of the log file being
//          generated.
// @rdesc   The full path name of log file.
string CyrLogFile::GetFullPath()
{
    return m_filename;
}

