/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc CyranoHelper
///////////////////////////////////////////////////////////////////////////////

#ifndef _CyrHelper_inl
#define _CyrHelper_inl

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes

///////////////////////////////////////////////////////////////////////////////
// Standard includes
#ifdef _WIN32
#include <windows.h>
#include <lmcons.h>
#include <io.h>
#include <process.h>
#endif

#if defined( UNIX )
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <fnmatch.h>
#include <pwd.h>
#include <unistd.h>
#include <limits.h>
#include <stdlib.h>
#endif

///////////////////////////////////////////////////////////////////////////////
#include <string>
#include <list>
using std::string;
using std::list;

///////////////////////////////////////////////////////////////////////////////
// Type definitions
// @type    NameList | A list of library module names.
typedef list<string>    NameList;

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function returns a full specified temporary file name.
// @rdesc   The full path to a temporary file.
inline string CyrGetTempFileName()
{
    string filename( "" );

#ifdef _WIN32
    // Create temporary file name
    string dir( "" );
    char buf[MAX_PATH];
    GetTempPath( MAX_PATH, buf );
    dir = buf;
    GetTempFileName( dir.c_str(), "Cyr", 0, buf );
    filename = buf;
#endif

#ifdef UNIX
    char buf[L_tmpnam+1];
    tmpnam(buf);
    filename = buf;
#endif

    return filename;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function deletes a file.
// @parm    The full path name to the file to be deleted.
inline void CyrDeleteFile( const char* filename )
{
#if defined (_WIN32) || defined( UNIX )
    remove( filename );
#endif
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function returns the name of the current executable.
// @rdesc   The full path name of the executable.
inline string CyrGetModuleName()
{

#ifdef _WIN32
    // Set current executable name
    string module( "" );
    unsigned long length = UNLEN + 1;
    char* buf = new char[length];

    HMODULE hModule = GetModuleHandle( NULL );
    if ( hModule != NULL )
    {
        // Get full path name of executable
        GetModuleFileName( hModule, buf, length );
        module = buf;
    }

    delete [] buf;
    return module;
#endif

#ifdef UNIX

    // two ways, neither easy

    // 1) the filename through which the program
    // was invoked is available as argv[0] to main().
    // will need to be saved at the start of main() in every program.

    // 2) per system calls

#ifdef Linux
    char * path=new char[PATH_MAX];
    int stat=readlink("/proc/self/exe", path, PATH_MAX-1);

    if ( stat < 0 )
    {
        /* alanw - TODO - find out how to return an error */
        path[0] = '\0';
    }
    else
    {
        path[stat] = '\0';
    }

    string Cmd(path);
    delete[] path;
    return Cmd;
#endif

#ifdef SOLARIS
    // read from struct in /proc/self/psinfo

    // GH PW code
    string execname( "" );
    string module( "" );
    string cwd( "" );

    execname = getexecname();

    // if / first then this is absolute pathname

    if(execname[0] != DIR_SLASH)
    {
        cwd = getcwd(NULL, PATH_MAX);
        module = cwd + DIR_SLASH + execname;
    }
    else
    {
        module = execname;
    }
    return module;
#endif

#ifdef HPUX
    // call pstat_getproc, use .pst_ucomm
    // TODO
    string module( "" );
    return module
#endif

#ifdef AIX
    // GOK
    // TODO
    string module( "" );
    return module
#endif

#endif	// UNIX
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function returns the name of a log file in the same directory
//          as the current executable.
// @rdesc   The log file name
inline string CyrGetLogFileName()
{
    string filename( "" );

#ifdef _WIN32
    // Remove filename from string
    char buf[20];
    filename = CyrGetModuleName();
    string::size_type sz=filename.find_last_of( '.' );
    filename.resize(sz);
    sprintf( buf, "_%d", _getpid() );
    filename += buf;
#endif

#ifdef UNIX

    // Remove filename from string
    char buf[20];
    filename = CyrGetModuleName();
    sprintf( buf, "_%d", getpid() );
    filename += buf;
#endif

    filename += ".log";
    return filename;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function returns the name of a statfile in the same directory
//          as the current executable.
// @parm	IP Address
// @parm	path - true return includes path
// @rdesc   The stat file name
inline string CyrGetStatFileName(const char* buf, bool path)
{
    string filename( "" );

#ifdef _WIN32
    char tbuf[128];

    // if we want the path
    if(path)
    {
        // Take the actual executer module name ...
        filename = CyrGetModuleName();

        // And find the last slash in the pathname returned from CyrGetModuleName
        filename.resize( filename.find_last_of( '\\' ) + 1);
    }
    else
    {
        filename = "";
    }

    // And generate a filename from the pid...
    sprintf( tbuf, "TEW_%s-%d",buf,_getpid());
    filename += tbuf;
#endif

#ifdef UNIX

    char tbuf[128];
    // if we want the path
    if(path)
    {
        // Take the actual executer module name ...
        filename = CyrGetModuleName();

        // And find the last slash in the pathname returned from CyrGetModuleName
        filename.resize( filename.find_last_of( '/' ) + 1);
    }
    else
    {
        filename = "";
    }
    // And generate a filename from the pid...
    sprintf( tbuf, "TEW_%s-%d",buf,getpid());

    //    sprintf( tbuf, "_%d-%d", getpid(),buf );
    filename += tbuf;
#endif

    filename += ".stat";
    return filename;
}




///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function returns the name of the currently logged on user.
// @rdesc   The user name.
inline string CyrGetCurrentUser()
{
    // Get current user name
    char buf[256];

#ifdef _WIN32
    DWORD length = 256;
    GetUserName( buf, &length );
#endif;

#ifdef UNIX
#if defined( HPUX)
    struct passwd ppwd;
    int rv = getpwuid_r(getuid(), &ppwd, buf, sizeof( buf ));
    if ( rv < 0 )
        sprintf( buf, "unknown" );
#elif defined(SOLARIS) && !defined(_POSIX_PTHREAD_SEMANTICS)
struct passwd ppwd,*ppwd2;
    ppwd2 = getpwuid_r(getuid(), &ppwd, buf, sizeof( buf ));
    if ( ppwd2 ==NULL)
        sprintf( buf, "unknown" );
#else
    struct passwd pwd, *ppwd;
    int rv = getpwuid_r(getuid(), &pwd, buf, sizeof( buf ), &ppwd);
    if ( rv < 0 || ppwd == NULL )
        sprintf( buf, "unknown" );
#endif
#endif;

    return string( buf );
}

inline NameList CyrFindFiles( const char* dir, const char* pattern )
{
    NameList names;

    // Calculate path for module
    string path( dir );

#ifdef _WIN32
    // Locate first Provider module that matches pattern.  If none do then
    // just return.  This is NOT an error condition as Providers may be located
    // elsewhere and will register later.
    _finddata_t fdt;
    long handle = _findfirst( (path+ pattern).c_str(), &fdt );
    if ( handle == -1 )
        return names;

    // Locate all Provider libraries that match the pattern
    int finish = 0;
    while ( finish == 0 )
    {
        // Add library name to list
        if ( strlen( pattern ) == strlen( fdt.name ))
            names.push_back( path + fdt.name );

        // Find next matching entry
        finish = _findnext( handle, &fdt );
    }

    // Free any resources related to file search
    _findclose( handle );
#endif

#ifdef UNIX

    printf("DIR = %s\n",dir);
    DIR* pdir = opendir(dir);
    if ( pdir == NULL )
        return names;

#if defined( HPUX)
    struct dirent ent;
    int rv = readdir_r(pdir, &ent);
    while( rv == 0 )
    {
        rv = fnmatch(pattern, ent.d_name, 0);
        if ( rv == 0 )
        {   // match
            names.push_back( path + ent.d_name );
            printf("CyrFindFiles :  %s  -->  which matches !!!!!!!\n",ent.d_name);
        }
        rv = readdir_r(pdir, &ent);
    }

#elif defined (SOLARIS)
    struct dirent *ent;
    int rv;

    while ((ent = readdir(pdir)) != NULL )
    {
        rv = fnmatch(pattern, ent->d_name, 0);
        if ( rv == 0 )
        {   // match
            names.push_back( path + ent->d_name );
            printf("1 CyrFindFiles :  %s  -->  which matches !!!!!!!\n",ent->d_name);
        }
        else
        {
            printf("1 CyrFindFiles :  %s  -->  which DOES NOT match !!!!!!!\n",ent->d_name);
        }
    }
#else
    struct dirent ent,*presult;
    int rv;

    while ((readdir_r(pdir, &ent, &presult))==0)
    {
        printf("ent = 0x%x\n",ent);
        printf("presult = 0x%x\n",presult);
        rv = fnmatch(pattern, ent.d_name, 0);
        if ( rv == 0 )
        {   // match
            names.push_back( path + ent.d_name );
            printf("2 CyrFindFiles :  %s  -->  which matches !!!!!!!\n",ent.d_name);
        }
        else
        {
            printf("2 CyrFindFiles :  %s  -->  which DOES NOT match !!!!!!!\n",ent.d_name);
        }
    }

    closedir(pdir);
#endif
#endif

    return names;
}

#endif

