/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#ifndef __GATEWAY_PARSE__
#define __GATEWAY_PARSE__


//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __DEFINITION__
#  include "definition.h"
#endif  // __DEFINITION__

#ifndef __MTSAFE__
#  include <mtsafe.hpp>
#endif  // __MTSAFE__

#ifndef __SOCKET__
#  include <socket.hpp>
#endif  // __SOCKET__


#ifndef __LOGFILE__
#  include <logfile.hpp>
#endif  // __LOGFILE__

#ifndef __CYRTOOLS_TIME__
#  include <ctime.hpp>
#endif  // __CYRTOOLS_TIME__


//===================================================================< Macros >
#ifndef DLLSPEC_GWPARSE
#  ifdef BUILDDLLGWORACLE
#    define DLLSPEC_GWPARSE __EXPORT__
#  else
#    define DLLSPEC_GWPARSE
#  endif    // BUILDDLLGWORACLE
#endif  // DLLSPEC_GWPARSE


//======================================================< Definition of class >

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////// Class CGWParseException //
///////////////////////////////////////////////////////////////////////////////
// Class used to keep class parse exception.
class CGWParseException : public CTException
{
    //-----------------------------------------------------------------------------
public:                     // Constructor(s) and destructor.
    CGWParseException (const string& clMsg, const string& clFilename, short sLineno, long lNumber = -1L)
: CTException (clMsg, clFilename, sLineno, lNumber)
    { }


public:                     // Operators.
    void                    operator= (const CGWParseException& exception)
    {
        CTException::operator= (exception) ;
    }

    //-----------------------------------------------------------------------------
}

; // CGWParseException

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// End of Class CGWParseException //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////// Class CGWParse //
///////////////////////////////////////////////////////////////////////////////
class DLLSPEC_GWPARSE CGWParse : public CTMTSafe
{
    //-----------------------------------------------------------------------------
    // Define Data.
#define NEWGATEWAYPARSE         "LGW_NewGatewayParse"   // Name of the library function to be used to instantiate the GWParse class.
#define SETCMDLINEARG           "LGW_SetCmdLineArg"     // Name of the library function to be used to pass command line arguments.
#define GATEWAYPARSEVERSION     "LGW_GWParseVersion"    // Name of the library version variable.

#define FROM_CLIENT             (int) 1 // Packet from client.
#define FROM_TARGET             (int) 2 // Packet from server.

public:                         // Types.
    // Type of the function used to instantiate the GWParse class.
    typedef void (*fnNewGatewayParse) (const void*, const string&, bool&, CTOLogFileList&, CTSocket*&, CTSocket*&, CGWParse*&) ;
    // Type of the function used to pass command line arguments.
    typedef void (*fnSetCmdLineArg) (const string&) ;

public:                         // Constructor(s) and destructor.
    // Construct the CGWParse object.
    CGWParse (const void* pConnectionId, const string& taskName, bool& bTraceOn, CTOLogFileList& logFileList, CTSocket*& pClient, CTSocket*& pTarget)
: pConnectionId_ ( (void*) pConnectionId), taskName_ (taskName), bTraceOn_ (bTraceOn), logFileList_ (logFileList), pClient_ (pClient), pTarget_ (pTarget)
    { }
    ;
    // Destroy the CGWParse object. Call the destructor methode if available.
    virtual ~CGWParse ()
    { }
    ;

public:                         // Operators.
    CGWParse&                   operator = (const CGWParse&) ;

public:                         // Control Methodes.
    void                        traceString (const string& str)
    {
        if (bTraceOn_)
        {
            CTLockGuard<CTMutex> guard (logFileList_.getMutex () ) ;
            logFileList_ << str ;
        }
    }
    void                        traceLine (const string& str)
    {
        if (bTraceOn_)
        {
            CTLockGuard<CTMutex> guard (logFileList_.getMutex () ) ;
            logFileList_ << str << endl ;
        }
    }
    void                        traceFlush (const string& str)
    {
        if (bTraceOn_)
        {
            CTLockGuard<CTMutex> guard (logFileList_.getMutex () ) ;
            logFileList_ << str << flush ;
        }
    }
    //// Required.
    // Gateway call this methode to parse the packet buffer 'pBuf' begining at
    // the offset 'lOffset', of a size 'lBufLen', coming from 'iDataFrom', and
    // which is 'bUrgent' data or not. The buffer 'pBuf' can be updated and
    // the size 'lBufLen' can be changed (from 0 bytes to a maximum of
    // 'lBufMaxLen' bytes). The output parameter 'lMoreData' can be used to
    // specify the gateway to receive 'lMoreData' data and add them at the end
    // of the buffer. This methode must return the size of the packet parsed
    // (because the buffer 'pBuf' can be one or more or a partial user
    // packet(s) ). This methode will be recalled for the same packet 'pBuf'
    // while all user packets are not parsed (while the returned size is less
    // than the buffer size 'lBufLen').
    virtual long                parseBuffer (const int iDataFrom, bool& bUrgent, const void* pBuf, long& lBufLen, const long lOffset, const long lBufMaxLen, long& lMoreData) = 0 ;
    // Optionnal.
    // Gateway call this methode to specify the library the receive time and
    // the send time for a packet.
    virtual void                setReceiveAndSendTimes (const CTTime&, const CTTime&)
    { }
    ;
    //// Pre-connection phase only !
    // Gateway call this methode to specify the library the new gateway listen
    // socket address 'gatewayListenAddress'. The library can then change the
    // packet 'pBuf' of a size 'lBufLen'. The buffer 'pBuf' can be updated and
    // the size 'lBufLen' can be changed (from 0 bytes to a maximum of
    // 'lBufMaxLen' bytes).
    virtual void                setGatewayListenAddress (string&, const void*, long&, const long)
    { }
    ;

public:                         // Get Methodes.
    //// Required.
    // Gateway call this methode to know if there is a pre-connection phase.
    virtual const bool          getPreconnectionMode (void) const = 0 ;
    //// Pre-connection phase only !
    // Gateway call this methode to know if the pre-connection phase is
    // terminated.
    virtual const bool          isConnected (void)
    {
        return true ;
    } ;
    // Gateway call this methode to get the new target listen socket address.
    // In a pre-connection phase, the library must get this from the packets.
    // Return if the 'new target listen socket address' information is
    // available.
    virtual const bool          getTargetListenAddress (string&)
    {
        return false ;
    } ;
    // Gateway call this methode to get the new maximum buffer size. In a
    // pre-connection phase, the library can get this from the packets. Return
    // the new maximum buffer size (-1 if not available).
    virtual const long          getNewMaxBufferSize (void)
    {
        return 0 ;
    } ;

    //-----------------------------------------------------------------------------
protected:                      // Members.
    void*                       pConnectionId_ ;            // Connection identifier.
    string                  taskName_ ;                 // Prefix to use when tracing.
    bool&                       bTraceOn_ ;                 // Reference to gateway trace flag used to know if trace is active.
    CTOLogFileList&             logFileList_ ;              // Reference to list of stream to trace into.
    CTSocket*&                  pClient_ ;                  // Reference to pointer to client socket.
    CTSocket*&                  pTarget_ ;                  // Reference to pointer to target socket.

    //-----------------------------------------------------------------------------
}

; // CGWParse

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// End of Class CGWParse //
///////////////////////////////////////////////////////////////////////////////


#endif  // __GATEWAY_PARSE__
