/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_CARDFILE_SUBS               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains a number of routines for recording useful       */
/*  addresses in a way that can be retrieved by PROCDMP.             */
/*                                       */
/*===========================================================================*/


#include  <psl_macros.h>        /* PSL Standard Macros */

/*    Define useful Include structures   */

#include  <string.h>            /* String functions */
#include  <stdlib.h>            /* Standard Library Routines */

/*    Define Function Prototypes for all external routines */
/*    used by the module */

#include  psl_common_file(psl_rtndef.h) /* Define ourselves */

/*    Define Function Prototypes for all internal routines */
/*    used by the module */

static struct  psl_card *LOCATE_ENTRY (char *text_ptr);

/*    Define the Cardfile Entry Structure */

struct  psl_card
{
    struct   psl_card   *fwdptr;    /* Queue forward pointer */
    struct   psl_card   *bckptr;    /* Queue backward pointer */
    char   prefix[4];  /* Prefix */
    char   descr[8];   /* Entry name */
    void   *data;      /* Pointer to data */
};

/*    Define the cardfile list header as common data */

static struct psl_card card_hdr =
    {
        NULL, NULL, {'*','C','D','*'},
        "**HDR**", NULL
    };

/************************************************************************/
/*                                  */
/*  PSL_CARDFILE_ADD - Add an entry to the Cardfile         */
/*                                  */
/*  This routine adds an entry to the Cardfile.         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    PSL_CARDFILE_ADD (data_ptr, text_ptr);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  data_ptr = Pointer to item to be logged in the cardfile     */
/*  text_ptr = Pointer to text to be associated with it.        */
/*                                  */
/************************************************************************/

void PSL_CARDFILE_ADD (void *data_ptr, char *text_ptr)
{

    /*   Local Data   */

    struct   psl_card   *new_ptr;       /* Pointer to new card */
    struct   psl_card   *old_ptr;       /* Pointer to old last card */
    size_t     txtlen;     /* Text length */

    /*   Allocate memory for the new card, exiting if none available */

    new_ptr = (struct psl_card *) malloc (sizeof(struct psl_card));
    if  (new_ptr == NULL) return;

    /*   Now fill in the user's details */

    (*new_ptr).data = data_ptr; /* Store data pointer */
    (void) memcpy ((*new_ptr).prefix, "*CD*", 4);
    /* Set up Prefix */
    (void) memset ((*new_ptr).descr, ' ', 8);
    /* Initialize description to spaces */
    txtlen = strlen (text_ptr); /* Get length of user text string */
    if  (txtlen > 8) txtlen = 8;    /* Constrain to eight characters */
    (void) memcpy ((*new_ptr).descr, text_ptr, txtlen);
    /* Copy in user data */

    /*   Lastly hook the new entry into the chain */

    if  (card_hdr.fwdptr == NULL)   /* If chain is empty */
    {
        card_hdr.fwdptr = new_ptr;  /* Set new entry as both first */
        card_hdr.bckptr = new_ptr;  /* and last */
        (*new_ptr).fwdptr = NULL;   /* And set no links from it */
        (*new_ptr).bckptr = NULL;
    }
    else                /* Else (chain not empty) */
    {
        old_ptr = card_hdr.bckptr;  /* Get last entry in chain */
        card_hdr.bckptr = new_ptr;  /* and reset last entry */
        (*old_ptr).fwdptr = new_ptr; /* Reset next pointer in old last */
        (*new_ptr).bckptr = old_ptr; /* And vice versa */
        (*new_ptr).fwdptr = NULL;   /* And no forward pointer */
    }
    return;             /* Return to caller */
}

/************************************************************************/
/*                                  */
/*  PSL_CARDFILE_REMOVE - Remove an entry from the Cardfile     */
/*                                  */
/*  This routine removes an entry from the Cardfile.        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    PSL_CARDFILE_REMOVE (text_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*  text_ptr = Pointer to text associated with item to remove.  */
/*                                  */
/************************************************************************/

void PSL_CARDFILE_REMOVE (char *text_ptr)
{

    /*   Local Data   */

    struct   psl_card   *ent_ptr;       /* Pointer to located card */
    struct   psl_card   *old_ptr;       /* Pointer to old last/next card */

    /*   Locate the entry via the standard routine */

    ent_ptr = LOCATE_ENTRY (text_ptr);
    /* Locate the entry */

    /*   If we found it, remove it from the chain and delete it */

    if  (ent_ptr != NULL)       /* If we found it */
    {
        old_ptr = (*ent_ptr).bckptr;
        /* Get previous item */
        if  (old_ptr == NULL)   /* If none */
        {
            card_hdr.fwdptr = (*ent_ptr).fwdptr;
            /* Reset header */
        }
        else            /* Else */
        {
            (*old_ptr).fwdptr = (*ent_ptr).fwdptr;
            /* Move entry back */
        }
        old_ptr = (*ent_ptr).fwdptr;
        /* Get next item */
        if  (old_ptr == NULL)   /* If none */
        {
            card_hdr.bckptr = (*ent_ptr).bckptr;
            /* Reset header */
        }
        else            /* Else */
        {
            (*old_ptr).bckptr = (*ent_ptr).bckptr;
            /* Move entry forward */
        }
        (void) memset (ent_ptr, '\0', sizeof(struct psl_card));
        /* Erase the memory */
        free (ent_ptr);     /* And release the memory */
    }
    return;             /* Return to caller */
}

/************************************************************************/
/*                                  */
/*  PSL_CARDFILE_LOCATE - Locate an entry in the Cardfile       */
/*                                  */
/*  This routine locates an entry in the Cardfile.          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    data_ptr = PSL_CARDFILE_LOCATE (text_ptr);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*  data_ptr = Pointer to data associated with located item     */
/*  text_ptr = Pointer to text associated with item to locate.  */
/*                                  */
/************************************************************************/

void *PSL_CARDFILE_LOCATE (char *text_ptr)
{

    /*   Local Data   */

    struct   psl_card   *ent_ptr;       /* Pointer to located card */

    /*   Locate the entry via the standard routine */

    ent_ptr = LOCATE_ENTRY (text_ptr);
    /* Locate the entry */

    /*   If we found it return the data pointer */

    if  (ent_ptr != NULL)       /* If we found it */
    {
        return (*ent_ptr).data; /* Return the Data Pointer */
    }
    else                /* Else (not found ) */
    {
        return NULL;        /* Return failure to caller */
    }
}

/************************************************************************/
/*                                  */
/*  LOCATE_ENTRY - Locate an entry in the Cardfile          */
/*                                  */
/*  This routine locates an entry in the Cardfile.          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*    ent_ptr = LOCATE_ENTRY (text_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*  ent_ptr = Pointer to located item               */
/*  text_ptr = Pointer to text associated with item to locate.  */
/*                                  */
/************************************************************************/

static struct psl_card *LOCATE_ENTRY (char *text_ptr)
{

    /*   Local Data   */

    struct   psl_card   *nxt_ptr;       /* Pointer to next card */
    size_t     txtlen;     /* Text length */

    /*   Check size of user's string and initialize card pointer */

    txtlen = strlen (text_ptr); /* Get length of user text string */
    if  (txtlen > 8) txtlen = 8;    /* Constrain to eight characters */

    /*   Sacn the list for the desired entry */

    for (nxt_ptr = card_hdr.fwdptr; /* Start at front of chain */
            nxt_ptr != NULL;       /* until end of chain */
            nxt_ptr = (*nxt_ptr).fwdptr)
        /* Following forward pointers */
    {
        if  (memcmp (text_ptr, (*nxt_ptr).descr, txtlen) == 0)
            /* If this entry matches */
        {
            return nxt_ptr;     /* Return the Structure Pointer */
        }
    }
    return NULL;            /* Else return failure to caller */
}

