/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_TAG_QSORT               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This routine forms a shell around the C RTL routine QSORT        */
/*  to implement a tag sort rather than a record sort.           */
/*                                       */
/*===========================================================================*/


/*    Define useful Include structures   */

#include  <psl_macros.h>        /* PSL Standard Macros */

/*    Define Function Prototypes for all external routines */
/*    used by the module */

#include    <stdlib.h>          /* Define RTL function prototypes */
#include  psl_common_file(psl_rtndef.h) /* PSL Function Prototypes */

/*    Define Function Prototypes for all internal routines */
/*    used by the module */

static int PSL_TAG_CMPRTN (const void *tag1_ptr, const void *tag2_ptr);

/*    Define the Global Data passed to the Compare Routine */

static  void * main_array_ptr;      /* Pointer to main array */
static  size_t main_array_size;     /* Size of each item in main array */
static  int    (*compare_rtn)(void *, void *, void *);
/* User comparison routine */
static  void * saved_user_ptr;      /* Saved user pointer */

/************************************************************************/
/*                                  */
/*  PSL_TAG_QSORT - Tag Sort shell around QSORT         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_TAG_QSORT (main_array, num_elem, elem_size,      */
/*            tag_array, compare_routine, user_ptr)     */
/*                                  */
/*  Where:                              */
/*                                  */
/*     main_array = Pointer to array that is to be sorted       */
/*     num_elem   = Number of elements in array         */
/*     elem_size  = Size of each element in main array      */
/*     tag_array  = Pointer to tag array                */
/*     compare_routine = pointer to comparison routine      */
/*     user_ptr   = Pointer to user data to pass to comp rtn    */
/*                                  */
/************************************************************************/

void PSL_TAG_QSORT (void *arr_base, size_t num_elem, size_t arr_size,
                    unsigned long *tag_base,
                    int (*compar)(void *, void *, void *),
                    void *user_ptr)
{

    /*   Save the data needed by the Comparison Routine */

    main_array_ptr = arr_base;      /* Save pointer to main array */
    main_array_size = arr_size;     /* and size of each item in main array */
    compare_rtn = compar;       /* And address of user routine */
    saved_user_ptr = user_ptr;      /* And user pointer */

    /*   Call QSORT to do all the hard work */

    qsort ((void *) tag_base, num_elem,
           sizeof (unsigned long), PSL_TAG_CMPRTN);
    /* Call QSORT to sort array */
}

/************************************************************************/
/*                                  */
/*  PSL_TAG_CMPRTN - Comparison Routine called by QSORT     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     ival = PSL_TAG_CMPRTN (tag1_ptr, tag2_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     ival     = Comparison Status:                */
/*          < 0 if items in order               */
/*            0 if items equal                  */
/*          > 0 if items in reverse order           */
/*     tag1_ptr = Pointer to tag for first element          */
/*     tag2_ptr = Pointer to tag for second element         */
/*                                  */
/*  This routine sits between QSORT and the user routine and    */
/*  simply calls the user routine having converted the tag pointers */
/*  into pointers to the associated items.              */
/*                                  */
/************************************************************************/

static int PSL_TAG_CMPRTN (const void *tag1_ptr, const void *tag2_ptr)
{

    /*   Local Data */

    void *arr_ptr1;        /* Pointer to first menu item */
    void *arr_ptr2;        /* Pointer to second menu item */
    unsigned long offset;           /* Intermediate array offset */
    unsigned long *tag_ptr;         /* Pointer to tag array */
    char *char_ptr;        /* Pointer for array calcs */

    /*   The pointers that are passed to us are pointers to our tag     */
    /*   array.  This, in turn, contains subscripts (starting at 1)     */
    /*   into the user's array, in which elements may be of any size.   */
    /*   We need to convert each tag pointer into a pointer to the      */
    /*   associated array item before asking the user to compare the    */
    /*   items.                             */

    char_ptr = (char *) main_array_ptr;
    /* cast array pointer */
    tag_ptr = (unsigned long *) tag1_ptr;
    /* and first tag pointer */
    offset = (((*tag_ptr) - 1) * main_array_size);
    /* Get offset of first item */
    arr_ptr1 = (void *) (char_ptr + offset);
    /* and hence pointer to it */

    tag_ptr = (unsigned long *) tag2_ptr;
    /* cast second tag pointer */
    offset = (((*tag_ptr) - 1) * main_array_size);
    /* Get offset of second item */
    arr_ptr2 = (void *) (char_ptr + offset);
    /* and hence pointer to it */
    return (*compare_rtn)(arr_ptr1, arr_ptr2, saved_user_ptr);
    /* Return result of user routine */
}
