/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/////////////////////////////////////////////////////////////////////////
//
// CECTssWriter.cpp
// ----------------
//
//    This is the implementation file for the CECTssWriter class
//
//////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CECTssWriter.h"
#include "CECRegistryTools.h"
#include "Solutions.h"
#include "CECTargetEnvironment.h"

//
//  PSL Macros.
//
extern "C"
{
#include "psl_macros.h"
#include "target_rtndef.h"  // Target/Executer routines
}


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CECTssWriter

CECTssWriter::CECTssWriter  ()
{}


CECTssWriter::~CECTssWriter()
{}

//
// GetNameDefinitionText
// =====================
//
// Get any Name Definition Text
//

CString CECTssWriter::GetNameDefinitionText(TSS_SECTION nType, CString szTestName/*=""*/, CString szDescription/*=""*/)
{
    CString szBuffer("");

    if (nType == ENVIRONMENT)
    {
        szBuffer += GetDescriptionStatement(nType, szDescription);
    }
    else if (nType == DEFINITIONS)
    {
        szBuffer += GetThreadStatement(nType, szTestName);
    }
    else if (nType == CODE)
    {
        szBuffer += GetTestStatement(nType, szTestName);
    }

    return szBuffer;
}

//
// GetDescriptionStatement
// =======================
//
// Get the Description
//
CString CECTssWriter::GetDescriptionStatement(TSS_SECTION nType, CString szDescription/*=""*/)
{
    CString szBuffer("");

    if (nType == ENVIRONMENT)
    {
        szBuffer += "\tDescription \t\"" + szDescription + "\"\n";
    }
    return szBuffer;
}

//
// GetTestStatement
// ================
//
// Get the Test Statements
//
CString CECTssWriter::GetTestStatement(TSS_SECTION nType, CString szTestName/*=""*/, CString szNode/*=""*/)
{
    CString szBuffer("");

    if (nType == DEFINITIONS)
    {
        szBuffer += "\tTest \t" + szTestName + "\n";
    }
    else if (nType == CODE)
    {
        szBuffer += "\tExecute Test \t" + szTestName;

        // Ignore Local Nodes
        if (szNode.CompareNoCase("<Local>") != 0 && !szNode.IsEmpty())
        {
            szBuffer += ",\tOn Node \"" + szNode + "\"";
        }
        szBuffer += "\n";
    }
    return szBuffer;
}

//
// GetThreadStatement
// ==================
//
// Get the Thread Statements
//
CString CECTssWriter::GetThreadStatement(TSS_SECTION nType, CString szTestName/*=""*/,
        int nUserThreads/*=1*/, int nInterval /*=0*/,
        CString szTargetName/*=""*/)
{
    CString szBuffer("");

    //
    //    If we have not been passed in the target environment then default to what ever is current
    //
    if(szTargetName.IsEmpty())
    {
        szTargetName = GetCurrentTargetEnvironment();
    }

    if (nType == DEFINITIONS)
    {
        szBuffer += "\tThread \t" + szTestName + " For Target " + szTargetName +"\n";
    }

    else if (nType == CODE)
    {
        //
        //    If the current target is COMMAND then we shall execute a thread USING SCRIPT, otherwise
        //    we shall execute USING TOF
        //
        char TargetString[32];
        Tgt_GetTarget (TARGET_COMMAND ,TargetString ,sizeof(TargetString) ,FALSE);
        if (szTargetName == TargetString)
        {
            szBuffer += "\tExecute Thread \t" + szTestName + "\tUsing Script \"" + szTestName + "\"";
        }
        else
        {
            szBuffer += "\tExecute Thread \t" + szTestName + "\tUsing Tof \"" + szTestName + "\"";
        }

        if ((nInterval>1) || (nUserThreads>1))
        {

            CString szUserThreads;
            CString szInterval;

            szUserThreads.Format("%d", nUserThreads);
            szInterval.Format("%d", nInterval);

            szBuffer +=  ", Number " + szUserThreads + ", Interval " + szInterval;
        }

        szBuffer += "\n";
    }
    return szBuffer;
}

//
// GetSynchronizationStatement
// ===========================
//
// Get the Thread Statements
//
CString CECTssWriter::GetSynchronizationStatement(TSS_SECTION nType, CStringArray* arSynchArray /*=NULL*/)
{
    CString szBuffer("");

    if (nType == CODE)
    {
        for (int i=0; i<arSynchArray->GetSize(); i++)
        {
            CString szTestName = arSynchArray->GetAt(i);

            if (i==0)
            {
                // Wait for the test
                szBuffer += "\tWait For Test \t" + szTestName;
            }
            else
            {
                // Add to list of tests being waited on
                szBuffer += " and " + szTestName;
            }
        }
        if (!szBuffer.IsEmpty()) szBuffer += "\n";
    }
    return szBuffer;
}
//
// GetDelayStatement
// =================
//
// Get the Delay Statement
//
CString CECTssWriter::GetDelayStatement(TSS_SECTION nType, int nDelay /*=10*/)
{
    CString szBuffer("");

    if (nType == CODE)
    {
        if (nDelay>0)
        {
            // Wait for the test
            szBuffer.Format("\tWait %d", nDelay);
        }
        if (!szBuffer.IsEmpty()) szBuffer += "\n";
    }
    return szBuffer;
}

//
// GetTemplateText
// ===============
//
// If a template file does not exist this function will create some default
// text for the desired section.

CString CECTssWriter::GetTemplateText(TSS_SECTION nType, CString szFileName /*=""*/)
{
    // Check to see if we have a template ...
    if (DoesTemplateExist((SECTION) nType))
    {
        // If so - use it
        return UseTemplate((SECTION) nType);
    }

    // Else create some default test
    return CECDefaultCodeWriter::GetDefaultText((SECTION) nType, szFileName);
}

//
// GetSectionTitle
// ===============
//
// Simply use the default

CString CECTssWriter::GetSectionTitle(TSS_SECTION nType)
{
    return CECDefaultCodeWriter::GetSectionTitle((SECTION) nType);
}


//
// GetCurrentTarget
// ================
//
// Get the current Target Environment
//
CString CECTssWriter::GetCurrentTargetEnvironment()
{
    CECTargetEnvironment te;
    CString szContext;

    te.GetCurrentContext(szContext);

    // and confirm that it is one that we understand.
    if ((Tgt_GetTargetID (szContext.GetBuffer(0)) == -1))
    {
        // we don't so blitz it
        szContext = "COMMAND";
    }

    return szContext;
}

//
//  GetOTAThreadStatement
//  =====================
//
//  Get the ota's thread statement
//
CString CECTssWriter::GetOTAThreadStatement(TSS_SECTION nType,CString szThreadName,CString szTofName,CString szTargetNode)
{
    CString szBuffer("");

    if (nType == DEFINITIONS)
    {
        szBuffer += "\tThread \t" + szThreadName + " For Target MON\n";
    }
    else
        if (nType == CODE)
        {
            szBuffer += "\tExecute Thread \t" + szThreadName + "\tUsing Tof \"" + szTofName + "\"" + "\t On Node \"" + szTargetNode + "\"\n";
        }
    return szBuffer;
}
