/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStApi CyrWrapLib
// @module  Test Output Wrappers | This module contains the implementation of
//          any simple Test Output types.  These output values are in effect
//          strings when they enter the OpenSTARep or when they are
//          transmitted via CORBA.  The objects are created to hide this
//          functioanlity from the developer.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECTestOutput.h>
///////////////////////////////////////////////////////////////////////////////
// RogueWave includes
#include <rw/ctoken.h>

///////////////////////////////////////////////////////////////////////////////
// Constants
#define COMMA ","

const char* CECTimerOutput::OUTPUT_TYPE     = "Timer.txt";
const char* CECTestLog::OUTPUT_TYPE         = "TestLog.txt";
const char* CECTestReport::OUTPUT_TYPE      = "TestRep.txt";
const char* CECTestDiagnostic::OUTPUT_TYPE  = "TestDiag.txt";
const char* CECTestHistory::OUTPUT_TYPE     = "TestHis.txt";
const char* CECOtaSimpleQuery::OUTPUT_TYPE  = "OtaStat.txt";
const char* CECOtaHeader::OUTPUT_TYPE       = "OtaHeader.txt";

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This base constructor simply generates a time value for the
//          creation of this object.  It should only be used when creating a
//          completely new test output object.
CECTestOutput::CECTestOutput()
{
    // Generate current time
    m_time = RWTime::now();
}

CECTestOutput::~CECTestOutput()
{}


RWTime CECTestOutput::GetCreatedTime() const
{
    return m_time;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a brand new instance of a Timer output.
//          The time of creation is logged internally, but all other values are
//          passed in.
// @parm    The description for this timer output.
// @parm    The value for this timer output.
// @parm    The number of active threads.
CECTimerOutput::CECTimerOutput( const char* desc, long value, long active )
        : CECTestOutput()
{
    // Assign member variables
    m_desc = desc;
    m_active = active;
    m_value = value;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor recreates a timer output based on the passed
//          string.  This string will be parsed in an attempt to find the
//          relevant strings.
// @parm    The string to be used uring recreation
CECTimerOutput::CECTimerOutput( const char* str )
{
    // Tokenize string
    RWCTokenizer token( str );

    // Extract relevant entries
    m_time = RWTime( atol( token( COMMA ).data()));
    m_value = atol( token( COMMA ).data() );
    m_active = atol( token( COMMA ).data() );
    m_desc = token( COMMA );
}

CECTimerOutput::~CECTimerOutput()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the timer output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.

RWCString CECTimerOutput::ToString()
{
    // Serialize all values into a single string
    RWCString ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d", m_time.seconds() );
    ret += buf;
    ret += COMMA;

    // Timer value
    sprintf( buf, "%d", m_value );
    ret += buf;
    ret += COMMA;

    // Number of active threads
    sprintf( buf, "%d", m_active );
    ret += buf;
    ret += COMMA;

    // Description
    ret += m_desc;
    ret += COMMA;

    return ret;
}

RWCString CECTimerOutput::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

RWCString CECTimerOutput::GetDescription() const
{
    return m_desc;
}

long CECTimerOutput::GetValue() const
{
    return m_value;
}

long CECTimerOutput::GetActiveThreads() const
{
    return m_active;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This is the one and only constructor for this type.  As there are
//          no distinguishing variables between a creation and retrieval for
//          this type, a boolean is added to the parameters.  By default this
//          boolean is true and the constructor will create a completely new
//          test log entry.  If set to fals then the passed string is parsed
//          and the relevant fields are extracted.
// @parm    The text to add are parse.
// @parm    Determines whether to parse text or not.
CECTestLog::CECTestLog( const char* text, bool create /*=true*/ )
        : CECTestOutput()
{
    // If this is a brand new test log entry then simply assign the
    // text value to the member variable.  otherwise we need to parse the
    // information out of the text variable.
    if ( create == true )
        m_text = text;
    else
    {
        // Tokenize string
        RWCTokenizer token( text );

        // Extract relevant entries
        m_time = RWTime( atol( token( COMMA ).data()));
        m_text = token( COMMA );
    }
}

CECTestLog::~CECTestLog()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the test log fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.

RWCString CECTestLog::ToString()
{
    // Serialize all values into a single string
    RWCString ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d", m_time.seconds() );
    ret += buf;
    ret += COMMA;

    // Text string
    ret += m_text;
    ret += COMMA;

    return ret;
}

RWCString CECTestLog::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

RWCString CECTestLog::GetText() const
{
    return m_text;
}

CECTestReport::CECTestReport( const char* text, bool create )
        : CECTestLog( text, create )
{}


RWCString CECTestReport::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

RWCString CECTestDiagnostic::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

CECTestHistory::CECTestHistory( const char* text, bool create )
        : CECTestLog( text, create )
{}


RWCString CECTestHistory::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor takes a line read in from the test output file
//          or transferred via CORBA and converts it into an instance of the
//          class.
// @parm    A string containing all information for construction of object.
CECOtaSimpleQuery::CECOtaSimpleQuery( const char* str )
        : CECTestOutput()
{
    // Extract the node name and other strings
    RWCTokenizer token( str );
    m_node = token( COMMA );
    RWCString index = token( COMMA );
    RWCString start = token( COMMA );
    RWCString stop  = token( COMMA );
    RWCString result = token( COMMA );

    m_index = atoi(index);
    m_start = atoi(start);
    m_stop = atoi(stop);
    m_result = atoi(result);
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new data object.
// @parm    Start time.
// @parm    Stop time.
// @parm    Result.
CECOtaSimpleQuery::CECOtaSimpleQuery( const char* node, int index, int start, int stop, int result )
        : CECTestOutput()
{
    // Assign member variables
    m_node = node;
    m_index = index;
    m_start = start;
    m_stop = stop;
    m_result = result;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor tidies up the allocated memory for the result array.
CECOtaSimpleQuery::~CECOtaSimpleQuery()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node where the query was
//          executed.
// @rdesc   The node name.

RWCString CECOtaSimpleQuery::GetNodeName()
{
    return m_node;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the query output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.
RWCString CECOtaSimpleQuery::ToString()
{
    // Serialize all data into a string
    RWCString ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%s%s%d%s%d%s%d%s%d%s",
             m_node, COMMA,
             m_index, COMMA,
             m_start, COMMA,
             m_stop, COMMA,
             m_result, COMMA );
    ret += buf;

    return ret;
}

RWCString CECOtaSimpleQuery::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description of the query.
// @rdesc   The query description.
RWCString CECOtaSimpleQuery::GetDescription()
{
    return m_desc;
}

int CECOtaSimpleQuery::GetStartTime()
{
    return m_start;
}

int CECOtaSimpleQuery::GetStopTime()
{
    return m_stop;
}

int CECOtaSimpleQuery::GetResult()
{
    return m_result;
}

int CECOtaSimpleQuery::GetIndex()
{
    return m_index;
}






















///////////////////////////////////////////////////////////////////////////////
CECOtaHeader::CECOtaHeader( const char* str )
        : CECTestOutput()
{
    // Extract the node name and other strings
    RWCTokenizer token( str );
    m_node = token( COMMA );
    RWCString strIndex = token( COMMA );
    m_index = atoi(strIndex);
    m_desc = token( COMMA );

}

CECOtaHeader::CECOtaHeader( const char* node, int index, const char* desc )
        : CECTestOutput()
{
    // Assign member variables
    m_index = index;
    m_node = node;
    m_desc = desc;
}

CECOtaHeader::~CECOtaHeader()
{}


RWCString CECOtaHeader::GetType()
{
    return RWCString( OUTPUT_TYPE );
}

RWCString CECOtaHeader::ToString()
{
    RWCString ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%s%s%d%s%s%s",
             m_node, COMMA,
             m_index, COMMA,
             m_desc, COMMA );
    ret += buf;

    return ret;
}

RWCString CECOtaHeader::GetNodeName()
{
    return m_node;
}

RWCString CECOtaHeader::GetDescription()
{
    return m_desc;
}

int CECOtaHeader::GetIndex()
{
    return m_index;
}
