/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "CConfigTable.h"
#ifdef __VMS
#include <ctype.h>
#else
#include <cctype>
#endif

/**
 * Constructor, initialise both table types to NULL.
 **/
CConfigTable::CConfigTable()
{
    _singleColTable = NULL;
    _multiColTable = NULL;
}

/**
 * Copy constructor
 **/
CConfigTable::CConfigTable(const CConfigTable& configIn)
{
    if (configIn._singleColTable != NULL)  // Single column map copy constructor OK
        _singleColTable = new _CONFIG_COLUMN(*configIn._singleColTable);
    else
        _singleColTable = NULL;

    if (configIn._multiColTable != NULL)
    { // Table requires copy for each column
        _multiColTable = new _CONFIG_TABLE;
        for(_CONFIG_TABLE::const_iterator i = (*configIn._multiColTable).begin();
                i!=(*configIn._multiColTable).end(); i++)
            _multiColTable->insert(_CONFIG_TABLE::value_type((*i).first,
                                   new _CONFIG_COLUMN( *((*i).second) )));
    }
    else
        _multiColTable = NULL;
}

/**
 * Destructor, ensures all maps are deleted.
 **/
CConfigTable::~CConfigTable()
{

    if ( _multiColTable != NULL)
    {
        for(_CONFIG_TABLE::const_iterator i = (*_multiColTable).begin();
                i!=(*_multiColTable).end(); i++)
            // Remove each column.
            delete (*i).second;
        // Remove table
        delete _multiColTable;
    }
    else if (_singleColTable != NULL)
        // Remove single column table.
        delete _singleColTable;
}

/**
 * Adds a value to a table. If table does not already exist it is created.
 * Parameter format which is dictated Atvf input use, is as follows :
 * 'rowString' will contain a numeric row identifier.
 * 'colValString' will contain an optional column token followed by a string
 * separated by whitespace.
 * If successful AddValue returns true, otherwise the result is false. 
 **/
addValueResult CConfigTable::AddValue(const string& rowString,const string& colValString)
{
    // Check for empty parameters
    if (rowString.empty() )
        return BAD_ROW_ID;
    if (colValString.empty())
        return BAD_COL_ID;

    // Convert row string to integer
    int intVal = 0;
    for (string::size_type index = 0; index < rowString.size();index++)
        if (isdigit(rowString[index])) // if a digit convert
            intVal = (intVal*((index==0)?1:10))+(rowString[index] - '0');
        else if (rowString[index]==' ' || rowString[index]=='\t')
            // Whitespace terinates row number
            break;
        else // Any other character complain
            return BAD_ROW_ID;

    // Convert column/value string to [column string] value string
    string colString;
    string valString;
    // Find last non-whitespace
    string::size_type endOfColVal;
    for ( endOfColVal = colValString.size();
            (colValString[endOfColVal-1]==' ')||(colValString[endOfColVal-1]=='\t');
            endOfColVal--);
    // Look for delimited value
    if (colValString[0] == '"' &&colValString[endOfColVal-1] == '"' && (endOfColVal-1)!=0)
        valString = colValString.substr(1,int(endOfColVal-2));
    else
    {
        string::size_type firParamEnd = colValString.find_first_of(" \t",0);
        if (firParamEnd != string::npos)
        { //  Check for column, and value
            string::size_type secParamStart = colValString.find_first_not_of(" \t",firParamEnd);
            if (secParamStart != string::npos)
            { // if column, and value
                colString=colValString.substr(0,firParamEnd);
                valString=colValString.substr(secParamStart,int(endOfColVal-secParamStart));
            }
            else
            { // only value
                valString=colValString.substr(0,firParamEnd);
            }
        }
        else // only value
            valString=colValString;
    }

    // Add to table
    if (colString.empty())
    { // Single colun table
        if (_multiColTable != NULL)
            return REDIM_ERROR;

        if (_singleColTable == NULL)
            _singleColTable = new _CONFIG_COLUMN;

        _singleColTable->insert(_CONFIG_COLUMN::value_type(intVal,valString));
        return OK;

    }
    // Multi column table

    // Check for existing single column table
    if (_singleColTable != NULL)
        return REDIM_ERROR;

    // Concert column header to uppercase
    string columnUC = ChangeCase(colString);

    // Check existing table
    if (_multiColTable == NULL)
        _multiColTable = new _CONFIG_TABLE;

    // Find or create column
    _CONFIG_COLUMN* currentColumn;
    if ((*_multiColTable).find(columnUC)== (*_multiColTable).end())
    {
        currentColumn = new _CONFIG_COLUMN;
        (*_multiColTable).insert(_CONFIG_TABLE::value_type(columnUC,currentColumn));
    }
    else
        currentColumn = (*_multiColTable)[columnUC];

    (*currentColumn).insert(_CONFIG_COLUMN::value_type(intVal,valString));

    return OK;


}

/**
 * Single column table GetValue. Returns value at 'row', otherwise returns "".
 **/
const string CConfigTable::GetValue(int row) const
{
    if ((_singleColTable == NULL) ||
            ((*_singleColTable).find(row)==(*_singleColTable).end()) )
        return "";
    return (*_singleColTable)[row];
}

/**
 * Multiple column table GetValue. Returns value at ['column','row'], otherwise
 * returns "". 
 **/
const string CConfigTable::GetValue(string column, int row) const
{
    string columnUC = ChangeCase(column);

    if ( (_multiColTable == NULL) ||
            (*_multiColTable).find(columnUC)==(*_multiColTable).end() ||
            (*_multiColTable)[columnUC]->find(row) ==(*_multiColTable)[columnUC]->end())
        return "";

    return (*(*_multiColTable)[columnUC])[row];
}

/**
 * Single column table GetColumn. Returns the address of the single column table,
 * (if undefined this will be NULL). NOTE column belongs to class do not delete!!!
 **/
const _CONFIG_COLUMN* CConfigTable::GetColumn() const
{
    return _singleColTable;
}

/**
 * Multiple column table GetColumn. Returns the address of the single column table
 * at 'colName' in the Multiple column table. (If undefined this will be NULL) 
 * NOTE column belongs to class do not delete !!!!!
 **/
const _CONFIG_COLUMN* CConfigTable::GetColumn(string colName) const
{
    string colNameUC = ChangeCase(colName);
    if ( (_multiColTable != NULL) &&
            (*_multiColTable).find(colNameUC)!= (*_multiColTable).end() )
        return (*_multiColTable)[colNameUC];

    return NULL;
}

/**
 * Helper funtion converts function column tokens to upper case.
 **/
const string CConfigTable::ChangeCase(const string& stingIn) const
{
    string stringUC;

    for (string::size_type index = 0; index < stingIn.size(); index++)
        stringUC+=toupper(stingIn[index]);

    return stringUC;
}

/**
 * Friend operator << for output.
 **/
ostream& operator<<(ostream& os, const CConfigTable& table)
{

    if (table._singleColTable!=NULL)
    {
        for (_CONFIG_COLUMN::const_iterator row = (*table._singleColTable).begin();
                row != (*table._singleColTable).end(); row ++)
        {
            if (row != (*table._singleColTable).begin())
                os << endl;
            os << "(" << (*row).first << ") " << (*row).second;
        }
    }
    else if (table._multiColTable!=NULL)
    {
        for(CConfigTable::_CONFIG_TABLE::const_iterator colIndex= table._multiColTable->begin();
                colIndex != table._multiColTable->end(); colIndex++)
        {
            if (colIndex != table._multiColTable->begin())
                os << endl;
            for (_CONFIG_COLUMN::const_iterator rowIndex = (*(*colIndex).second).begin();
                    rowIndex != (*(*colIndex).second).end(); rowIndex++)
            {
                if (rowIndex != (*(*colIndex).second).begin())
                    os << endl;
                os << "(" << (*colIndex).first << "," << (*rowIndex).first
                << ") " << (*rowIndex).second;
            }
        }
    }

    return os;
}
