/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TParamRule.cpp
/**
 * Method definitions for Transaction Parameter Rule class
 **/

#ifdef WIN32
#pragma warning(disable:4786)
#endif
#include "TParamRule.h"

/**
 * Default Constructor used when duplicating TParamRule types 
 **/

TParamRule::TParamRule()
        :   _aRRuleP(NULL),
        _defVal(""),
        _configs(NULL)
{}

/**
 * Constructor
 **/
TParamRule::TParamRule(AtvfRecRule* aRRuleP)
        :   _aRRuleP(aRRuleP),
        _defVal(""),
        _configs(NULL)
{}

/**
 * Destructor
 *
 * TParamRule is no longer responsible for the destruction of the associated
 * AtvfRecRule. Instead with the new single AtvfRecRule to multiple TParamRule's
 * model, which is now used, the TParamRule tells AtvfRecRule to decremented it's
 * reference count.
 **/
TParamRule::~TParamRule()
{
    _aRRuleP->removeReference();
#ifndef __VAX
    if(_configs)
        delete _configs;
#endif
}

/**
 * Method for accessing the inner atvf record rule
 **/
const AtvfRecRule* TParamRule::innerRule() const
{
    return _aRRuleP;
}

/**
 * Method for accessing the key for this rule
 **/
AtvfKey TParamRule::key() const
{
    return _aRRuleP->key();
}

/**
 * Method for accessing the description for this parameter
 **/
string TParamRule::token() const
{
    return _aRRuleP->token();
}

/**
 * Method for accessing the options of this parameter
 **/
const TOptions& TParamRule::options() const
{
    return _options;
}

/**
 * Method for accessing configuration table
 **/
const CConfigTable* TParamRule::config() const
{
    return _configs;
}

/**
 * Method for accessing  the default value of this parameter
 * as a string
 **/
string TParamRule::defStrVal() const
{
    return _defVal;
}

/**
 * Method for accessing the default value of this parameter
 * as an integer, 
 **/
int TParamRule::defIntVal() const
{
    return intVal(defStrVal());
}

/**
 * Method to access a conversion of string to a possible integer
 * representation using this rule. Fairly meaningless in this base
 * class so we'll always return zero.
 **/
int TParamRule::intVal(const string&) const
{
    return 0;
}

/**
 * Method to check whether the parameter that this rule describes
 * supports integer values. Overriden in classes that do.
 **/
bool TParamRule::intVal() const
{
    return false;
}

/**
 * Method to access a conversion of integer to possible string
 * repesentation.  In this base class we'll always be returning
 * the default string value the real integer rule may check attribute
 * and format in a specific base.
 **/
string TParamRule::intVal(int) const
{
    return defStrVal();
}

/**
 * Method to produce a segmented string from a value if this
 * parameter
 **/
TSegStr* TParamRule::segStr(const string& value) const
{
    return (hasAttr("varsubs")) ? TSegStr::createIfNeeded(value) :
#ifndef __VMS
           static_cast<TSegStr*>(NULL);
#else
           (TSegStr*)NULL;
#endif
}

/**
 * Method to access whether this rule has had a particular
 * attribute set.  Attributes although passed in as strings
 * are actually stored as AtvfKeys and therefore only the
 * first four chars matter in a no-case-dependent manner.
 **/
bool TParamRule::hasAttr(const string& attr) const
{
    attrset_t::const_iterator i = _attrs.find(AtvfKey(attr));
    if( i != _attrs.end()) return true;
    else return false;
}

/**
 * Protected accessor for copying our contents to another
 * rule of similar type
 **/
void TParamRule::giveContents(TParamRule* pRP) const
{
    pRP->addOption(_options);
    pRP->defStrVal(_defVal);
    pRP->setAttr(_attrs);
    if (_configs!=NULL)
        pRP->addConfigs(new CConfigTable(*_configs));
    // Share the same Atvf record rule
    pRP->setAtvfRule(_aRRuleP);
    return;
}

/**
 * Method for giving this parameter another option
 **/
void TParamRule::addOption(const string& text)
{
    _options.add(text);
    return;
}

/**
 * Method for giving this parameter group of options
 **/
void TParamRule::addOption(const TOptions& opts)
{
    _options.add(opts);
    return;
}

/**
 * Method for setting the default value of this parameter
 **/
void TParamRule::defStrVal(const string& defVal)
{
    _defVal = defVal;
    return;
}

/**
 * Method for setting an attribute of a particular rule
 **/
void TParamRule::setAttr(const string& attr)
{
    _attrs.insert(AtvfKey(attr));
    return;
}

/**
 * Method for setting copying set of attributes into ours
 **/
void TParamRule::setAttr(const attrset_t& attrs)
{
    for(attrset_t::const_iterator i = attrs.begin();
            i != attrs.end();
            i++)
    {
        _attrs.insert(*i);
    }
    return;
}

/**
 * Add configuration table
 **/
void TParamRule::addConfigs(const CConfigTable* configIn)
{
    _configs = configIn;
}


/**
 *  Associates parameter with the appropriate previously defined Atvf record rule.
 *  Used when when creating a duplicate parameter.
 **/
void TParamRule::setAtvfRule(AtvfRecRule* const sharedRule )
{
    _aRRuleP = sharedRule;
    //  Update rule's reference count
    _aRRuleP->addReference();
}


/**
 * Global function for printing out this class
 **/
ostream& operator<<(ostream& os, const TParamRule& pR)
{
    os << "Name = " << pR.token() << "\tType = " << pR.type() <<
    "(" << pR._aRRuleP->type() << ")";

    if(pR._defVal!="")
        os << endl << "Default = " << pR._defVal;

    if(!pR._attrs.empty())
        for(attrset_t::const_iterator attrset_it = pR._attrs.begin();
                attrset_it != pR._attrs.end();
                attrset_it++)
            (attrset_it == pR._attrs.begin()) ?
            os << endl << "Attributes = " << *attrset_it :
            os << ", " << *attrset_it;

    os << endl << pR._options;

    if (pR._configs != NULL)
        os << endl << "Configuration Table :" << endl << *(pR._configs);

    return os;
}
