/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TPrmRealRules.cpp
/**
 * Method definitions for Parameter rule classes found in
 * TPrmRealRules.h
 **/
#include "TPrmRealRules.h"
#include "AtvfRealRecRules.h"

/**
 * Token Parameter methods
 **/
const string TPrmToknRule::TYPE("tokn");
TPrmToknRule::TPrmToknRule(const string& token)
        :   TParamRule(new AtvfNoValRule(token))
{}

TParamRule* TPrmToknRule::duplicate(const string& name,
                                    const string& type) const
{
    if(type.compare(TYPE) != 0) return NULL;
    TParamRule* pRP = new TPrmToknRule(name);
    giveContents(pRP);
    return pRP;
}
// Description method
string TPrmToknRule::type() const
{
    return "Token";
}

/**
 * Date Parameter methods
 **/
const string TPrmDateRule::TYPE("date");
TPrmDateRule::TPrmDateRule(const string& token)
        :   TParamRule(new AtvfSimplRule(token))
{}

TParamRule* TPrmDateRule::duplicate(const string& name,
                                    const string& type) const
{
    if(type.compare(TYPE) != 0) return NULL;
    TParamRule* pRP = new TPrmDateRule(name);
    giveContents(pRP);
    return pRP;
}
// Description method
string TPrmDateRule::type() const
{
    return "Date Text";
}

/**
 * Short text Parameter methods
 **/
const string TPrmSTxtRule::TYPE("stxt");
TPrmSTxtRule::TPrmSTxtRule(const string& token)
        :   TParamRule(new AtvfSimplRule(token))
{}

TParamRule* TPrmSTxtRule::duplicate(const string& name,
                                    const string& type) const
{
    if(type.compare(TYPE) != 0) return NULL;
    TParamRule* pRP = new TPrmSTxtRule(name);
    giveContents(pRP);
    return pRP;
}
string TPrmSTxtRule::type() const
{
    return "Short Text";
}

/**
 * Integer Number parameter methods
 **/
const string TPrmINumRule::TYPE("inum");
const string TPrmINumRule::VALREF("0123456789ABCDEF");
TPrmINumRule::TPrmINumRule(const string& token)
        :   TParamRule(new AtvfSimplRule(token)),
        _base(10)
{}

TParamRule* TPrmINumRule::duplicate(const string& name,
                                    const string& type) const
{
    if(type.compare(TYPE) != 0) return NULL;
    TParamRule* pRP = new TPrmINumRule(name);
    giveContents(pRP);
    return pRP;
}
bool TPrmINumRule::intVal() const
{
    return true;
}
// Basic string to int conversion
int TPrmINumRule::intVal(const string& str) const
{
    // No coding for parsing different base yet
    string ref = VALREF.substr(0, _base);

    string::size_type val;
    int result = 0;
    int sign = 1;
    for(string::const_iterator it = str.begin();
            it != str.end();
            it++)
    {
        if((result == 0) && (*it =='-'))
        {
            sign = -1;
        }
        else
        {
            val = ref.find(*it);
            if(val == string::npos)
            {
                break;
            }
            result = (_base * result) + val;
        }
    }
    return result * sign;
}
// And the other way around, taking notice of attributes to set base later.
string TPrmINumRule::intVal(int val) const
{
    string strVal;
    if(val < 0)
    {
        strVal.append("-");
        val *= -1;
    }
    // Need base signifier appended here for none decimal
    int modulus = val%_base;
    if(val > modulus) strVal.append(intVal(val/_base)); // recursive bit
    strVal.append(1, VALREF.at(modulus));

    return strVal;
}
// Description method
string TPrmINumRule::type() const
{
    return string("Integer Number");
}

/**
 * Long Text Parameter methods
 **/
const string TPrmLTxtRule::TYPE("ltxt");
TPrmLTxtRule::TPrmLTxtRule(const string& token)
        :   TParamRule(new AtvfDelimRule(token))
{}

TParamRule* TPrmLTxtRule::duplicate(const string& name,
                                    const string& type) const
{
    if(type.compare(TYPE) != 0) return NULL;
    TParamRule* pRP = new TPrmLTxtRule(name);
    giveContents(pRP);
    return pRP;
}
string TPrmLTxtRule::type() const
{
    return "Long Text";
}
