/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TSegStr.cpp
// This module contains the implementation of the
// TSegStr and associated classes. The TSegStr class handles strings
// that may contain embedded variable references. The latter are
// identified by a single character, used to delimit the variable
// reference, e.g:
//
// <PRE>
// Input string:   Customer code = %cuscod[1]%; name = %cusnam%
// </PRE>
//
// The above string contains two embedded variable references to
// cuscod[1] and cusnam.
//
// A literal '%' is identified by two adjacent percent characters
// ("%%").
//
#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif
#include "TContext.h"
#include "TSegStr.h"
#include "mdbexcept.h"
#include "CECTofTxt.h"

const char TSegStr::VAR_DELIM = '%';
const char TSegStr::CURVAR_MODIFIER = '$';
const char TSegStr::LFT_INDEX_DELIM = '[';
const char TSegStr::RGT_INDEX_DELIM = ']';
const char TSegStr::RANDOM_FLAG = '?';

/**
 * Constructor for variable details
 **/
TVarDetails::TVarDetails(const string& orig)
        :   _orig(orig),
        _current(false),
        _index("")
{
    if(!_orig.empty())
    {
        _current = (_orig.at(0) == TSegStr::CURVAR_MODIFIER);
        _varname.assign(_orig.substr((_current?1:0)));

        if(_varname.at(_varname.size()-1) == TSegStr::RGT_INDEX_DELIM)
        {
            string::size_type pos =_varname.find_first_of(TSegStr::LFT_INDEX_DELIM);
            if(pos != string::npos)
            {
                ++pos;
                string idxstr(_varname.substr(pos));
                _varname.resize(pos-1);
                {
                    _index=idxstr.substr(1,idxstr.size()-2);
                }
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Static method returning a pointer to a new segmented string object
//	    (TSegStr) if the string passed in contains one or more variable
//	    references or percent literal specifiers. A variable reference is
//	    delimited by the "%" character, e.g. Total customers = %cuscnt%
//	    and a percent literal is specified by two adjacent percent
//	    characters ("%%").
//
// @rdesc   Pointer to new segmented string object (TSegStr) produced
//	    from supplied string value, otherwise NULL pointer.
// @parm    String from which to create segmented string object, if possible.
//
TSegStr* TSegStr::createIfNeeded(const string& val)
{
    // Look for two variable/percent literal delimiters in the supplied string.
    // If we find them create a segmented string object and return a pointer to
    // it, otherwise return a NULL pointer.
    string::size_type pos1 = val.find_first_of(VAR_DELIM);

    if (pos1 != string::npos)
    {
        string::size_type pos2 = val.find_first_of(VAR_DELIM, pos1+1);
        if (pos2 != string::npos) return new TSegStr(val, pos1);
    }

    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor. A segmented string object is created from the
//	    supplied string, starting to look for a variable reference or a
//	    percent literal specifier at the position spcified by the second
//	    parameter. The object contains a vector of 'text strings' (_txts)
//	    and a vector of 'TVarDetails' (_vars) describing the text and
//	    variable segments of the input string respectively.
//
//	    An unmatched variable/percent literal delimiter is treated as a
//	    literal character.
//
// @parm    String from which to create segmented string object.
// @parm    Position within input string at which to start search for variable
//	    delimiter or percent literal specifier. Default 0. Provided for
//	    efficiency.
//
TSegStr::TSegStr(const string& val, string::size_type searchPos)
        :   initialSize(val.size())
{
    for(string::size_type pos = 0;pos < initialSize;pos = searchPos)
    {
        TVarDetails* varP = NULL;
        string* txtP = NULL;
        string::size_type txtlen = string::npos; // text rest of string

        // Look for opening variable/percent literal delimiter.
        string::size_type delimPos = val.find_first_of(VAR_DELIM, searchPos);
        if (delimPos != string::npos) // If there is an opening delimiter
        {
            txtlen = delimPos - pos;
            string::size_type varStart = delimPos + 1;

            // Look for closing variable/percent literal delimiter.
            delimPos = val.find_first_of(VAR_DELIM, varStart);
            if(delimPos != string::npos) // If there is a closing delimiter
            {
                varP = new TVarDetails(val.substr(varStart, (delimPos-varStart)));
            }
            else // Treat opening delimiter as part of text.
            {
                txtlen = string::npos; // Text to rest of string - no more delimiters
            }

        } // End If there is an opening delimiter

        if(txtlen > 0)txtP = new string(val.substr(pos, txtlen));
        _txts.push_back(txtP);
        _vars.push_back(varP);
        searchPos = (delimPos < initialSize) ? delimPos+1 : initialSize;
    }//end for
}

/**
 * Destructor
 **/
TSegStr::~TSegStr()
{
    for(vector<string*>::iterator t = _txts.begin();
            t != _txts.end();
            t++) delete *t;

    _txts.erase(_txts.begin(), _txts.end());
    for(vector<TVarDetails*>::iterator v = _vars.begin();
            v != _vars.end();
            v++) delete *v;

    _vars.erase(_vars.begin(), _vars.end());
}

/**
 * Accessor to persuade a segmented string to produce
 * the resulting string generated by filling in the variables
 * The variables are requested using the TContext
 **/
#ifdef __VAX
string TSegStr::value(TContext& ctxt)
#else
string TSegStr::value(TContext& ctxt) const
#endif
{
    try
    {
        vector<string*>::const_iterator t = _txts.begin();
        vector<TVarDetails*>::const_iterator v = _vars.begin();
        string result;
        // Hopefully get area of memory big enough to hold string first time
        result.reserve(initialSize);
        // Build up the string.
        while((t != _txts.end()) || (v != _vars.end()))
        {
            if(t != _txts.end())
            {
                if(*t != NULL) result.append(**t);
                t++;
            }
            if(v != _vars.end())
            {
                if(*v != NULL)
                {
                    if((*v)->noname())
                    {
                        result.append(1, VAR_DELIM);
                    }
                    else
                    {
                        try
                        {
                            TtVarRef varref(&ctxt);		    // Variable reference object
                            string varnam;			    // Variable name: excluding indices
                            string varidx;			    // Variable index, e.g. "2,3"
                            (void) varref.Parse((*v)->text());  // Parse variable reference
                            varnam = varref.Varnam();	    // ...Get variable name
                            varidx = varref.Evaluate(ctxt);	    // ...and variable index
                            result.append(ctxt.get_variable(varnam,
                                                            (*v)->current(),
                                                            varidx));
                        }
                        catch(mdbParseException&)
                        {
                            result.append(1, VAR_DELIM);
                            result.append((*v)->text());
                            result.append(1, VAR_DELIM);
                        }
                    }
                }
                v++;
            }
        }
        // Adjust our guess of how big string should be if needed.
        if(initialSize < result.capacity()) initialSize = result.capacity();
        return result;
    }
    catch(...)//TODO: remove
    {
        cout << "ERROR in TSegStr::value" << endl;
        throw;
    }
}
//@mfunc    find matching ']' after position of '[%' in variable
//@parm	    string to find ] in
//@parm	    position to start
//@retn	    position of ']'
string::size_type TSegStr::find_right_brkt(string str,string::size_type pos)
{
    //iterate through entire string counting brackets to account for
    //array indecies within indecies
    int brckt_count=1;//have 1 '[' so far
    for(string::size_type i=pos;i<str.length();i++)
    {
        if(str[i]=='[')brckt_count++;
        if(str[i]==']')
        {
            brckt_count--;
            if(brckt_count==0)return i;
        }
    }
    return -1;//No matching bracket , error
}
/**
 * Friend of TSegStr used to write out its contents to an ostream
 **/
ostream& operator<<(ostream& os, const TSegStr& s)
{
    vector<string*>::const_iterator t = s._txts.begin();

    vector<TVarDetails*>::const_iterator v = s._vars.begin();
    while((t != s._txts.end()) || (v != s._vars.end()))
    {
        if(t != s._txts.end())
        {
            if(*t != NULL) os << "txt: '" << **t << "'" << endl;
            t++;
        }
        if(v != s._vars.end())
        {
            if(*v != NULL) os << "var: '" << (*v)->text() << "'" << endl;
            v++;
        }
    }
    return os;
}


//
// Implementation of TSegStr Iterator...
//

// Constructor, from a TSegStr, initialises to begining
TSegStr_iterator::TSegStr_iterator(const TSegStr* segStr) :
        _segStr(segStr), _ctxt(NULL)
{
    if (_segStr)
    {
        _litIter = segStr->_txts.begin();
        _varIter = segStr->_vars.begin();

        // 1st retrievalable value is a literal ?
        _literal = (!(_segStr->_txts.empty()) && (*_litIter != NULL));

        if (!_literal && _litIter != _segStr->_txts.end())
            ++_litIter;
    }
    else
    {
#ifdef USE_STL_PORT
        //        _litIter = string::end();
        //        _varIter = string::end();
#else
        _litIter = NULL;
        _varIter = NULL;
#endif
        _literal = true;
    }
};


bool
TSegStr_iterator::operator==(const TSegStr_iterator& rhs)
{
    if (this == &rhs) // physically the same object
        return true;

    // Points at the same position within the same TSegStr object
    return (_litIter == rhs._litIter && _varIter == rhs._varIter);
}


bool
TSegStr_iterator::operator!=(const TSegStr_iterator& rhs)
{
    if (&rhs == this) // physically the same object
        return false;

    // Different position within the same or different TSegStr object
    return (_litIter != rhs._litIter || _varIter != rhs._varIter);
}


// Pre Increment, ie ++Iter
TSegStr_iterator&
TSegStr_iterator::operator++()
{
    if (_segStr)
    {
        if (_literal)
        {
            if (_litIter != _segStr->_txts.end())
                ++_litIter;
            _literal = false;
        }
        else
        {
            if (_varIter != _segStr->_vars.end())
                ++_varIter;
            if (_litIter != _segStr->_txts.end())
            {
                _literal = (*_litIter != NULL);
                if (!_literal)
                    ++_litIter;
            }
        }
    }

    return *this;
}


// Post Increment, ie Iter++
TSegStr_iterator TSegStr_iterator::operator++(int)
{
    TSegStr_iterator oldValue(*this); // Save old value for return
    ++(*this); // Increment
    return oldValue;
}


// Dereference Iterator
string
TSegStr_iterator::operator*()
{
    static string nullString = "";

    if (!_segStr)
        return nullString;

    if (_literal)
    {
        if (_litIter != _segStr->_txts.end() && (*_litIter != NULL))
            return(**_litIter);
    }
    else
    {
        if (_varIter != _segStr->_vars.end() && (*_varIter != NULL))
        {
            if((*_varIter)->noname())
            {
                return string(1, TSegStr::VAR_DELIM);
            }
            else if (_ctxt)
            {
                try
                {
                    TtVarRef varref(_ctxt);		    // Variable reference object
                    string varnam;			    // Variable name: excluding indices
                    string varidx;			    // Variable index, e.g. "2,3"
                    (void) varref.Parse((*_varIter)->text());
                    // Parse variable reference
                    varnam = varref.Varnam();	    // ...Get variable name
                    varidx = varref.Evaluate(*_ctxt);    // ...and variable index
                    return _ctxt->get_variable(varnam,
                                               (*_varIter)->current(),
                                               varidx);
                }
                catch(mdbParseException&)
                {
                    return (TSegStr::VAR_DELIM +
                            (*_varIter)->text() +
                            TSegStr::VAR_DELIM);
                }
            }
            else
                return (TSegStr::VAR_DELIM +
                        (*_varIter)->text() +
                        TSegStr::VAR_DELIM);
        }
    }

    return nullString;
}
