/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TSegStr.h
#ifndef CYR_TSEGSTR_H
#define CYR_TSEGSTR_H
/**
 * Object to represent a segmented string
 * This is a way of storing a string that contains variable
 * substitution information.
 **/

#if defined (WIN32)
#pragma warning(disable:4786)
#endif

#include <string>
#include <vector>
#include <iterator>

#ifndef CYR_NO_STD_NAMESPACE
using namespace std;
#endif
#ifdef __VAX
#define mutable
#endif
#if defined (WIN32)
//++
class TSegStr;
ostream& operator <<(ostream&, const TSegStr&);
//--
#endif

class TContext;

/**
 * Class to hold the details of a variable
 **/
class TVarDetails
{
    string _orig;
    bool _current;
    string _index;
    string _varname;
public:
    // Constructors and destructor
    TVarDetails(const string& orig);

    // Accessors
    const string& name() const
    {
        return _varname;
    };
    const string& text() const
    {
        return _orig;
    };
    string index() const
    {
        return _index;
    };
    bool current() const
    {
        return _current;
    };
    bool noname() const
    {
        return _orig.empty() ? true : false;
    };
}
; // TVarDetails


class TSegStr; // forward declaration.

//
// LINUX Port Changes
//
#ifdef LINUX
class TSegStr_iterator : public std::bidirectional_iterator<input_iterator_tag, string>
#elif defined(HPUX)
class TSegStr_iterator : public bidirectional_iterator<input_iterator_tag, string>
#else
class TSegStr_iterator : public std::iterator<input_iterator_tag, string>
#endif
{
public:
    TSegStr_iterator(vector<string*>::const_iterator litIter,
                 vector<TVarDetails*>::const_iterator varIter) :
    _segStr(NULL), _ctxt(NULL), _literal(true),
    _litIter(litIter), _varIter(varIter)
    {}

    ; // Empty (end) iterator

TSegStr_iterator(const TSegStr_iterator& rhs) :
    //std::iterator<input_iterator_tag, T>(rhs),

    //
    // LINUX Port Changes
    //
#ifdef LINUX

    std::bidirectional_iterator<input_iterator_tag, string>(rhs),
    _segStr(rhs._segStr), _ctxt(rhs._ctxt), _literal(rhs._literal),
    _litIter(rhs._litIter), _varIter(rhs._varIter)
    {}
    ;
#elif defined(HPUX)

bidirectional_iterator<input_iterator_tag, string>(rhs),
    _segStr(rhs._segStr), _ctxt(rhs._ctxt), _literal(rhs._literal),
    _litIter(rhs._litIter), _varIter(rhs._varIter)
    {}
    ;
#else
    std::iterator<input_iterator_tag, string>(rhs),
    _segStr(rhs._segStr), _ctxt(rhs._ctxt), _literal(rhs._literal),
    _litIter(rhs._litIter), _varIter(rhs._varIter)
    {}
    ;

#endif

    // Constructor, from a TSegStr, initialises to begining
    TSegStr_iterator(const TSegStr* segStr);

    bool operator==(const TSegStr_iterator& rhs);
    bool operator!=(const TSegStr_iterator& rhs);

    TSegStr_iterator& operator++();		// Pre Increment
    TSegStr_iterator  operator++(int);	// Post Increment

    //T operator*(); // Dereference Iterator
    string operator*(); // Dereference Iterator

    void SetContext(TContext* ctxt)
    {
        _ctxt = ctxt;
    }

private:
    const TSegStr*  _segStr;
    TContext* _ctxt;
    bool _literal;
    vector<string*>::const_iterator _litIter;
    vector<TVarDetails*>::const_iterator _varIter;
};


// class definition
class TSegStr
{

    vector<string*> _txts;
    vector<TVarDetails*> _vars;

    mutable string::size_type initialSize;
public:


    friend class TSegStr_iterator;

    static const char VAR_DELIM;
    static const char CURVAR_MODIFIER;
    static const char LFT_INDEX_DELIM;
    static const char RGT_INDEX_DELIM;
    static const char RANDOM_FLAG;

    // constructors & destructor
    TSegStr(const string& val, string::size_type searchPos = 0);
    virtual ~TSegStr();

    // accessors
    static TSegStr* createIfNeeded(const string& value);
#ifdef __VAX
    string value(TContext& ctxt);
#else
    string value(TContext& ctxt) const;
#endif
    //Public function for string parsing of array expressions
    string::size_type find_right_brkt(string str,string::size_type pos);
    // mutators

    // iterators
    typedef TSegStr_iterator iterator;
    typedef const TSegStr_iterator const_iterator;
    iterator begin()
    {
        return iterator(this);
    }
    const_iterator begin() const
    {
        return iterator(this);
    }
    iterator end()
    {
        return iterator(_txts.end(), _vars.end());
    }
    const_iterator end() const
    {
        return iterator(_txts.end(), _vars.end());
    }

    // friends
    friend ostream& operator<<(ostream& os, const TSegStr& s);
}
; // TSegStr



#endif // CYR_TSEGSTR_H
