/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CECTofTxt.cpp
// @module  CECTofTxt and derived classes | This module contains the
//	    implementations of these classes. The classes handle the parsing,
//	    encoding and evaluation of text parameters (stxt and ltxt) within
//	    TOF records. Text parameters may contain, for example, boolean,
//	    character and integer strings. The CECTofTxt base class provides
//	    common functions for parsing and a common routine for evaluation.
//	    The derived classes simply provide the high-level parameter
//	    specific parsing, e.g. that for parsing a boolean expression.
//
//	    Parsing produces an encoded string in postfix format,
//	    representing the input string - this makes for easier and more
//	    efficient evaluation, e.g.:
//
//	    <PRE>
//	    Input string:   (!((Iret = 0) & (Jret = 0)))
//	    Encoded string: iret, 0, =, jret, 0, =, &, !
//	    </PRE>
//
//	    The classes in this module use a local class (ParseException) to
//	    signal parsing errors, which should be trapped and processed by
//	    the main, public member functions. When a ParseException is thrown,
//	    a description of the parsing error is placed in CECTofTxt::_reason.
//	    The reasons for encoding and exection falures are also placed in
//	    this field. The accessor function CECTofTxt::Reason() should be
//	    used to retrieve the _reason string in the event of a parsing,
//	    encoding or execution error.
///////////////////////////////////////////////////////////////////////////////

#include <psl_macros.h>

//Function prototypes.
#ifdef __cplusplus
extern "C"
{
#include <psl_rtndef.h>
}
#endif // _cplusplus

#include <cmath>
#include <string>

#include "CECOperand.h"
#include "CECTofTxt.h"
#include "CECVarOptions.h"
#include "TContext.h"
#include "TScript.h"
#include "TPLog.h"
#include "assert.h"

// Constants used externally to indicate success or failure.
const string CECTofTxt::stFALSE = "0";
const string CECTofTxt::stTRUE = "1";

// Constants for symbols used in parsing.
const string ps_squote		    = "\'";	// Single quote
const string ps_dquote		    = "\"";	// Double quote
const string ps_stcond		    = "(";	// Start condtion
const string ps_econd		    = ")";	// End condition
const string ps_starray_indices	    = "[";	// Start array indices
const string ps_earray_indices	    = "]";	// End array indices
const string ps_stprm		    = "(";	// Start parameter list
const string ps_eprm		    = ")";	// End parameter list
const string ps_stopt		    = ",";	// Start option clause
const string ps_prmsep		    = ",";	// Parameter separator
const string ps_idxsep		    = ",";	// Array index separator
const char   ps_idxsepc		    = ',';	// Array index separator as character
const string ps_valsep		    = ",";	// Value separator
const char   ps_valsepc		    = ',';	// Value separator as character
const string ps_fnc_extract	    = "~EXTRACT"; // Function ~EXTRACT
const string ps_fnc_locate	    = "~LOCATE"; // Function ~LOCATE
const string ps_opr_ne		    = "!=";	// Binary NOT EQUALS operator
const string ps_opr_not		    = "!";	// Unary logical NOT operator
const string ps_opr_and		    = "&";	// Binary logical AND operator
const string ps_opr_or		    = "|";	// Binary logical OR operator
const string ps_opr_eq		    = "=";	// Binary EQUALS operator
const string ps_opr_not_contains    = "<^>";	// Binary NOT CONTAINS operator
const string ps_opr_le		    = "<=";	// Binary LESS THAN OR EQUAL TO operator
const string ps_opr_lt		    = "<";	// Binary LESS THAN operator
const string ps_opr_ge		    = ">=";	// Binary GREATER THAN OR EQUAL TO operator
const string ps_opr_gt		    = ">";	// Binary GREATER THAN operator
const string ps_opr_contains	    = "^";	// Binary CONTAINS operator
const string ps_opr_add		    = "+";	// Binary ADD operator
const string ps_opr_sub		    = "-";	// Binary SUBTRACT operator
const string ps_opr_mul		    = "*";	// Binary MULTIPLY operator
const string ps_opr_div		    = "/";	// Binary DIVIDE operator
const string ps_opr_mod		    = "|";	// Binary MODULO operator
const char   ps_stchr_cmndc	    = '<';	// Start character command specifier
const char   ps_echr_cmndc	    = '>';	// End character command specifier
const string ps_key_case_blind	    = "CASE_BLIND";
// Keyword: CASE_BLIND

const char   ps_cmnd_chr	    = '~';	// The command character
const char   ps_ctrl_chr	    = '^';	// The control character

// List of 'arithmetic' operators valid for character operands. In list of string operators
// longer variants with the same leading character(s) as another operator MUST appear
// first, e.g. "<=" must be before "<", so that string comparisons match the correct
// values.
const int chrops_end = 2;  // Number of operators
const string chrops[] =    // List of operators as strings
    {
        ps_opr_add, ps_opr_sub
    };
const Operator chropr[] =  // List of operator codes (must correspond to above)
    {
        opr_add, opr_sub
    };

// List of arithmetic operators valid for integer operands. In list of string operators
// longer variants with the same leading character(s) as another operator MUST appear
// first, e.g. "<=" must be before "<", so that string comparisons match the correct
// values.
const int intops_end = 5;  // Number of operators
const string intops[] =    // List of operators as strings
    {
        ps_opr_add, ps_opr_sub, ps_opr_mul, ps_opr_div, ps_opr_mod
    };
const Operator intopr[] =  // List of operator codes (must correspond to above)
    {
        opr_add, opr_sub, opr_mul, opr_div, opr_mod
    };

// List of logical operators valid for character operands. In list of string operators
// longer variants with the same leading character(s) as another operator MUST appear
// first, e.g. "<=" must be before "<", so that string comparisons match the correct
// values.
const int logchrops_end = 8;  // Number of operators
const string logchrops[] =    // List of operators as strings
    {
        ps_opr_not_contains, ps_opr_contains, ps_opr_le, ps_opr_lt,
        ps_opr_ne, ps_opr_eq, ps_opr_ge, ps_opr_gt
    };
const Operator logchropr[] =  // List of operator codes (must correspond to above)
    {
        opr_not_contains, opr_contains, opr_le, opr_lt,
        opr_ne, opr_eq, opr_ge, opr_gt
    };

// List of logical operators valid for integer operands. In list of string operators
// longer variants with the same leading character(s) as another operator MUST appear
// first, e.g. "<=" must be before "<", so that string comparisons match the correct
// values.
const int logintops_end = 6;  // Number of operators
const string logintops[] =    // List of operators as strings
    {
        ps_opr_le, ps_opr_lt, ps_opr_ne, ps_opr_eq, ps_opr_ge, ps_opr_gt
    };
const Operator logintopr[] =  // List of operator codes (must correspond to above)
    {
        opr_le, opr_lt, opr_ne, opr_eq, opr_ge, opr_gt
    };

// Function option mask values.
#define FNCOPT_CASE_BLIND     (unsigned long)0x00000001	/* Case Blind Bit */

// ERROR: Throw 'Premature End of String' exception.
#define ERROR_PREMEOS throw ParseException("Premature end of string.");

// Skip whitespace from position 'p' in string 's'.
//#define SKIP_WHITESPACE(s, p)   while(isspace(s[p])){p++;if(p>s.length())break;}
inline void SKIP_WHITESPACE(const string &s, int &p)
{
    while(p < s.length() && isspace(s[p]))
    {
        p++;
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor.
//
// @parm
//
CECExprDsc::CECExprDsc(DataType dTyp, bool isVariable, string varNam,
                       list<CECOperand>::const_iterator exprIt) :
        _dTyp(dTyp), _isVariable(isVariable), _varNam(varNam), _exprIt(exprIt)
{
    ; // Do nothing
} // End CECExprDsc()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Default destructor - does nothing.
//

CECExprDsc::~CECExprDsc()
{
    ; // Do nothing
} // End ~CECExprDsc()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor. Save pointer to the TScript record for the current TOF
//	    file.
//
// @parm    Pointer to TScript record for current TOF file.
//

CECTofTxt::CECTofTxt(const TScript* script) :
        _scriptPtr(script), _idxLevel(0), _opdCnt(0), _isVariable(false), _varNam(""),
        _notOprSpecified(false)
{
    ; // Do nothing
} // End CECTofTxt()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor. Save pointer to the TContext
//
// @parm    Pointer to TContext
//

CECTofTxt::CECTofTxt(TContext* ctxt) :
        _idxLevel(0), _opdCnt(0), _isVariable(false), _varNam(""), _notOprSpecified(false)
{
    assert(ctxt);
    _scriptPtr = ctxt->Script();
} // End CECTofTxt()

///////////////////////////////////////////////////////////////////////////////
// @mfunc Copy Constructor,
//    Lists are empty during construction because of Problems on solaris
CECTofTxt::CECTofTxt(const CECTofTxt& tftxt)
{
    operator=(tftxt);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Default destructor - does nothing.
//
CECTofTxt::~CECTofTxt()
{
    ; // Do nothing
} // End ~CECTofTxt()


const CECTofTxt& CECTofTxt::operator=(const CECTofTxt& tt)
{
    _reason         = tt._reason;
    _idxLevel       = tt._idxLevel;
    _opdCnt         = tt._opdCnt;
    _isVariable     = tt._isVariable;
    _varNam         = tt._varNam;
    _dTyp           = tt._dTyp;
    _scriptPtr      = tt._scriptPtr;
    _exprIt         = tt._exprIt;
    _prsStack       = tt._prsStack;
    _notOprSpecified= tt._notOprSpecified;
    _exprEncoded    = tt._exprEncoded;
    _exprList       = tt._exprList;

    return *this;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This mutator function evaluates an encoded string set up by a
//	    ::Parse() function. An encoded string consists of a list of
//	    operands and operators in reverse Polish format, e.g.
//
//	    <PRE>
//	    Input string:   (!((Iret = 0) & (Jret = 0)))
//	    Encoded string: iret, 0, =, jret, 0, =, &, !
//	    </PRE>
//
//	    This function traverses the list from left to right evaluating each
//	    part of the expression and pushing the result onto a local stack. If
//	    there are no operands in the list, they are popped from the local
//	    stack. When evaluation is complete the result should be on the top of
//	    the stack.
//
//	    Operands may be variables, immediate character values or immediate
//	    integer values. Operators may operate on integer or character
//	    operands.
//
//	    If any parsing/evaluating error occurs then throw an mdbParseException
//	    with as much information as is to hand. this should cause the thread
//	    executing this statement to shut down neatly (there will probably
//	    be some memory leakage but as this is a fairly catstrophic error
//	    anyway it will often result in the Process ending)
// @rdesc   Result - as a string - of evaluating the encoded expression.
// @parm    Rerefernce to current context.
//
string CECTofTxt::Evaluate(TContext& ctxt) const
{
    static const string noResult = "";
    if (_exprEncoded.empty())
    {
        return noResult;
    }

    // Process encoded expression.
    stack<CECOperand*> operands;    // Stack of operand pointers
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
    {
        if (!_evaluateOpd(ctxt, &operands, &it)) return noResult;
    }

    if (operands.empty())
    {
        string s("Illformed expression, no result left on stack at end");
        ctxt.SetErrorMessage (s);
        throw mdbParseException(s);// Underflow
    }

    else if (operands.size() > 1)
    {
        string s("Illformed expression, too many results left on stack at end");
        ctxt.SetErrorMessage (s);
        throw mdbParseException(s);// Underflow
    }

    // Get the result, and convert to a string representation (no trailing zero's).

    string resultStr;
    CECOperand* result = operands.top();
    if (result->IsChr())
    {
        resultStr = result->ChrVal();
    }
    else
    {
        char resStr[32];
        sprintf(resStr, "%li", result->IntVal());
        resultStr = string(resStr);
    }

    delete result;
    return resultStr;
} // End CECTofTxt::Evaluate()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to process an encoded expression containing
//	    an ~EXTRACT function reference. The routine returns the result of the
//	    function as a string. An empty string is returned if the current
//	    token is not an ~EXTRACT function operand. On return, the encoded
//	    expression iterator (used to obtain operands) passed to this
//	    routine will be updated, as necessary, so that it always refers to
//	    the last operand procesed by this routine.
//
//	    The following illustrates the format of the operands for the
//	    ~EXTRACT function.
//
//	    <PRE>
//	    ~EXTRACT (0, 15, ERRSTR)
//
//	    OPD:    FUNCTION ~EXTRACT
//	    OPD:    0
//	    OPD:    NULL (terminates 'offset' parameter)
//	    OPD:    15
//	    OPD;    NULL (terminates 'length' parameter)
//	    OPD:    ERRSTR
//	    OPD:    NULL (terminates 'string' parameter)
//	    </PRE>
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
// @parm    Reference to current context.
// @parm    Pointer to encoded expression iterator. Assumed to be pointing to
//	    the ~EXTRACT function operand that immediately precedes the first
//	    function parameter operand.
// @parm    Pointer to string to hold function result string on exit. Empty
//	    string returned if not an ~EXTRACT function.
bool CECTofTxt::_evaluateFncExtract(TContext& ctxt,
                                    list<CECOperand>::const_iterator* itP,
                                    string* resstrP) const
{
    CECOperand token = **itP;	    // Current token
    stack<CECOperand*> operands;    // Local stack of operand pointers
    const long prmcnt = 3;	    // Number of parameters
    long extoff;		    // ~EXTRACT offset
    long extlen;		    // ~EXTRACT length
    string extstr;		    // ~EXTRACT string to extract from
    string resstr = "";		    // Local result string

    // Iterate through the parameters in the encoded expression, saving each
    // parameter value.
    for (long isub = 0; (isub < prmcnt);)
    {
        // Bump iterator. If we have reached the end of the encoded expression
        // report an error, as we have not had all the required subscripts.
        (*itP)++;
        if (*itP == _exprEncoded.end())
        {
            ctxt.SetErrorMessage ("Illformed ~EXTRACT function reference, too few parameters");
            return false;		// Invalid ~EXTRACT reference
        }


        token = **itP;

        if (token.Addrm() == addr_none)
        {
            if (operands.size() != 1)
            {
                ctxt.SetErrorMessage ("Illformed ~EXTRACT function parameter");
                return false;		// Invalid ~EXTRACT function reference
            }


            CECOperand* prmvalP;	// Pointer to parameter value operand
            prmvalP = operands.top();	// Pop subscript value operand off stack
            operands.pop();		// ...

            if (isub == 0)
            {
                extoff = prmvalP->IntVal();
            }
            else if (isub == 1)
            {
                extlen = prmvalP->IntVal();
            }
            else
            {
                extstr = prmvalP->ChrVal();
            }

            isub += 1;		    // One more parameter evaluated
            delete prmvalP;
        }
        else
        {
            _evaluateOpd (ctxt, &operands, itP);
        }
    }

    // The local operands stack should be empty - as we should empty the stack after
    // processing each parameter. If the stack is not empty, report an error.
    if (!operands.empty())
    {
        ctxt.SetErrorMessage ("Illformed ~EXTRACT function reference");
        return false;
    }

    // Do the 'extract'.
    resstr = extstr.substr(extoff, extlen);

    // Return function value.
    *resstrP = resstr;
    return true;
} // End CECTofTxt::_evaluateFncExtract()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to process an encoded expression containing
//	    a ~LOCATE function reference. The routine returns the result of the
//	    function as a string. An empty string is returned if the current
//	    token is not a ~LOCATE function operand. On return, the encoded
//	    expression iterator (used to obtain operands) passed to this
//	    routine will be updated, as necessary, so that it always refers to
//	    the last operand procesed by this routine.
//
//	    The following illustrates the format of the operands for the
//	    ~LOCATE function.
//
//	    <PRE>
//	    ~LOCATE ("*Warning", ERRSTR)
//
//	    OPD:    FUNCTION ~LOCATE
//	    OPD:    "*Warning"
//	    OPD:    NULL (terminates first parameter)
//	    OPD:    ERRSTR
//	    OPD;    NULL (terminates second subscript expression)
//	    OPD:    0 (function options mask)
//	    OPD:    NULL (terminates options mask)
//	    </PRE>
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
// @parm    Reference to current context.
// @parm    Pointer to encoded expression iterator. Assumed to be pointing to
//	    the ~LOCATE function operand that immediately precedes the first
//	    function parameter operand.
// @parm    Pointer to string to hold function result string on exit. Empty
//	    string returned if not a ~LOCATE function.
bool CECTofTxt::_evaluateFncLocate(TContext& ctxt,
                                   list<CECOperand>::const_iterator* itP,
                                   string* resstrP) const
{
    CECOperand token = **itP;	    // Current token
    stack<CECOperand*> operands;    // Local stack of operand pointers
    const long prmcnt = 3;	    // Number of parameters (inc. option mask)
    string locsubstr;		    // ~LOCATE substring
    string locstr;		    // ~LOCATE string to search
    long locopt;		    // ~LOCATE options mask
    string resstr = "";		    // Local result string

    // Iterate through the parameters in the encoded expression, saving each
    // parameter value.
    for (long isub = 0; (isub < prmcnt);)
    {
        // Bump iterator. If we have reached the end of the encoded expression
        // report an error, as we have not had all the required subscripts.
        (*itP)++;
        if (*itP == _exprEncoded.end())
        {
            ctxt.SetErrorMessage ("Illformed ~LOCATE function reference, too few parameters");
            return false;		// Invalid ~LOCATE reference
        }


        token = **itP;

        if (token.Addrm() == addr_none)
        {
            if (operands.size() != 1)
            {
                ctxt.SetErrorMessage ("Illformed ~LOCATE function parameter");
                return false;		// Invalid ~LOCATE function reference
            }


            CECOperand* prmvalP;	// Pointer to parameter value operand
            prmvalP = operands.top();	// Pop subscript value operand off stack
            operands.pop();		// ...

            if (isub == 0)
            {
                locsubstr = prmvalP->ChrVal();
            }
            else if (isub == 1)
            {
                locstr = prmvalP->ChrVal();
            }
            else
            {
                locopt = prmvalP->IntVal();
            }

            isub += 1;		    // One more parameter evaluated
            delete prmvalP;
        }
        else
        {
            _evaluateOpd (ctxt, &operands, itP);
        }
    }

    // The local operands stack should be empty - as we should empty the stack after
    // processing each parameter. If the stack is not empty, report an error.
    if (!operands.empty())
    {
        ctxt.SetErrorMessage ("Illformed ~LOCATE function reference");
        return false;
    }

    // Convert strings to same case if a case blind locate is required.
    if (locopt && FNCOPT_CASE_BLIND)
    {
        string_toupper(locsubstr);		// UPPERCASE substring
        string_toupper(locstr);			// UPPERCASE string
    }

    // Do the 'locate'. If string found, result is the offset of the first
    // character in the target string, otherwise result is "-1".

    long resval = locstr.find(locsubstr);
    if (resval == string::npos) resval = -1;	// String not found

    // Set up result as a string.
    char valbuf[32];				// Result value buffer
    sprintf(valbuf, "%li", resval);		// Get result as string
    resstr = valbuf;				// ...

    // Function evaluated successfully, return function value.
    *resstrP = resstr;
    return true;
} // End CECTofTxt::_evaluateFncLocate()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to process an operand during expression
//	    evaluation. A pointer to a stack of operands is passed to this
//	    routine and used to maintain the 'results' of evaluation. Some
//	    operands require multiple operands to describe them (e.g.
//	    subscripted variables and function references). This routine will
//	    process all the operands associated with such operands. On return,
//	    the encoded expression iterator passed to this routine will be
//	    updated, as necessary so that it always refers to the last
//	    operand procesed by this routine.
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
// @parm    Reference to current context.
// @parm    Pointer to stack of operand pointers holding evaluation 'results'.
// @parm    Pointer to encoded expression iterator.
//
bool CECTofTxt::_evaluateOpd(TContext& ctxt,
                             stack<CECOperand*>* operandsP,
                             list<CECOperand>::const_iterator* itP) const
{
    CECOperand token = **itP;	    // Current token

    // If variable, get and push variable value on operand stack.
    if (token.Addrm() == addr_variable)
    {
        string varval;		    // Variable value as string
        string varIdx = "";	    // Variable index string
        if (token.VarDims() != 0) _evaluateVarIdx(ctxt, itP, &varIdx);

        try
        {
            varval = ctxt.get_variable(token.VarNam(), false, varIdx);
        }
        catch (mdbParseException&)
        {
            ctxt.SetErrorMessage("No value available for variable: " + token.VarNam());
            return false;
        }

        if (token.IsInt())				// If integer variable
        {
            long intval = atol(varval.c_str());	// Get integer value
            (*operandsP).push(new CECOperand(intval));	// Push integer value
        }

        else if (token.IsChr())			// Else if character variable
        {
            (*operandsP).push(new CECOperand(varval, addr_immediate));
        }
        else
        {
            ctxt.SetErrorMessage("Unrecognised data tye for variable: " +
                                 token.VarNam());
            return false;
        }
    }

    // Else if immediate operand, push operand on operand stack.
    else if (token.Addrm() == addr_immediate)
    {
        (*operandsP).push(new CECOperand(token));
    }

    // Else if operator, operate on operand(s) on top of stack and push result on stack.
    else if (token.Addrm() == addr_operator)
    {
        // Always pop at least one operand, reporting an error if we cannot.
        if ((*operandsP).empty())		// Underflow
        {
            ctxt.SetErrorMessage ("Stack underflow, insufficient operands for operator");
            return false;
        }
        CECOperand* rhs = (*operandsP).top();	// Pop first operand
        (*operandsP).pop();				// ...

        // If we have a binary operator, pop another operand.
        CECOperand* lhs = NULL;
        if (token.Opr() != opr_not)		// If binary operator
        {
            if ((*operandsP).empty())		// Underflow
            {
                ctxt.SetErrorMessage("Stack underflow, insufficient operands for operator");
                return false;
            }
            lhs = (*operandsP).top();		// Pop second operand
            (*operandsP).pop();			// ...
        }

        // Perform operation on operand(s), and push result back onto stack.

        CECOperand* tmpopd = NULL;
        switch (token.Opr())
        {
        case opr_ne:	    // NOT EQUAL TO operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    tmpopd = new CECOperand(lhs->ChrVal() != rhs->ChrVal());
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() != rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_not:	    // Logical NOT operator
            {
                (*operandsP).push(new CECOperand(!rhs->IntVal()));
                break;
            }

        case opr_le:	    // LESS THAN OR EQUAL TO operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    tmpopd = new CECOperand(lhs->ChrVal() <= rhs->ChrVal());
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() <= rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_lt:	    // LESS THAN operator
            {
                if (lhs->IsChr())
                {
                    tmpopd = new CECOperand(lhs->ChrVal() < rhs->ChrVal());
                }
                else
                {
                    tmpopd = new CECOperand(lhs->IntVal() < rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_ge:	    // GREATER THAN OR EQUAL TO operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    tmpopd = new CECOperand(lhs->ChrVal() >= rhs->ChrVal());
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() >= rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_gt:	    // GREATER THAN operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    tmpopd = new CECOperand(lhs->ChrVal() > rhs->ChrVal());
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() > rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_eq:	    // EQUAL TO operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    tmpopd = new CECOperand(lhs->ChrVal() == rhs->ChrVal());
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() == rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_contains:	    // CONTAINS operator
            {
                tmpopd = new CECOperand(lhs->ChrVal().find(rhs->ChrVal()) != string::npos);
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_and:	    // Logical AND operator
            {
                (*operandsP).push(new CECOperand(lhs->IntVal() && rhs->IntVal()));
                break;
            }

        case opr_or:	    // Logical OR operator
            {
                (*operandsP).push(new CECOperand(lhs->IntVal() || rhs->IntVal()));
                break;
            }

        case opr_add:	    // ADDITION operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    string opdstr = lhs->ChrVal() + rhs->ChrVal();
                    tmpopd = new CECOperand(opdstr, addr_immediate);
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() + rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_sub:	    // SUBTRACTION operator
            {
                if (lhs->IsChr())   // If character operands
                {
                    string opdstr = lhs->ChrVal();		// Operand string
                    int rpos = opdstr.find (rhs->ChrVal());	// Pos of string to remove
                    if (rpos != string::npos)			// If sub-string found
                    {
                        int rlen = rhs->ChrVal().length();	// Remove it
                        opdstr.replace (rpos, rlen, "");		// ...
                    }

                    tmpopd = new CECOperand(opdstr, addr_immediate);
                }
                else		    // Else must be integer operands
                {
                    tmpopd = new CECOperand(lhs->IntVal() - rhs->IntVal());
                }
                (*operandsP).push(tmpopd);
                break;
            }

        case opr_mul:	    // MULTIPLY operator
            {
                (*operandsP).push(new CECOperand(lhs->IntVal() * rhs->IntVal()));
                break;
            }

        case opr_div:	    // DIVIDE operator
            {
                if (rhs->IntVal() != 0)
                {
                    (*operandsP).push(new CECOperand(lhs->IntVal() / rhs->IntVal()));
                }
                else
                {
                    ctxt.SetErrorMessage("Attempted divide by ZERO");
                    return false;
                }
                break;
            }

        case opr_mod:	    // MODULUS operator, i.e. remainder when lhs / rhs
            {
                if (rhs->IntVal() != 0)
                {
                    (*operandsP).push(new CECOperand(lhs->IntVal() % rhs->IntVal()));
                }
                else
                {
                    ctxt.SetErrorMessage("Attempted modulus divide by ZERO");
                    return false;
                }
                break;
            }

        default:		    // Unrecognised operator
            {
                ctxt.SetErrorMessage("Unknown operator encountered");
                return false;
            }
        } // End switch (token.Opr())

        delete lhs;
        delete rhs;
    }

    // Else if function, get and push function value on operand stack.
    else if (token.Addrm() == addr_function)
    {
        string fncVal;		    // Function value as string
        switch (token.FncCod())
        {
        case fnc_extract:	    // ~EXTRACT function
            {
                _evaluateFncExtract(ctxt, itP, &fncVal);
                break;
            }
        case fnc_locate:	    // ~LOCATE function
            {
                _evaluateFncLocate(ctxt, itP, &fncVal);
                break;
            }
        default:		    // Unrecognised function
            {
                ctxt.SetErrorMessage("Unknown function encountered");
                return false;
            }
        } // End switch (token.FncCod())

        if (token.IsInt())			// If integer function
        {
            long intval = atol(fncVal.c_str());	// Get integer value
            (*operandsP).push(new CECOperand(intval));	// Push integer value
        }

        else if (token.IsChr())			// Else if character function
        {
            (*operandsP).push(new CECOperand(fncVal, addr_immediate));
        }
        else
        {
            ctxt.SetErrorMessage("Unrecognised data type for function.");
            return false;
        }
    }

    return true;
} // End CECTofTxt::_evaluateOpd()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to process an encoded expression containing
//	    a list variable indices/subscripts. The routine returns the
//	    evaluated array subscripts (if any) as a character separated list
//	    of values, e.g. "24,2". An empty string is returned if the
//	    variable is not an array variable. On return, the encoded
//	    expression iterator (used to obtain operands) passed to this
//	    routine will be updated, as necessary, so that it always refers to
//	    the last operand procesed by this routine.
//
//	    The following illustrates the format of the operands for a
//	    subscripted variable CUSTOMERS [cuscod, isub + 2]:
//
//	    <PRE>
//	    OPD:    CUSTOMERS, 2 dimensions
//	    OPD:    CUSCOD
//	    OPD:    NULL (terminates first subscript expression)
//	    OPD:    ISUB
//	    OPD:    2
//	    OPD:    +
//	    OPD;    NULL (terminates second subscript expression)
//	    </PRE>
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
// @parm    Reference to current context.
// @parm    Pointer to encoded expression iterator. Assumed to be pointing to
//	    the subscripted variable operand that immediately precedes the first
//	    index operand.
// @parm    Pointer to string to hold variable index string on exit. Empty
//	    string returned if not an array variable.
bool CECTofTxt::_evaluateVarIdx(TContext& ctxt,
                                list<CECOperand>::const_iterator* itP,
                                string* idxstrP) const
{
    CECOperand token = **itP;	    // Current token
    stack<CECOperand*> operands;    // Local stack of operand pointers
    long vardims = token.VarDims(); // Number of variable dimensions
    string wrkstr = "";		    // Local index string

    // Iterate through the variable subscript operands in the encoded expression,
    // building up a string of comma separated array subscripts, e.g, "24,20".
    for (long isub = 0; (isub < vardims);)
    {
        // Bump iterator. If we have reached the end of the encoded expression
        // report an error, as we have not had all the required subscripts.
        (*itP)++;
        if (*itP == _exprEncoded.end())
        {
            ctxt.SetErrorMessage ("Illformed array reference, too few subscripts");
            return false;		// Invalid subscripted variable reference
        }


        token = **itP;

        if (token.Addrm() == addr_none)
        {
            if (operands.size() != 1)
            {
                ctxt.SetErrorMessage ("Illformed array subscript");
                return false;		// Invalid subscripted variable reference
            }


            CECOperand* subvalP;	// Pointer to subscript value operand
            subvalP = operands.top();	// Pop subscript value operand off stack
            operands.pop();		// ...

            if (isub != 0) wrkstr += ps_idxsep;	// Append subscript separator to index

            char substr[32];	    // Variable subscript value buffer
            sprintf(substr, "%li", subvalP->IntVal());
            wrkstr += substr;	    // Append subscript value to index string
            isub += 1;		    // One more array subscript evaluated
            delete subvalP;
        }
        else
        {
            _evaluateOpd (ctxt, &operands, itP);
        }
    }

    // The local operands stack should be empty - as we should empty the stack after
    // processing each subscript. If the stack is not empty, report an error.
    if (!operands.empty())
    {
        ctxt.SetErrorMessage ("Illformed array variable reference");
        return false;
    }

    // Subscripts evaluated successfully, return index string.
    *idxstrP = wrkstr;
    return true;
} // End CECTofTxt::_evaluateVarIdx()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Check that the expression is complete. This function is called
//	    after expression parsing has been completed to check that the
//	    operator stack has been correctly unwound, if it has not the
//	    expression is invalid - either too many or too few brackets.
//
void CECTofTxt::_chkExprComplete()
{
    // If operator stack has not been fully unwound the expression is invalid.
    if (_prsStack.size() != 1)
    {
        throw ParseException("Invalid expression structure");
    }
    return;
} // End _chkExprComplete()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Reset all the member data that needs to be reset in order to parse
//	    an expression string.
//
void CECTofTxt::_clrExpr()
{
    // Reset operator stack, encoded expression and 'NOT' operator specified flag.
    while (!_prsStack.empty()) _prsStack.pop();
    _prsStack.push(opr_none);
    _exprEncoded.erase(_exprEncoded.begin(), _exprEncoded.end());
    _notOprSpecified = false;
    return;
} // End _clrExpr()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Decrement expression level. If there is an operator at the current
//	    level encode it and pop it off the operator stack. If the condition
//	    at this level is to be negated there will be a 'NOT' operator on
//	    the stack; if there is encode a NOT operand and pop the NOT
//	    operator off the stack. If there is an operator defined for the new
//	    level, encode it and replace it with a NULL operator.
//
void CECTofTxt::_decExprLevel()
{
    // If operator stack is empty got too many ')'s - report an error.
    if (_prsStack.empty())
    {
        throw ParseException("Invalid expression - too many ')'s");
    }

    // Get and pop operator for current level - if there is one encode it.
    Operator opr = _prsStack.top();	// Get operator
    _prsStack.pop();			// Pop it off the stack

    if (opr != opr_none)		// If there is one encode it
    {
        CECOperand opd(opr);
        _exprEncoded.push_back(opd);
    }

    // If operator stack empty got too many ')'s - report an error.
    if (_prsStack.empty())
    {
        throw ParseException("Invalid expression - too many ')'s");
    }

    // If operator at top of stack is a NOT, encode it and pop it off the stack.
    opr = _prsStack.top();
    if (opr == opr_not)
    {
        CECOperand opd(opr);
        _exprEncoded.push_back(opd);
        _prsStack.pop();
        opr = _prsStack.top();
    }

    // If there is an operator at the new level, encode it and reset the operator to NONE.
    if (opr != opr_none)
    {
        CECOperand opd(opr);
        _exprEncoded.push_back(opd);
        _prsStack.pop();
        _prsStack.push(opr_none);
    }

    return;
} // End _decExprLevel()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Get the number of dimesnions in the supplied variable dimensions
//	    string. A variable dimesnions string consists of a character
//	    separated list of indices. Each index which may specify the range
//	    of values that the index may take, e.g. 10,100-999 represents 2
//	    indices, the first in the range 1 to 10 and the second in the range
//	    100 to 999.
//
// @rdesc   Number of dimesnions in supplied variable dimensions string.
// @parm    Dimensions string.
//
long CECTofTxt::_dimCnt (const string dimstr)
{
    // If there is no dimension string, we do not have an array - dimension count is zero.
    if (dimstr.length() == 0) return 0;

    // Here if we have a non-empty dimension string.
    long dimcnt = 1;		    // Dimensions count - must be at least one
    int dimlen = dimstr.length();   // Dimensions string length

    // The number of dimensions corresponds to the number of index separator characters + 1.
    for (int pos = 0;
            pos < dimlen;
            pos++)
    {
        if (dimstr[pos] == ps_idxsepc) dimcnt++;
    }

    return dimcnt;
} // End _dimCnt()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Create a function operand for a specified function and append it to
//	    the encoded expression buffer.
//
// @parm    Function code identifying function operand to be encoded.
//
void CECTofTxt::_encFnc(const FncCode fnccod)
{
    // Encode operand and update expression descriptor data. We do an 'insert'
    // rather than a push_back to retrieve the iterator for the new item.
    CECOperand opd(fnccod);
    list<CECOperand>::const_iterator it = _exprEncoded.insert(_exprEncoded.end(), opd);

    if (_idxLevel == 0)				// If not processing an array subscript
    {
        _isVariable = false;			// NOT a variable operand
        _varNam = "";				// ...
        _dTyp = opd.DTyp();			// Expression type confirmed
        if (_opdCnt == 0) _exprIt = it;		// Save iterator of first operand
        _opdCnt += 1;				// One more main expression operand
    }


    return;
} // End _encFnc()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   If there is an operator at the current level in the operator stack
//	    encode the operator and reset the operator for the current level to
//	    none.
//
void CECTofTxt::_encOpr()
{
    // If there is an operator at the current level, encode it and reset operator to NONE.
    Operator opr = _prsStack.top();

    if (opr != opr_none)
    {
        CECOperand opd(opr);
        _exprEncoded.push_back(opd);
        _prsStack.pop();
        _prsStack.push(opr_none);
    }

    return;
} // End _encOpr()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the substring formed by expanding the current
//	    token in the supplied string to the next blank - or end of string
//	    if no blank found. It is used to identify the current token when
//	    reporting an error.
//
// @rdesc   String formed by expanding current token in input string to next
//	    blank.
// @parm    Input string.
// @parm    Current position in input string..
//
const string CECTofTxt::_expandToBlank (const string inStr, const int pos)
{
    int newpos = inStr.find_first_of(" \t", pos);

    if (newpos == inStr.npos)	// If blank not found, substring extends to end of string
    {
        return inStr.substr(pos, inStr.length() - pos);
    }
    else    			// Else it extends to next blank
    {
        return inStr.substr(pos, newpos - pos);
    }
} // End _expandToBlank()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the substring formed by expanding the current
//	    token in the supplied string to end of string. It is typically used
//	    to identify extra text at the end of an otherwise valid string in
//	    reporting an error.
//
// @rdesc   String formed by expanding current token in input string to end of
//	    string.
// @parm    Input string.
// @parm    Current position in input string..
//
const string CECTofTxt::_expandToEos(const string inStr, const int pos)
{
    return inStr.substr(pos, inStr.length() - pos);
} // End _expandToEos()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Increment expression level. This just consists of pushing a value
//	    onto the operator stack to show that there is no operator current
//	    at the new level.
//
void CECTofTxt::_incExprLevel()
{
    _prsStack.push(opr_none);
    return;
} // End _incExprLevel()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses an array indices string. An array indices
//	    string has the following format:
//
//	    <PRE>
//	    [ index1 {,index2 {,index3}} ]
//	    </PRE>
//
//	    where 'indexn' is an integer expression.
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing array indices
// @parm    Position in input string at which to start parsing.
// @parm    Number of indices parsed and encode.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psArrayIndices (const string& inStr, int* pos, int* idxcnt,
                                 const bool updpos)
{
    enum States {array_indices_1, array_indices_2, array_indices_3};

    int	wrkpos = *pos;	// Get copy of string position for local update
    int wrkidxcnt = 0;	// Local index count - only update parameter if successful

    // Save starting stack size, so that we can check later whether the indices are complete
    // and push a null operand on stack for parsing the indices (remember to pop on exit).
    _prsStack.push(opr_none);				    // Add null operand
    stkOperator::size_type start_stksiz = _prsStack.size(); // Save starting stack size

    // Loop through the tokens in the input string encoding them as necessary.
    States state = array_indices_1;
    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE (inStr, wrkpos);

        switch (state)
        {
            // Start array indices. Expect "[".
        case array_indices_1:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_starray_indices.length()) == ps_starray_indices)
                {
                    wrkpos += ps_starray_indices.length();
                    _idxLevel += 1;
                    state = array_indices_2;
                }
                else
                {
                    _reason = "Expected '[', got \"" + _expandToBlank (inStr, wrkpos) + "\"";
                    return false;
                }
                break;
            }

            // Now expecting index (integer expression) - terminated by "," or "]".
        case array_indices_2:
            {
                if (wrkpos >= inStr.length())	    // If EOS
                {
                    ERROR_PREMEOS;
                }
                else if (_psIntExpr (inStr, &wrkpos))   // Else If index OK
                {
                    wrkidxcnt++;			    // Increment index count
                    CECOperand opd;			    // Encode terminating NULL operand
                    _exprEncoded.push_back(opd);	    // ...
                    state = array_indices_3;
                }
                else				    // Else index NOT OK
                {
                    throw ParseException("Invalid array subscript, \"" +
                                         _expandToBlank (inStr, wrkpos) + "\"");
                }
                break;
            }
            // Have parsed index, now expect "," or "]".
        case array_indices_3:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_earray_indices.length()) == ps_earray_indices)
                {
                    wrkpos += ps_earray_indices.length();
                    _idxLevel -= 1;
                    goto indices_done;
                }
                else if (inStr.substr(wrkpos, ps_idxsep.length()) == ps_idxsep)
                {
                    wrkpos += ps_idxsep.length();
                    state = array_indices_2;
                }
                else
                {
                    throw ParseException(
                        "Expected ']' or ',', got \"" + _expandToBlank (inStr, wrkpos) + "\"");
                }
                break;
            }
        default:
            {
                throw ParseException("Internal error - unrecognised state");
            }
        } // End switch (state)
    } // End while (pos < inStr.length())

    // Array indices parsed successfully, update string position (if required)
    // and index count.

indices_done:
    ;
    // If operator stack is not as it was on entry - report an error.
    if (_prsStack.size() != start_stksiz)
    {
        throw ParseException("Invalid indices - incorrect parentheses.");
    }
    _prsStack.pop();		// Remove operand pushed on entry

    if (updpos) *pos = wrkpos;
    *idxcnt = wrkidxcnt;
    return true;
} // End _psArrayIndices()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for a character expression string at the
//	    specified position within the input string. It optionally updates
//	    the string position to point to the character after the expression.
//	    A character expression string has the following format:
//
//	    <PRE>
//	    operand1 {operator operandn}
//	    </PRE>
//
//	    Any number of operator/operands may appear in an expression
//	    and parentheses may be used to aid legibility and/or change the
//	    evaluation order - by default evaluation is left to right.
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetects that the
//	    string being parsed is not a character expression a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
//	    Note that this routine does NOT ensure that the FULL string from
//	    the specified position to the end of the string is a valid character
//	    expression. There may be extraneous characters following the
//	    character expression. This allows this routine to be used to parse
//	    character expressions in SET commands etc.
//
// @rdesc   Return status. 'true' if character expression at specified position
//	    in input string, 'false' otherwise.
// @parm    String containing string to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the character expression.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psChrExpr (const string& inStr, int* pos, const bool updpos)
{
    int	wrkpos = *pos;	// Get copy of string position for local update

    enum States {chr_expr_1, chr_expr_2, chr_expr_3, chr_expr_4};

    //
    //  Save starting stack size, so that we can check later whether the expression is complete.
    //
    stkOperator::size_type start_stksiz = _prsStack.size();  // Save starting stack size

    //
    //  Loop through the input string tokens from the specified starting position
    //  and produce a corresponding postfix format operand list.
    //
    States state = chr_expr_1;  // Set initial parsing state

    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE(inStr, wrkpos);

        // Process current parse state.
        switch (state)
        {
            // Start a character expression. Expect "(" or operand.
        case chr_expr_1:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    // No state change
                }

                else if (_psChrValue(inStr, &wrkpos))
                {
                    state = chr_expr_2;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    state = chr_expr_2;
                }
                else if (_psChrFnc(inStr, &wrkpos))
                {
                    state = chr_expr_2;
                }
                else
                {
                    return false;	    // Not a character expression
                }

                break;
            }
            // Here after operand. Expect EOS or character operator.
        case chr_expr_2:
            {
                if (wrkpos >= inStr.length())
                {
                    ; // End of string valid, do nothing.
                }

                else
                {
                    int isub;   // Loop control
                    for (isub = 0; isub < chrops_end; isub++)
                    {
                        if (inStr.substr(wrkpos, chrops[isub].length()) == chrops[isub])
                        {
                            wrkpos += chrops[isub].length();
                            _saveOpr(chropr[isub]);
                            state = chr_expr_3;
                            break;
                        }
                    }
                    // If operator not found, we have a valid expression with trailing characters.
                    if (isub >= chrops_end) goto ChrExpr_parsed;
                }
                break;
            }
            // Here after operator. Expect "(" or operand.
        case chr_expr_3:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    // No state change
                }

                else if (_psChrValue(inStr, &wrkpos))
                {
                    _encOpr();
                    state = chr_expr_4;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    _encOpr();
                    state = chr_expr_4;
                }
                else if (_psChrFnc(inStr, &wrkpos))
                {
                    _encOpr();
                    state = chr_expr_2;
                }
                else
                {
                    throw ParseException("Operator must be followed by operand or '(', got \"" +
                                         _expandToBlank (inStr, wrkpos) + "\"");
                }
                break;
            }
            // Here after operand following an operator. Expect EOS, ")" or operator.
        case chr_expr_4:
            {
                if (wrkpos >= inStr.length())
                {
                    ; // End of string valid, do nothing.
                }

                else if (inStr.substr(wrkpos, ps_econd.length()) == ps_econd)
                {
                    wrkpos += ps_econd.length();
                    _decExprLevel();
                    // No state change
                }

                else
                {
                    state = chr_expr_2;
                }
                break;
            }
        default:
            {
                throw ParseException("Internal error - unrecognised state");
            }
        } // End switch (state)
    } // End while (wrkpos < inStr.length())


ChrExpr_parsed:
    ;
    // If operator stack has not been fully unwound the expression is invalid.
    if (_prsStack.size() != start_stksiz)
    {
        throw ParseException("Invalid expression structure");
    }

    if (updpos) *pos = wrkpos;

    return true;
} // End _psChrExpr()


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a predefined character function reference.
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing character function reference.
// @parm    Position in input string at which to start parsing.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position
//
bool CECTofTxt::_psChrFnc (const string& inStr, int* pos, const bool updpos)
{
    // Check for each of the predefined character functions.
    // Currently only support ~EXTRACT.
    if (_psFncExtract (inStr, pos, updpos)) return true;

    // Not a character function.
    return false;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for a character value at the specified
//	    position within the input string. It optionally updates the string
//	    position to point to the character after the character value. A
//	    character value is character string enclosed in either single or
//	    double quotes, i.e.
//
//	    <PRE>
//	    "..." or '...'
//	    </PRE>
//
// @rdesc   Return status. 'true' if character value at specified position
//	    in input string, 'false' otherwise.
// @parm    String to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the character value.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psChrValue(const string& inStr, int* pos, const bool updpos)
{
    int	wrkpos = *pos;	// Get copy of string position for local update
    string delim;	// Character value delimiter

    // See if we have first (required) character value delimiter.
    if (inStr.substr(wrkpos, ps_dquote.length()) == ps_dquote)
    {
        wrkpos += ps_dquote.length();	// Position to character after delimiter
        delim = ps_dquote;		// Delimiter is double quote
    }

    else if (inStr.substr(wrkpos, ps_squote.length()) == ps_squote)
    {
        wrkpos += ps_squote.length();	// Position to character after delimiter
        delim = ps_squote;		// Delimiter is sin;gle quote
    }

    else
    {
        return false;			// Not a character value
    }

    // Locate terminating character value delimiter. If not found - bad character value.

    int newpos = inStr.find(delim, wrkpos);
    if (newpos == inStr.npos)	// If delimiter not found
    {
        throw ParseException("No terminating delimiter for character value");
    }

    // Get the 'raw' character value represented by the parsed character value
    // by calling a routine to translate character and control commands into their
    // corresponding 'raw' characters.
    string chrval = _getChrValue(inStr, wrkpos, newpos);

    // Reset working position to first character after value.
    wrkpos = newpos + 1;

    // Encode operand and update expression descriptor data. We do an 'insert'
    // rather than a push_back to retrieve the iterator for the new item.
    CECOperand opd(chrval, addr_immediate);
    list<CECOperand>::const_iterator it = _exprEncoded.insert(_exprEncoded.end(), opd);

    if (_idxLevel == 0)				// If not processing an array subscript
    {
        _isVariable = false;			// NOT a variable operand
        _varNam = "";				// ...
        _dTyp = opd.DTyp();			// Expression type confirmed
        if (_opdCnt == 0) _exprIt = it;		// Save iterator of first operand
        _opdCnt += 1;				// One more main expression operand
    }

    // Update string position (if required) and return success.

    if (updpos) *pos = wrkpos;

    return true;
} // End _psChrValue()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method creates a 'raw' character value string from a character
//	    value string potentially containing character and control commands.
//	    It interprets all character and control commands to produce a
//	    string containing only 'raw' characters (i.e. no commands). The
//	    character and control commands have the following format:
//
//	    <PRE>
//	    ~<mnemonic>	    Mnemonic character, e.g. ~<CR> carriage return.
//	    ~<xx>	    Hexadecimal ASCII code, e.g. ~<0D> carriage return
//	    ~~		    Literal command character ("~")
//	    ~^		    Literal control character ("^")
//	    ^m		    Control command, e.g. ^M carriage return.
//	    </PRE>
//
//	    where "~" is the command character and "^" is the control
//	    character. e.g "Fred~<CR>".
//
//	    A ParseException is thrown if an error is encountered, e.g.
//	    an invalid command.
//
// @rdesc   'Raw' character string (with all character and control commands
//	    translated).
// @parm    Input string containing character value - potentially including
//	    character and control commands - but excluding delimiting quotes.
// @parm    Position in input string at which value starts (first character
//	    of value).
// @parm    Position in input string at which value string ends (last
//	    character of value).
//
string CECTofTxt::_getChrValue(const string& inStr, const int valStartPos,
                               const int valEndPos)
{
    string outStr;  // Output string
    for (int inPos = valStartPos; inPos < valEndPos; )
    {
        if (inStr[inPos] == ps_cmnd_chr)	    // If character command
        {
            inPos++;	// Move past  command character

            if (inPos > valEndPos)
            {
                char offbuf[32];
                sprintf(offbuf, "%li", (long) inPos);
                throw ParseException("Invalid character command in: \"" + inStr +
                                     "\" at offset: " + string(offbuf) + " Premature end of string");
            }

            if (inStr[inPos] == ps_cmnd_chr)	    // If have "~~"
            {
                outStr += inStr[inPos++];	    // Command character literal
            }

            else if (inStr[inPos] == ps_ctrl_chr)   // If have "~^"
            {
                outStr += inStr[inPos++];	    // Control character literal
            }

            else if (inStr[inPos] == ps_stchr_cmndc) // If have "~<..."
            {
                // Move past opening delimiter and look for closing delimiter.
                inPos++;
                int newPos = inStr.find(ps_echr_cmndc, inPos);

                if ((newPos == inStr.npos)	// If closing delimiter not found
                        ||  (newPos > valEndPos))	// ...or beyond end of value!
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) inPos);
                    throw ParseException("Invalid character command in: \"" + inStr +
                                         "\" at offset: " + string(offbuf) +
                                         " No terminating delimiter for character command");
                }

                // Get characters "ccc" between delimiters ("~<ccc>") in uppercase
                // for case_blind comparison.
                string cmdStr = inStr.substr(inPos, newPos - inPos);
                string_toupper (cmdStr);

                // See if we have mnemonic format, e.g. ~<CR>.
                if (cmdStr == "BEL")
                {
                    outStr += '\a';    // Bell character
                }

                else if (cmdStr == "BS")
                {
                    outStr += '\b';    // Backspace character
                }

                else if (cmdStr == "CR")
                {
                    outStr += '\r';    // Carriage return character
                }

                else if (cmdStr == "DEL")
                {
                    outStr += '\x7F';  // Delete character
                }

                else if (cmdStr == "ESC")
                {
                    outStr += '\x1B'; // Escape character
                }

                else if (cmdStr == "FF")
                {
                    outStr += '\f';    // Form-feed character
                }

                else if (cmdStr == "HT")
                {
                    outStr += '\t';    // Horizontal tab character
                }

                else if (cmdStr == "LF")
                {
                    outStr += '\n';    // Line feed character
                }

                else if (cmdStr == "SP")
                {
                    outStr += ' ';	// Space character
                }

                else if (cmdStr == "VT")
                {
                    outStr += '\v';	// Vertical tab character
                }

                else // Not mnemonic, may be hexadecimal code
                {
                    char* endP;
                    long hexval = strtol (inStr.c_str() + inPos, &endP, 16);
                    if (endP >= inStr.c_str() + newPos)
                    {
                        char c = hexval;
                        outStr += c;
                    }
                    else // Else not a valid command
                    {
                        char offbuf[32];
                        sprintf(offbuf, "%li", (long) inPos);
                        throw ParseException("Invalid character command in: \"" + inStr +
                                             "\" at offset: " + string(offbuf) +
                                             " Unrecognised mnemonic/hexadecimal code");
                    }
                }
                inPos = newPos + 1;	// Move past character command
            }

            else // Else not a valid command
            {
                char offbuf[32];
                sprintf(offbuf, "%li", (long) inPos);
                throw ParseException("Invalid character command in: \"" + inStr +
                                     "\" at offset: " + string(offbuf) + " Invalid character after command character");
            }
        }
        else if (inStr[inPos] == ps_ctrl_chr)	    // Else if control command
        {
            inPos++;				    // Move to control specifier
            outStr += (inStr[inPos] & (~64));	    // Mask off bit 7 to get character
            inPos++;				    // Move to next character in input
        }

        else					    // Else must be a literal character
        {
            outStr += inStr[inPos];		    // Copy input character unchanged
            inPos++;				    // Move to next character in input
        }

    } // End of for...

    return outStr;
} // End _getChrValue()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for a character variable at the specified
//	    position within the input string. It optionally updates the string
//	    position to point to the character after the character variable.
//
// @rdesc   Return status. 'true' if character variable reference at specified
//	    position in input string, 'false' otherwise.
// @parm    String to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the character variable.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string..
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psChrVariable(const string& inStr, int* pos, const bool updpos)
{
    return _psVariable (inStr, pos, VTCHAR);
} // End _psChrVariable()


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses an ~EXTRACT function reference. This has the
//	    following format:
//
//	    <PRE>
//	    ~EXTRACT (offset, length, string)
//	    </PRE>
//
//	    where 'offset' and 'length' may be an integer value or an integer
//	    variable and 'string' is a character value or character variable.
//
//	    The function produces an operand list in postfix (Polish) format.
//	    This has the following format:
//
//	    <PRE>
//	    OPD: ~EXTRACT FUNCTION
//	    OPD: offset
//	    OPD: NULL operand (terminating offset)
//	    OPD: length
//	    OPD: NULL operand (terminating length)
//	    OPD: string
//	    OPD: NULL operand (terminating string)
//	    </PRE>
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing ~EXTRACT function reference.
// @parm    Position in input string at which to start parsing.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position
//
bool CECTofTxt::_psFncExtract (const string& inStr, int* pos, const bool updpos)
{
    enum States {extract_1, extract_2, extract_3, extract_4,
                 extract_5, extract_6, extract_7, extract_8};

    int	wrkpos = *pos;	// Get copy of string position for local update
    long locopt = 0;	// ~EXTRACT options mask

    // Loop through the tokens in the input string encoding them as necessary.
    States state = extract_1;
    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE (inStr, wrkpos);

        switch (state)
        {
            // Start ~EXTRACT reference. Expect "~EXTRACT".
        case extract_1:
            {
                if (wrkpos >= inStr.length())   // If EOS, NOT a ~EXTRACT reference
                {
                    return false;
                }
                else if (inStr.substr(wrkpos, ps_fnc_extract.length()) == ps_fnc_extract)
                {
                    wrkpos += ps_fnc_extract.length();
                    state = extract_2;
                }
                else			    // Else NOT a ~EXTRACT reference
                {
                    return false;
                }
                break;
            }
            // Have parsed "~EXTRACT", now expect "(".
        case extract_2:
            {
                _encFnc(fnc_extract);	    // Encode ~EXTRACT function operand

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered; Expecting \"(\".");
                }
                else if (inStr.substr(wrkpos, ps_stprm.length()) == ps_stprm)
                {
                    wrkpos += ps_stprm.length();
                    state = extract_3;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; \"(\" not found.");
                }
                break;
            }
            // Have parsed "(", now expect 'offset' operand (P1).
        case extract_3:
            {
                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered " +
                                         "; Expecting 'offset' operand (P1).");
                }
                else if (_psIntValue(inStr, &wrkpos))
                {
                    state = extract_4;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    state = extract_4;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Invalid 'offset' operand (P1).");
                }
                break;
            }
            // Have parsed 'offset' operand (P1), now expect ",".
        case extract_4:
            {
                CECOperand opd;		    // Encode terminating NULL operand
                _exprEncoded.push_back(opd);    // ...

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting parameter separator.");
                }
                else if (inStr.substr(wrkpos, ps_prmsep.length()) == ps_prmsep)
                {
                    wrkpos += ps_prmsep.length();
                    state = extract_5;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Parameter separator not found.");
                }
                break;
            }
            // Have parsed ",", now expect 'length' operand (P2).
        case extract_5:
            {
                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered " +
                                         "; Expecting 'length' operand (P2).");
                }
                else if (_psIntValue(inStr, &wrkpos))
                {
                    state = extract_6;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    state = extract_6;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Invalid 'length' operand (P2).");
                }
                break;
            }
            // Have parsed 'length' operand (P2), now expect ",".
        case extract_6:
            {
                CECOperand opd;		    // Encode terminating NULL operand
                _exprEncoded.push_back(opd);    // ...

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting parameter separator.");
                }
                else if (inStr.substr(wrkpos, ps_prmsep.length()) == ps_prmsep)
                {
                    wrkpos += ps_prmsep.length();
                    state = extract_7;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Parameter separator not found.");
                }
                break;
            }
            // Have parsed ",", now expect 'string' operand (P3).
        case extract_7:
            {
                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered " +
                                         "; Expecting 'string' operand (P3).");
                }
                else if (_psChrValue(inStr, &wrkpos))
                {
                    state = extract_8;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    state = extract_8;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Invalid 'string' operand (P3).");
                }
                break;
            }
            // Have parsed 'string' operand (P3), now expect ")".
        case extract_8:
            {
                CECOperand opd;		    // Encode terminating NULL operand
                _exprEncoded.push_back(opd);    // ...

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting \")\".");
                }
                else if (inStr.substr(wrkpos, ps_eprm.length()) == ps_eprm)
                {
                    wrkpos += ps_eprm.length();
                    goto FncExtract_parsed;
                }
                else // Not a valid ~EXTRACT reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~EXTRACT function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; \")\" not found.");
                }
                break;
            }
        default:
            {
                char offbuf[32];
                sprintf(offbuf, "%li", (long) wrkpos);
                throw ParseException(
                    "Internal error - unrecognised state in _psFncExtract parsing: \"" +
                    inStr + "\" at offset: " + string(offbuf) + ".");
            }
        } // End switch (state)
    } // End while (pos < inStr.length())

    // ~EXTRACT function reference parsed successfully.

FncExtract_parsed:
    ;

    // Encode terminating NULL operand.
    CECOperand nulopd;
    _exprEncoded.push_back(nulopd);

    // Update string position (if required) and return success.
    if (updpos) *pos = wrkpos;

    return true;
} // End _psFncExtract()



///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a ~LOCATE function reference. This has the
//	    following format:
//
//	    <PRE>
//	    ~LOCATE (substring, string) {,CASE_BLIND}
//	    </PRE>
//
//	    where 'substring' and 'string' may be a character value or a
//	    character variable.
//
//	    The function produces an operand list in postfix (Polish) format.
//	    This has the following format:
//
//	    <PRE>
//	    OPD: ~LOCATE FUNCTION
//	    OPD: substring
//	    OPD: NULL operand (terminating substring)
//	    OPD: string
//	    OPD: NULL operand (terminating string)
//	    OPD: Function options mask
//	    OPD: NULL operand (terminating options mask)
//	    </PRE>
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing ~LOCATE function reference.
// @parm    Position in input string at which to start parsing.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position
//
bool CECTofTxt::_psFncLocate (const string& inStr, int* pos, const bool updpos)
{
    enum States {locate_1, locate_2, locate_3, locate_4,
                 locate_5, locate_6, locate_7, locate_8};

    int	wrkpos = *pos;	// Get copy of string position for local update
    long locopt = 0;	// ~LOCATE options mask

    // Loop through the tokens in the input string encoding them as necessary.
    States state = locate_1;
    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE (inStr, wrkpos);

        switch (state)
        {
            // Start ~LOCATE reference. Expect "~LOCATE".
        case locate_1:
            {
                if (wrkpos >= inStr.length())   // If EOS, NOT a ~LOCATE reference
                {
                    return false;
                }
                else if (inStr.substr(wrkpos, ps_fnc_locate.length()) == ps_fnc_locate)
                {
                    wrkpos += ps_fnc_locate.length();
                    state = locate_2;
                }
                else			    // Else NOT a ~LOCATE reference
                {
                    return false;
                }
                break;
            }
            // Have parsed "~LOCATE", now expect "(".
        case locate_2:
            {
                _encFnc(fnc_locate);	    // Encode ~LOCATE function operand

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered; Expecting \"(\".");
                }
                else if (inStr.substr(wrkpos, ps_stprm.length()) == ps_stprm)
                {
                    wrkpos += ps_stprm.length();
                    state = locate_3;
                }
                else // Not a valid ~LOCATE reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; \"(\" not found.");
                }
                break;
            }
            // Have parsed "(", now expect 'substring' operand (P1).
        case locate_3:
            {
                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered " +
                                         "; Expecting 'substring' operand (P1).");
                }
                else if (_psChrValue(inStr, &wrkpos))
                {
                    state = locate_4;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    state = locate_4;
                }
                else // Not a valid ~LOCATE reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Invalid 'substring' operand (P1).");
                }
                break;
            }
            // Have parsed 'substring' operand (P1), now expect ",".
        case locate_4:
            {
                CECOperand opd;		    // Encode terminating NULL operand
                _exprEncoded.push_back(opd);    // ...

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting parameter separator.");
                }
                else if (inStr.substr(wrkpos, ps_prmsep.length()) == ps_prmsep)
                {
                    wrkpos += ps_prmsep.length();
                    state = locate_5;
                }
                else // Not a valid ~LOCATE reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Parameter separator not found.");
                }
                break;
            }
            // Have parsed ",", now expect 'string' operand (P2).
        case locate_5:
            {
                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered " +
                                         "; Expecting 'string' operand (P2).");
                }
                else if (_psChrValue(inStr, &wrkpos))
                {
                    state = locate_6;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    state = locate_6;
                }
                else // Not a valid ~LOCATE reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; Invalid 'string' operand (P2).");
                }
                break;
            }
            // Have parsed 'string' operand (P2), now expect ")".
        case locate_6:
            {
                CECOperand opd;		    // Encode terminating NULL operand
                _exprEncoded.push_back(opd);    // ...

                if (wrkpos >= inStr.length())   // If EOS, Premature EOL
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting \")\".");
                }
                else if (inStr.substr(wrkpos, ps_eprm.length()) == ps_eprm)
                {
                    wrkpos += ps_eprm.length();
                    state = locate_7;
                }
                else // Not a valid ~LOCATE reference
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; \")\" not found.");
                }
                break;
            }
            // Have parsed ")", now expect EOS, ",CASE_BLIND" or something else.
        case locate_7:
            {
                if (wrkpos >= inStr.length())
                {
                    ;   // End of string OK, do nothing.
                }

                else if (inStr.substr(wrkpos, ps_stopt.length()) == ps_stopt)
                {
                    wrkpos += ps_stopt.length();
                    state = locate_8;
                }
                else // We have a valid ~LOCATE with trailing characters.
                {
                    goto FncLocate_parsed;
                }
                break;
            }
            // Have parsed ",", now expect "CASE_BLIND".
        case locate_8:
            {
                if (wrkpos >= inStr.length())
                {
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\"; Unexpected end of line encountered" +
                                         "; Expecting \"CASE_BLIND\".");
                }
                else if (inStr.substr(wrkpos, ps_key_case_blind.length()) == ps_key_case_blind)
                {
                    wrkpos += ps_key_case_blind.length();
                    locopt |= FNCOPT_CASE_BLIND;		// Set Case Blind option
                    goto FncLocate_parsed;
                }
                else
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    throw ParseException("Invalid ~LOCATE function reference: \"" +
                                         inStr + "\" at offset: " + string(offbuf) +
                                         "; \"CASE_BLIND\" not found.");
                }
                break;
            }
        default:
            {
                char offbuf[32];
                sprintf(offbuf, "%li", (long) wrkpos);
                throw ParseException(
                    "Internal error - unrecognised state in _psFncLocate parsing: \"" +
                    inStr + "\" at offset: " + string(offbuf) + ".");
            }
        } // End switch (state)
    } // End while (pos < inStr.length())

    // ~LOCATE function reference parsed successfully.

FncLocate_parsed:
    ;

    // Encode ~LOCATE options mask operand and terminating NULL operand.
    CECOperand opd(locopt);
    _exprEncoded.push_back(opd);

    CECOperand nulopd;		    // Encode terminating NULL operand
    _exprEncoded.push_back(nulopd); // ...

    // Update string position (if required) and return success.
    if (updpos) *pos = wrkpos;

    return true;
} // End _psFncLocate()


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for an integer expression string at the
//	    specified position within the input string. It optionally updates
//	    the string position to point to the character after the expression.
//	    An integer expression string has the following format:
//
//	    <PRE>
//	    operand1 {operator operandn}
//	    </PRE>
//
//	    Any number of operator/operands may appear in an expression
//	    and parentheses may be used to aid legibility and/or change the
//	    evaluation order - by default evaluation is left to right.
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetects that the
//	    string being parsed is not an integer expression a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
//	    Note that this routine does NOT ensure that the FULL string from
//	    the specified position to the end of the string is a valid integer
//	    expression. There may be extraneous characters following the
//	    integer expression. This allows this routine to be used to parse
//	    integer expressions in SET commands and array indices.
//
// @rdesc   Return status. 'true' if integer expression at specified position
//	    in input string, 'false' otherwise.
// @parm    String containing integer expression.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the integer expression.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psIntExpr(const string& inStr, int* pos, const bool updpos)
{
    int	wrkpos = *pos;	    // Get copy of string position for local update

    enum States {int_expr_1, int_expr_2, int_expr_3, int_expr_4};

    //
    //  Save starting stack size, so that we can check later whether the expression is complete.
    //
    stkOperator::size_type start_stksiz = _prsStack.size();  // Save starting stack size

    //
    //  Loop through the input string tokens from the specified starting position
    //  and produce a corresponding postfix format operand list.
    //
    States state = int_expr_1;  // Set initial parsing state

    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE(inStr, wrkpos);

        // Process current parse state.
        switch (state)
        {
            // Start an integer expression. Expect "(" or operand.
        case int_expr_1:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    // No state change
                }

                else if (_psIntValue(inStr, &wrkpos))
                {
                    state = int_expr_2;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    state = int_expr_2;
                }
                else if (_psIntFnc(inStr, &wrkpos))
                {
                    state = int_expr_2;
                }
                else
                {
                    return false;		    // Not an integer expression
                }

                break;
            }
            // Here after operand. Expect EOS or integer operator.
        case int_expr_2:
            {
                if (wrkpos >= inStr.length())
                {
                    ;   // End of string OK, do nothing.
                }

                else
                {
                    int isub;   // Loop control
                    for (isub = 0; isub < intops_end; isub++)
                    {
                        if (inStr.substr(wrkpos, intops[isub].length()) == intops[isub])
                        {
                            wrkpos += intops[isub].length();
                            _saveOpr(intopr[isub]);
                            state = int_expr_3;
                            break;
                        }
                    }
                    // If operator not found, we have a valid expression with trailing characters.
                    if (isub >= intops_end) goto IntExpr_parsed;
                }
                break;
            }
            // Here after operator. Expect "(" or operand.
        case int_expr_3:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    // No state change
                }

                else if (_psIntValue(inStr, &wrkpos))
                {
                    _encOpr();
                    state = int_expr_4;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    _encOpr();
                    state = int_expr_4;
                }
                else if (_psIntFnc(inStr, &wrkpos))
                {
                    _encOpr();
                    state = int_expr_4;
                }
                else
                {
                    throw ParseException("Operator must be followed by operand or '(', got \"" +
                                         _expandToBlank (inStr, wrkpos) + "\"");
                }
                break;
            }
            // Here after operand following an operator. Expect EOS, ")" or another operator.
        case int_expr_4:
            {
                if (wrkpos >= inStr.length())
                {
                    ;   // End of string valid, do nothing
                }

                else if (inStr.substr(wrkpos, ps_econd.length()) == ps_econd)
                {
                    wrkpos += ps_econd.length();
                    _decExprLevel();
                    // No state change
                }

                else
                {
                    state = int_expr_2;
                }
                break;
            }
        default:
            {
                throw ParseException("Internal error - unrecognised state");
            }
        } // End switch (state)
    } // End while (wrkpos < inStr.length())


IntExpr_parsed:
    ;
    // If operator stack has not been fully unwound the expression is invalid.
    if (_prsStack.size() != start_stksiz)
    {
        throw ParseException("Invalid expression structure");
    }

    if (updpos) *pos = wrkpos;

    return true;
} // End TtIntExpr::_psIntExpr()


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a predefined integer function reference.
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing integer function reference.
// @parm    Position in input string at which to start parsing.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position
//
bool CECTofTxt::_psIntFnc (const string& inStr, int* pos, const bool updpos)
{
    // Check for each of the predefined integer functions.
    // Currently only support ~LOCATE.
    if (_psFncLocate (inStr, pos, updpos)) return true;

    // Not an integer function.
    return false;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for an integer value at the specified
//	    position within the input string. It optionally updates the string
//	    position to point to the character after the character value. An
//	    integer value has the following format:
//
//	    <PRE>
//	    {+/-}n
//	    </PRE>
//
//	    where 'n' is an integer number. For compatibility with SCL,
//	    whitespace is permitted between the sign and digits.
//
// @rdesc   Return status. 'true' if integer value at specified position
//	    in input string, 'false' otherwise.
// @parm    String to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the integer value.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psIntValue(const string& inStr, int* pos, const bool updpos)
{
    int	wrkpos = *pos;	    // Get copy of string position for local update

    // Look for and save sign, if specified.
    string signstr = "";    // Sign string.
    if (inStr[wrkpos] == '-')
    {
        wrkpos++;	    // Position after sign
        signstr = "-";	    // Save sign
    }

    else if (inStr[wrkpos] == '+')
    {
        wrkpos++;	    // Position after sign
    }

    // If only whitespace left, not an integer value, return with failure status.

    SKIP_WHITESPACE (inStr, wrkpos);
    if (wrkpos >= inStr.length()) return false;

    // Get integer value and position of first character after value.
    long intval;	    // Integer value
    string::size_type newpos = inStr.find_first_not_of("0123456789", wrkpos);

    // If not an integer value, return with failure status.
    if(newpos == wrkpos)
    {
        return false;
    }

    // Else if integer value extends to end of string.
    else if (newpos == inStr.npos)
    {
        string valstr = signstr + inStr.substr(wrkpos, inStr.length());
        intval = atol(valstr.c_str());
        wrkpos = inStr.length();
    }

    // Else integer value does not extend to end of string.
    else
    {
        string valstr = signstr + inStr.substr(wrkpos, newpos);
        intval = atol(valstr.c_str());
        wrkpos = newpos;
    }

    // Encode operand and update expression descriptor data. We do an 'insert'
    // rather than a push_back to retrieve the iterator for the new item.
    CECOperand opd(intval);
    list<CECOperand>::const_iterator it = _exprEncoded.insert(_exprEncoded.end(), opd);

    if (_idxLevel == 0)				// If not processing an array subscript
    {
        _isVariable = false;			// NOT a variable operand
        _varNam = "";				// ...
        _dTyp = opd.DTyp();			// Expression type confirmed
        if (_opdCnt == 0) _exprIt = it;		// Save iterator of first operand
        _opdCnt += 1;				// One more main expression operand
    }

    // Update string position (if required) and return success.

    if (updpos) *pos = wrkpos;
    return true;
} // End _psIntValue()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function checks for an integer variable reference at the
//	    specified position within the input string. It optionally updates
//	    the string position to point to the character after the integer
//	    variable.
//
// @rdesc   Return status. 'true' if integer variable reference at specified
//	    position in input string, 'false' otherwise.
// @parm    String to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the integer variable.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string..
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psIntVariable(const string& inStr, int* pos, const bool updpos)
{
    return _psVariable (inStr, pos, VTINT);
} // End _psIntVariable()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a simple boolean condition string. A simple
//	    boolean expression string has the following format:
//
//	    <PRE>
//	    operand1 operator operand2
//	    </PRE>
//
//	    where 'operand1' is an integer or character operand, 'operator' is
//	    a logical operator that may be used on the specified operand types
//	    and 'operand2' is an operand of the same type as 'operand1'.
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing boolean expression.
// @parm    Position in input string at which to start parsing.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psSimpleCondition (const string& inStr, int* pos, const bool updpos)
{
    enum States {simple_condition_1, simple_condition_2, simple_condition_3,
                 simple_condition_4, simple_condition_5};

    int	wrkpos = *pos; // Get copy of string position for local update

    //
    //  Loop through the tokens in the input string encoding them as necessary.
    //
    States state = simple_condition_1;
    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE (inStr, wrkpos);

        switch (state)
        {
            // Parse 'operand1'.
        case simple_condition_1:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (_psIntValue(inStr, &wrkpos))
                {
                    state = simple_condition_2;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    state = simple_condition_2;
                }
                else if (_psChrValue(inStr, &wrkpos))
                {
                    state = simple_condition_3;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    state = simple_condition_3;
                }
                else
                {
                    _reason = "Unrecognised operand \"" +
                              _expandToBlank (inStr, wrkpos) +
                              "\"";
                    return false;
                }
                break;
            }
            // Have parsed integer operand, now need a logical operator for such operands.
        case simple_condition_2:
            {
                if (wrkpos >= inStr.length()) ERROR_PREMEOS;

                int isub;	// Loop control
                for (isub = 0; isub < logintops_end; isub++)
                {
                    if (inStr.substr(wrkpos, logintops[isub].length()) == logintops[isub])
                    {
                        wrkpos += logintops[isub].length();
                        _saveOpr(logintopr[isub]);
                        state = simple_condition_4;
                        break;
                    }
                }
                // If operator not found report an error.
                if (isub >= logintops_end)
                {
                    throw ParseException("Invalid operator for integer operands");
                }
                break;
            }
            // Have parsed character operand, now need a logical operator for such operands.
        case simple_condition_3:
            {
                if (wrkpos >= inStr.length()) ERROR_PREMEOS;

                int isub;	// Loop control
                for (isub = 0; isub < logchrops_end; isub++)
                {
                    if (inStr.substr(wrkpos, logchrops[isub].length()) == logchrops[isub])
                    {
                        wrkpos += logchrops[isub].length();
                        _saveOpr(logchropr[isub]);
                        state = simple_condition_5;
                        break;
                    }
                }
                // If operator not found report an error.
                if (isub >= logchrops_end)
                {
                    throw ParseException("Invalid operator for character operands");
                }
                break;
            }
            // Parse integer 'operand2'.
        case simple_condition_4:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (_psIntValue(inStr, &wrkpos))
                {
                    if (updpos) *pos = wrkpos;
                    return true;
                }
                else if (_psIntVariable(inStr, &wrkpos))
                {
                    if (updpos) *pos = wrkpos;
                    return true;
                }
                else
                {
                    throw ParseException("Invalid integer operand");
                }
                break;
            }
            // Parse character 'operand2'.
        case simple_condition_5:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (_psChrValue(inStr, &wrkpos))
                {
                    if (updpos) *pos = wrkpos;
                    return true;
                }
                else if (_psChrVariable(inStr, &wrkpos))
                {
                    if (updpos) *pos = wrkpos;
                    return true;
                }
                else
                {
                    throw ParseException("Invalid character operand");
                }
                break;
            }
        default:
            {
                throw ParseException("Internal error - unrecognised state");
            }
        } // End switch (state)
    } // End while (pos < inStr.length())


    ERROR_PREMEOS;
} // End _psSimpleCondition()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a value list string. A value list string has
//	    the following format:
//
//	    <PRE>
//	    value1 {, valuen}
//	    </PRE>
//
//	    where "value1" and "valuen" are integer or character expressions.
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing value list.
// @parm    Position in input string at which to start parsing.
// @parm    Number of values parsed and encoded.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string.
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psValList (const string& inStr, int* pos, long* valcnt,
                            const bool updpos)
{
    enum States {val_list_1, val_list_2};

    int	wrkpos = *pos;	// Get copy of string position for local update
    int	wrkvalcnt = 0;	// Local value count - only update parameter if successful

    //
    //  Loop through the tokens in the input string encoding them as necessary.
    //
    States state = val_list_1;
    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE (inStr, wrkpos);

        switch (state)
        {
            // Start value list. Expect value (integer or character expression) - terminated by
            // "," or EOS or something else.
        case val_list_1:
            {
                if (wrkpos >= inStr.length())	    // If EOS
                {
                    ERROR_PREMEOS;
                }
                else if (_psIntExpr (inStr, &wrkpos))   // Else If integer value
                {
                    wrkvalcnt++;			    // Increment value count
                    CECOperand opd;			    // Encode terminating NULL operand
                    _exprEncoded.push_back(opd);	    // ...
                    CECExprDsc exprDsc (_dTyp, _isVariable, _varNam, _exprIt);
                    _exprList.push_back(exprDsc);
                    _opdCnt = 0;
                    state = val_list_2;
                }
                else if (_psChrExpr (inStr, &wrkpos))   // Else If character value
                {
                    wrkvalcnt++;			    // Increment value count
                    CECOperand opd;			    // Encode terminating NULL operand
                    _exprEncoded.push_back(opd);	    // ...
                    CECExprDsc exprDsc (_dTyp, _isVariable, _varNam, _exprIt);
                    _exprList.push_back(exprDsc);
                    _opdCnt = 0;
                    state = val_list_2;
                }
                else				    // Else NOT a valid value
                {
                    char offbuf[32];
                    sprintf(offbuf, "%li", (long) wrkpos);
                    string msgstr = "Invalid value in value list: \"" + inStr +
                                    "\" at offset: " + string(offbuf);
                    throw mdbParseException(msgstr);
                }
                break;
            }
            // Have parsed value, now expect "," or EOS or something else.
        case val_list_2:
            {
                if (wrkpos >= inStr.length()) // If EOS
                {
                    ; // End of Stream is valid, do nothing
                }

                else if (inStr.substr(wrkpos, ps_valsep.length()) == ps_valsep)
                {
                    wrkpos += ps_valsep.length();
                    state = val_list_1;
                }
                else // Else valid value list with trailing characters.
                {
                    goto vallist_done;
                }
                break;
            }
        default:
            {
                char offbuf[32];
                sprintf(offbuf, "%li", (long) wrkpos);
                throw mdbParseException(				//RAC:04/07/00
                    "Internal error - unrecognised state in _psValList parsing: \"" +
                    inStr + "\" at offset: " + string(offbuf) + ".");
            }
        } // End switch (state)
    } // End while (pos < inStr.length())

    // Value List parsed successfully, update string position (if required)
    // and value count.

vallist_done:
    ;
    if (updpos) *pos = wrkpos;
    *valcnt = wrkvalcnt;
    return true;
} // End _psValList()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method checks for a character or integer variable at the
//	    specified position within the input string. It optionally updates
//	    the string position to point to the character after the variable.
//
// @rdesc   Return status. 'true' if character variable reference at specified
//	    position in input string, 'false' otherwise.
// @parm    String to be parsed.
// @parm    Pointer to position in input string at which to start parsing.
//	    Optionally updated if parsing successful, to point to the character
//	    after the character variable.
// @parm    Options mask indicating the required variable options.
// @parm    Flag indicating whether the supplied string position is to be
//	    updated to point to the next character in the string..
//	    true = update position, false = do not update position.
//
bool CECTofTxt::_psVariable(const string& inStr, int* pos, const long reqopt,
                            const bool updpos)
{
    int	wrkpos = *pos; // Get copy of string position for local update

    // See if there is a dataname at the specified string position; if not, cannot be
    // a variable reference - return 'false'.
    short namlen = inStr.length();
    short status;
    //
    // LINUX Port Changes
    //
    // Pass in a const char as opposed to a const string.
    (void) CHK_DATANAME(inStr.c_str() + wrkpos, &namlen, &status);
    if (status == -1) return false;

    // Get potential variable name and position of first character after name.
    int tknlen = inStr.length() - status;
    string varnam = inStr.substr(wrkpos, tknlen);
    wrkpos += tknlen;

    // Check that the name is that of a defined variable; if not return false.
    // Also get variable options mask, so that we can check the variable type.
    // Also get variable dimensions count (if any), so that we can check that we have the
    //   correct number of indices in an array reference.
    long options;		// Variable options mask
    long dimensions = 0;	// Variable dimensions count (assume not an array)

    assert(_scriptPtr);

    if(_scriptPtr != NULL)	// If using Script object to obtain variable details
    {
        TRecord* trec_ptr = (*_scriptPtr).GetVardef(varnam);
        // If the name is NOT that of a defined variable log a message - I do not
        // think that it is ever valid to have a dataname that is NOT a defined
        // variable here, but just in case there is - or may be in the future - we log
        // a message rather than throw an exception.
        if (trec_ptr == NULL)
        {
            char offbuf[32];
            sprintf(offbuf, "%li", (long) wrkpos);
            string msgstr = "Warning parsing variable in: \"" + inStr +
                            "\" at offset: " + string(offbuf) +
                            "; Possible variable \"" + varnam + "\" is NOT a defined variable.";
            if(logr) logr->AddTrace(DBGSTD, DEBUGLEVEL, msgstr.c_str());
            return false;	// Not a defined variable
        }

        // Get options mask - should always be present, so throw an exception if it is not.

        const TParam* prm_ptr = trec_ptr->param(TParam::stOptions);
        if (prm_ptr == NULL)
        {
            char offbuf[32];
            sprintf(offbuf, "%li", (long) wrkpos);
            string msgstr = "Error parsing variable in: \"" + inStr +
                            "\" at offset: " + string(offbuf) +
                            "; 'options' parameter missing from variable \"" +
                            varnam + "\" definition.";
            throw mdbParseException(msgstr);
        }

        options = prm_ptr->intVal();		// Get variable options mask

        const TParam* dimprm_ptr = trec_ptr->param(TParam::stDimensions);
        if (dimprm_ptr != NULL)
        {
            string dimstr = dimprm_ptr->strVal();
            dimensions = _dimCnt (dimstr);
        }
    }
    else // Else not a valid script object pointer - should never get here!
    {
        char offbuf[32];
        sprintf(offbuf, "%li", (long) wrkpos);
        string msgstr = "Error parsing variable in: \"" + inStr +
                        "\" at offset: " + string(offbuf) +
                        "; NULL script object pointer parsing variable \"" +
                        varnam + "\".";
        throw mdbParseException(msgstr);
    }

    // Return false, if not a character variable.
    if (!(options & reqopt)) return false;

    // Variable IS a variable with the required options - encode variable operand.
    // and update expression descriptor data. We do an 'insert' rather than a
    // push_back to retrieve the iterator for the new item. If this is the first
    // operand in an expression (excluding indices) we may have a variable expression
    // (set _isVariable), otherwise we do not have a variable expression.
    enum DataType dtyp = dtyp_none;
    if (options & VTCHAR) dtyp = dtyp_character;
    if (options & VTINT) dtyp = dtyp_integer;

    CECOperand opd(varnam, dtyp, dimensions);
    list<CECOperand>::const_iterator it = _exprEncoded.insert(_exprEncoded.end(), opd);

    if (_idxLevel == 0)				// If not processing an array subscript
    {
        if (_opdCnt == 0)			// If first operand, may be a variable,
        {
            _isVariable = true;			// May be a variable operand
            _varNam = opd.VarNam();		// Variable name
        }

        else					// Else, not a variable
        {
            _isVariable = false;		// IS NOT a variable operand
            _varNam = "";			// ...
        }

        _dTyp = opd.DTyp();			// Expression type confirmed
        if (_opdCnt == 0) _exprIt = it;		// Save iterator of first operand
        _opdCnt += 1;				// One more main expression operand
    }

    // Skip any white space following variable name and process any expected indices.

    SKIP_WHITESPACE (inStr, wrkpos);
    if (dimensions > 0)
    {
        int idxcnt;	// Number of array indices
        if (_psArrayIndices (inStr, &wrkpos, &idxcnt))
        {
            if (dimensions != idxcnt)
            {
                throw ParseException("Number of subscripts in reference, do not match"
                                     "number in declaration.");
            }
        }
        else
        {
            throw ParseException("Invalid array indices:\"" + _expandToEos(inStr, wrkpos) +
                                 "\"");
        }
    }

    // Update string position (if required) and return success.
    if (updpos) *pos = wrkpos;
    return true;
} // End _psVariable()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Save operator on operator stack. If the operator is a 'NOT'
//	    operator just push it into the stack, otherwise replace the
//	    operator on the top of the stack (that for the current level) with
//	    the new one. If we are already at the top of the stack,
//	    the expression is invalid a ParseException is thrown.
//
// @parm    Operator to be saved.
//
void CECTofTxt::_saveOpr(const Operator opr)
{
    if (opr == opr_not)
    {
        _prsStack.push(opr);
    }
    else
    {
        if (_prsStack.empty())
        {
            throw ParseException("Cannot save operator on stack, stack empty");
        }
        _prsStack.pop();
        _prsStack.push(opr);
    }

    return;
} // End _saveOpr()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a boolean expression string. A boolean
//	    expression string has the following format:
//
//	    <PRE>
//	    ( {!}{(}simple_condition{)} {&/| {!}(simple_condition)} )
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//	    The ::Reason() method can be used to obtain a string describing
//	    the reason for any failure.
//
//	    The function delegates the real work to the protected function
//	    ::_parse, which will generate a ParseException if the supplied
//	    string is not a boolean expression. If a ParseException is
//	    thrown by ::_parse this routine sets the _reason field to
//	    indicate the reason for the failure and returns 'false',
//	    otherwise it returns 'true'.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing boolean expression.
// @parm    Position in input string at which to start parsing.
//
bool TtBoolExpr::Parse(const string& inStr, const int inPos)
{
    try
    {
        _parse(inStr, inPos);
    }
    catch (ParseException& exc)
    {
        _reason = exc.Reason();
        return false;
    }
    return true;
} // End TtBoolExpr::Parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a boolean expression string. A boolean
//	    expression string has the following format:
//
//	    <PRE>
//	    simple_condition  (no parentheses)  OR
//	    {!}(simple_condition) {&/| {!}(simple_condition)}
//	    </PRE>
//
//	    Additional parentheses may be added as desired. Examples:
//
//	    <PRE>
//	    ISUB < 10
//	    ((ISUB < 10) & ((JSUB > 1) | !(KSUB = -2)))
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//
// @parm    String containing boolean expression.
// @parm    Pointer to position in input string at which to start parsing.
//
void TtBoolExpr::_parse(const string& inStr, const int pos)
{
    int	wrkpos = pos; // Get copy of string position for local update

    enum States {bool_expr_1, bool_expr_2, bool_expr_3, bool_expr_4, bool_expr_5,
                 bool_expr_end};

    //
    //  Loop through the input string tokens from the specified starting position
    //  and produce a corresponding postfix format operand list.
    //
    _clrExpr();			// Clear expression parsing data
    States state = bool_expr_1;	// Set initial parsing state

    while (wrkpos < inStr.length())
    {
        SKIP_WHITESPACE(inStr, wrkpos);

        // Process current parse state.
        switch (state)
        {
            // Start boolean expresion. Expect "(", "!" or simple condition.
        case bool_expr_1:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    state = bool_expr_2;
                }
                else if (inStr.substr(wrkpos, ps_opr_not.length()) == ps_opr_not)
                {
                    wrkpos += ps_opr_not.length();
                    _saveOpr(opr_not);
                    state = bool_expr_3;
                }
                else if (_psSimpleCondition(inStr, &wrkpos))
                {
                    _encOpr();				// Encode operator
                    SKIP_WHITESPACE(inStr, wrkpos);		// Skip trailing whitespace
                    if (wrkpos >= inStr.length()) return;	// OK, if all done
                    state = bool_expr_end;			// Must have extra text in string
                }

                else
                {
                    throw ParseException("Invalid operand \"" +
                                         _expandToBlank (inStr, wrkpos) + "\" after \"(\"");
                }
                break;
            }
            // Start boolean expression (or part thereof). Expect "(", "!" or simple condition.
        case bool_expr_2:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    // No state change
                }

                else if (inStr.substr(wrkpos, ps_opr_not.length()) == ps_opr_not)
                {
                    wrkpos += ps_opr_not.length();
                    _saveOpr(opr_not);
                    state = bool_expr_3;
                }
                else if (_psSimpleCondition(inStr, &wrkpos))
                {
                    state = bool_expr_4;
                }
                else
                {
                    throw ParseException("Invalid operand \"" +
                                         _expandToBlank (inStr, wrkpos) + "\" after \"(\"");
                }
                break;
            }
            // Here after "!" (NOT operator). Expect "(".
        case bool_expr_3:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_stcond.length()) == ps_stcond)
                {
                    wrkpos += ps_stcond.length();
                    _incExprLevel();
                    state = bool_expr_2;
                }
                else
                {
                    throw ParseException("Expected \"(\" to start an expression, got " +
                                         _expandToBlank (inStr, wrkpos));
                }
                break;
            }
            // Here after simple condition. Expect ")".
        case bool_expr_4:
            {
                if (wrkpos >= inStr.length())
                {
                    ERROR_PREMEOS;
                }
                else if (inStr.substr(wrkpos, ps_econd.length()) == ps_econd)
                {
                    wrkpos += ps_econd.length();
                    _decExprLevel();
                    state = bool_expr_5;
                }
                else
                {
                    throw ParseException("Condition must be terminated by \")\", got \"" +
                                         _expandToBlank (inStr, wrkpos) + "\"");
                }
                break;
            }
            // Here after terminated simple condition. Expect "&", "|", or ")".
        case bool_expr_5:
            {
                if (wrkpos >= inStr.length())
                {
                    ; // End of Stream is valid, do nothing
                }

                else if (inStr.substr(wrkpos, ps_opr_and.length()) == ps_opr_and)
                {
                    wrkpos += ps_opr_and.length();
                    _saveOpr(opr_and);
                    state = bool_expr_2;
                }
                else if (inStr.substr(wrkpos, ps_opr_or.length()) == ps_opr_or)
                {
                    wrkpos += ps_opr_or.length();
                    _saveOpr(opr_or);
                    state = bool_expr_2;
                }
                else if (inStr.substr(wrkpos, ps_econd.length()) == ps_econd)
                {
                    wrkpos += ps_econd.length();
                    _decExprLevel();
                    // No state change
                }

                else
                {
                    throw ParseException("Expecting '&', '|', or ')'; got \"" +
                                         _expandToBlank(inStr, wrkpos) + "\"");
                }
                break;
            }

            // Here if extra text following valid expression.
        case bool_expr_end:
            {
                throw ParseException("Unexpected text after valid expression \"" +
                                     _expandToEos (inStr, wrkpos) + "\".");
            }
        default:
            {
                throw ParseException("Internal error - unrecognised state");
            }
        } // End switch (state)
    } // End while (wrkpos < inStr.length())

    //
    // LINUX Port Changes
    //

#if defined(LINUX)||defined (HPUX)
    // Display resultant postfix format operand list.
    for (list<CECOperand>::iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#else
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#endif
    {
        CECOperand opd = *it;
    }

    // Expression parsed successfully. Now perform any consistency checks that can only
    // be performed after the expression has been parsed.
    _chkExprComplete();

    return;
} // End TtBoolExpr::_parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a character expression string. A character
//	    expression string has the following format:
//
//	    <PRE>
//	    operand1{,option} {operator operand{,option}}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//	    The ::Reason() method can be used to obtain a string describing
//	    the reason for any failure.
//
//	    The function delegates the real work to the protected function
//	    ::_parse, which will generate a ParseException if the supplied
//	    string is not a character expression. If a ParseException is
//	    thrown by _chrExpr this routine sets the _reason field to
//	    indicate the reason for the failure and returns 'false',
//	    otherwise it returns 'true'.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing character expression.
// @parm    Position in input string at which to start parsing.
//
bool TtChrExpr::Parse(const string& inStr, const int inPos)
{
    try
    {
        _parse(inStr, inPos);
    }
    catch (ParseException& exc)
    {
        _reason = exc.Reason();
        return false;
    }
    return true;
} // End TtChrExpr::Parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a character expression string. A character
//	    expression string has the following format:
//
//	    <PRE>
//	    operand1 {operator operand}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetect that the
//	    string being parsed is not a character expression a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
// @parm    String containing string to parse.
// @parm    Pointer to position in input string at which to start parsing.
//
void TtChrExpr::_parse(const string& inStr, const int pos)
{
    int	wrkpos = pos; // Get copy of string position for local update

    //
    //  Call a helper function to do the real work. Note that we need to check that
    //  there are no extraneous characters following a valid character expression.
    //
    bool status;    // _psChrExpr status
    _clrExpr();	    // Clear expression parsing data.
    status = _psChrExpr (inStr, &wrkpos);

    if (status && (wrkpos != (inStr.length())))
    {
        throw ParseException(
            "Extra text following valid character expression: " +
            _expandToEos (inStr, wrkpos));
    }
    else if (!status)
    {
        throw ParseException(
            "Invalid Character Expression, \"" + _expandToEos (inStr, 0) + "\"");
    }
    //
    // LINUX Port Changes
    //
#if defined(LINUX)||defined (HPUX)
    // Display resultant postfix format operand list.
    for (list<CECOperand>::iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#else
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#endif
    {
        CECOperand opd = *it;
    }

    return;
} // End TtChrExpr::_parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses an integer expression string. An integer
//	    expression string has the following format:
//
//	    <PRE>
//	    operand1 {operator operand}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//	    The ::Reason() method can be used to obtain a string describing
//	    the reason for any failure.
//
//	    The function delegates the real work to the protected function
//	    ::_parse, which will generate a ParseException if the supplied
//	    string is not an integer expression. If a ParseException is
//	    thrown by a called routine this routine sets the _reason field to
//	    indicate the reason for the failure and returns 'false',
//	    otherwise it returns 'true'.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing integer expression.
// @parm    Position in input string at which to start parsing.
//
bool TtIntExpr::Parse(const string& inStr, const int inPos)
{
    try
    {
        _parse(inStr, inPos);
    }
    catch (ParseException& exc)
    {
        _reason = exc.Reason();
        return false;
    }
    catch(...)//TODO remove
    {
        cout << "ERROR in TtIntExpr::Parse" << endl;
        throw;
    }
    return true;
} // End TtIntExpr::Parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses an integer expression string. An integer
//	    expression string has the following format:
//
//	    <PRE>
//	    operand1 {operator operand}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetect that the
//	    string being parsed is not an integer expression a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
// @parm    String containing integer expression.
// @parm    Pointer to position in input string at which to start parsing.
//
void TtIntExpr::_parse(const string& inStr, const int pos)
{
    int	wrkpos = pos; // Get copy of string position for local update

    //
    //  Call a helper function to do the real work. Note that we need to check that
    //  there are no extraneous characters following a valid integer expression.
    //
    bool status;    // _psIntExpr status
    _clrExpr();	    // Clear expression parsing data.
    status = _psIntExpr (inStr, &wrkpos);
    if (status && (wrkpos != (inStr.length())))
    {
        throw ParseException(
            "Extra text following valid integer expression: " +
            _expandToEos (inStr, wrkpos));
    }
    else if (!status)
    {
        throw ParseException(
            "Invalid Integer Expression, \"" + _expandToEos (inStr, 0) + "\"");
    }
/*
    //
    // LINUX Port Changes
    //
#if defined(LINUX)||defined (HPUX)
    // Display resultant postfix format operand list.
    for (list<CECOperand>::iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#else
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#endif
    {
        CECOperand opd = *it;
    }
*/
    return;
} // End TtIntExpr::_parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function evaluates a value in the encoded value list string
//	    set up by the TtValList::Parse() function. The encoded string
//	    consists of a list of operands and operators in reverse Polish
//	    format describing a list of values - with a NULL operand separating
//	    the values, e.g.
//
//	    <PRE>
//	    Input string:      cuscod, 100 + isub, cusnam
//	    Encoded string:    cuscod, NULL, 00, isub, +, NULL, cusnam
//	    Return (example):  "123","199","Jones" (list of strings)
//	    </PRE>
//
//	    This function returns the string resulting from the evaluation of
//	    the value in the value list identified by the supplied iterator.
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
//
// @parm    Reference to current context.
// @parm    reference to iterator identifying the first operand of the value
//	    to be evaluated in _exprEncoded.
// @parm    Reference to string to hold returned value.
//
bool TtValList::EvaluateValue(TContext& ctxt, list<CECOperand>::const_iterator& valItr,
                              string& retVal) const
{
    if (_exprEncoded.empty())
    {
        //_reason = "No values to evaluate";
        return false;
    }

    // Evaluate value identified by supplied value iterator.
    bool status = _evaluateValue(ctxt, valItr, retVal);
    if (!status) return false;

    return true;
} // End TtValList::EvaluateValue()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to evaluate a value in an encoded
//	    expression (a value being either a character or integer
//	    expression). The routine returns the evaluated value as a string.
//	    On return, the encoded expression iterator (used to obtain operands)
//	    passed to this routine will be updated, as necessary, so that it
//	    always refers to the last operand procesed by this routine.
//
//	    The following illustrates the format of the operands for a
//	    value list - cuscod, isub + 2, "FRED":
//
//	    <PRE>
//	    OPD:    CUSCOD
//	    OPD:    NULL (terminates first value expression)
//	    OPD:    ISUB
//	    OPD:    2
//	    OPD:    +
//	    OPD;    NULL (terminates second value expression)
//	    OPD:    "FRED"
//	    OPD:    NULL (terminates last value expression)
//	    </PRE>
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
//
// @parm    Reference to current context.
// @parm    Pointer to encoded expression iterator. Assumed to point at first
//	    operand of value to be evaluated.
// @parm    Pointer to strings to hold resultant value on exit.
//
bool TtValList::_evaluateValue(TContext& ctxt,
                               list<CECOperand>::const_iterator& itr,
                               string& retVal) const
{
    CECOperand token = *itr;	    // Current token
    stack<CECOperand*> operands;    // Local stack of operand pointers

    // Iterate through the operands for the value, until terminating NULL operand
    // enountered.
    for (;; itr++)
    {
        // If we have reached the end of the encoded expression
        // report an error, as we have not had all the expected values.
        if (itr == _exprEncoded.end())
        {
            ctxt.SetErrorMessage ("Corrupt value list, too few values");
            return false;		// Invalid value list
        }


        token = *itr;

        if (token.Addrm() == addr_none)
        {
            if (operands.size() != 1)
            {
                ctxt.SetErrorMessage ("Illformed value list");
                return false;		// Invalid value list
            }


            CECOperand* valP;		// Pointer to value operand
            valP = operands.top();	// Pop value operand off stack
            operands.pop();		// ...

            if (valP->IsChr())
            {
                retVal = valP->ChrVal();
            }
            else
            {
                char resStr[32];
                sprintf(resStr, "%li", valP->IntVal());
                retVal = string(resStr);
            }
            break;			// Evaluation complete
        }

        else
        {
            _evaluateOpd (ctxt, &operands, &itr);
        }
    }

    // The local operands stack should be empty after processing a value. If
    // it is not, report an error.
    if (!operands.empty())
    {
        ctxt.SetErrorMessage ("Illformed value list, stack not empty");
        return false;
    }

    return true;
} // End TtValList::_evaluateValue()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function evaluates the values in the encoded value list string
//	    set up by the TtValList::Parse() function. The encoded string
//	    consists of a list of operands and operators in reverse Polish
//	    format describing a list of values - with a NULL operand separating
//	    the values, e.g.
//
//	    <PRE>
//	    Input string:      cuscod, 100 + isub, cusnam
//	    Encoded string:    cuscod, NULL, 00, isub, +, NULL, cusnam
//	    Return (example):  "123","199","Jones" (list of strings)
//	    </PRE>
//
//	    This function returns a list of strings, where each string is that
//	    resulting from the evaluation of the corresponding value in the
//	    value list.
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
//
// @parm    Reference to current context.
// @parm    Reference to list of strings to hold returned list of values.
//
bool TtValList::EvaluateValues(TContext& ctxt, list<string>& retVals) const
{
    if (_exprEncoded.empty())
    {
        //_reason = "No values to evaluate";
        return false;
    }

    // Evaluate values in value list.
    list<CECOperand>::const_iterator it = _exprEncoded.begin();
    bool status;		    // Evaluation status

    status = _evaluateValues(ctxt, &it, retVals);

    if (status && (it != _exprEncoded.end()))
    {
        ctxt.SetErrorMessage ("Illformed value list");
        return false;
    }

    return status;
} // End TtValList::EvaluateValues()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This routine is invoked to process an encoded expression containing
//	    one or more values (each value being either a character or integer
//	    expression). The routine returns a list of the evaluated values. On
//	    return, the encoded expression iterator (used to obtain operands)
//	    passed to this routine will be updated, as necessary, so that it
//	    always refers to the last operand procesed by this routine.
//
//	    The following illustrates the format of the operands for a the
//	    value list - cuscod, isub + 2, "FRED":
//
//	    <PRE>
//	    OPD:    CUSCOD
//	    OPD:    NULL (terminates first value expression)
//	    OPD:    ISUB
//	    OPD:    2
//	    OPD:    +
//	    OPD;    NULL (terminates second value expression)
//	    OPD:    "FRED"
//	    OPD:    NULL (terminates last value expression)
//	    </PRE>
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
//
// @parm    Reference to current context.
// @parm    Pointer to encoded expression iterator. Assumed to point to first
//	    operand of the value list.
// @parm    Pointer to list of strings to hold resultant value strings on exit. Empty
//	    list returned if no values.
bool TtValList::_evaluateValues(TContext& ctxt,
                                list<CECOperand>::const_iterator* itP,
                                list<string>& retVals) const
{
    CECOperand token = **itP;	    // Current token
    stack<CECOperand*> operands;    // Local stack of operand pointers

    // Iterate through the operands in the encoded expression,
    // building up a list of evaluated value strings.
    for (long isub = 0; (isub < _valCnt); (*itP)++)
    {
        // If we have reached the end of the encoded expression
        // report an error, as we have not had all the expected values.
        if (*itP == _exprEncoded.end())
        {
            ctxt.SetErrorMessage ("Corrupt value list, too few values");
            return false;		// Invalid value list
        }


        token = **itP;

        if (token.Addrm() == addr_none)
        {
            if (operands.size() != 1)
            {
                ctxt.SetErrorMessage ("Illformed value list");
                return false;		// Invalid value list
            }


            CECOperand* valP;		// Pointer to value operand
            valP = operands.top();	// Pop value operand off stack
            operands.pop();		// ...

            string wrkStr = "";		// Local value string
            if (valP->IsChr())
            {
                wrkStr = valP->ChrVal();
            }
            else
            {
                char resStr[32];
                sprintf(resStr, "%li", valP->IntVal());
                wrkStr = string(resStr);
            }
            retVals.push_back(wrkStr);	// Append value to list of evaluated values
            isub += 1;			// One more value evaluated
            delete valP;
        }
        else
        {
            _evaluateOpd (ctxt, &operands, itP);
        }
    }

    // The local operands stack should be empty - as we should empty the stack after
    // processing each value. If the stack is not empty, report an error.
    if (!operands.empty())
    {
        ctxt.SetErrorMessage ("Illformed value list, stack not empty");
        return false;
    }

    return true;
} // End TtValList::_evaluateValues()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function evaluates a value in the encoded value list string
//	    set up by the TtValList::Parse() function. The encoded string
//	    consists of a list of operands and operators in reverse Polish
//	    format describing a list of values - with a NULL operand separating
//	    the values, e.g.
//
//	    <PRE>
//	    Input string:      cuscod, 100 + isub, cusnam
//	    Encoded string:    cuscod, NULL, 00, isub, +, NULL, cusnam
//	    Return (example):  "123","199","Jones" (list of strings)
//	    </PRE>
//
//	    This function returns the string resulting from the evaluation of
//	    the value in the value list identified by the supplied iterator.
//
// @rdesc   Return status. 'true' if evaluation successful, 'false' otherwise.
//
// @parm    Reference to current context.
// @parm    reference to iterator identifying the first operand of the value
//	    to be evaluated in _exprEncoded.
// @parm    Reference to string to hold returned varible index string..
//
bool TtValList::EvaluateVarIdx(TContext& ctxt, list<CECOperand>::const_iterator* valItrP,
                               string* retValP) const
{
    string wrkVal;	// Local value string to be returned

    if (_exprEncoded.empty())
    {
        //_reason = "No values to evaluate";
        return false;
    }

    // Evaluate index for variable identified by supplied value iterator.
    CECOperand token = **valItrP;	    // Current token
    bool status = true;
    if (token.VarDims() != 0) status = _evaluateVarIdx(ctxt, valItrP, &wrkVal);
    if (!status) return false;

    // Copy variable index string to return parameter.
    *retValP = wrkVal;
    return true;
} // End TtValList::EvaluateVarIdx()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a value list string. The string may consist of
//	    one or more character or integer expressions each separated by a
//	    comma, i.e. the format is:
//
//	    <PRE>
//	    value1 {,valuen}
//	    </PRE>
//
//	    where "value1" and "valuen" are integer or character expressions.
//
//	    The function produces an operand list in postfix (Polish) format.
//	    The ::Reason() method can be used to obtain a string describing
//	    the reason for any failure.
//
//	    The function delegates the real work to the protected function
//	    ::_parse, which will generate a ParseException if the supplied
//	    string is not a valid value list. If a ParseException is
//	    thrown by a called routine this routine sets the _reason field to
//	    indicate the reason for the failure and returns 'false',
//	    otherwise it returns 'true'.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing value list.
// @parm    Position in input string at which to start parsing.
//
bool TtValList::Parse(const string& inStr, const int inPos)
{
    try
    {
        _parse(inStr, inPos);
    }
    catch (ParseException& exc)
    {
        _reason = exc.Reason();
        return false;
    }
    return true;
} // End TtValList::Parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a value list string. A value list string has
//	    the following format:
//
//	    <PRE>
//	    value1 {,valuen}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetect that the
//	    string being parsed is not an integer expression a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
// @parm    String containing value list.
// @parm    Pointer to position in input string at which to start parsing.
//
void TtValList::_parse(const string& inStr, const int pos)
{
    int	wrkpos = pos; // Get copy of string position for local update

    //
    //  Call a helper function to do the real work. Note that we need to check that
    //  there are no extraneous characters following a valid integer expression.
    //
    bool status;    // _psIntExpr status
    _clrExpr();	    // Clear expression parsing data.
    status = _psValList (inStr, &wrkpos, &_valCnt);
    if (status && (wrkpos != (inStr.length())))
    {
        throw ParseException(
            "Extra text following valid value list: " +
            _expandToEos (inStr, wrkpos));
    }
    else if (!status)
    {
        char offbuf[32];
        sprintf(offbuf, "%li", (long) wrkpos);
        throw ParseException("Invalid Value List in : \"" + inStr +
                             "\" at offset: " + string(offbuf) + ".");
    }
    //
    // LINUX Port Changes
    //
#if defined(LINUX)||defined (HPUX)
    // Display resultant postfix format operand list.
    for (list<CECOperand>::iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#else
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#endif
    {
        CECOperand opd = *it;
    }

    return;
} // End TtValList::_parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This mutator function evaluates an encoded string set up by the
//	    TtVarRef::Parse() function. The encoded string consists of a list of
//	    operands and operators in reverse Polish format describing a
//	    subscripted variable, e.g.
//
//	    <PRE>
//	    Input string:      Customers [cuscod, 100 + isub]
//	    Encoded string:    Customers(2), cuscod, NULL, 100, isub, +, NULL
//	    Return (example):  "54,199"
//	    </PRE>
//
//	    This function returns a character separated string containing the
//	    subscript values.
//
// @rdesc   Variable subscript string, e.g. "1,24,99".
// @parm    Rereference to current context.
//
string TtVarRef::Evaluate(TContext& ctxt) const
{
    static const string noResult = "";

    if (_exprEncoded.empty())
    {
        //_reason = "No indices to evaluate";
        return noResult;
    }

    // Process encoded expression.
    string varIdx;		    // Variable index string
    list<CECOperand>::const_iterator it = _exprEncoded.begin();
    bool status;		    // Evaluation status

    status = _evaluateVarIdx(ctxt, &it, &varIdx);

    if (!status)
    {
        return noResult;
    }
    else if (++it != _exprEncoded.end())
    {
        ctxt.SetErrorMessage ("Illformed expression");
        return noResult;
    }

    return varIdx;
} // End TtVarRef::Evaluate()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a variable reference string. A variable
//	    reference string has the following format:
//
//	    <PRE>
//	    variable_name {[ index1 {,index2 {,indexn}} ]}
//	    </PRE>
//
//	    where, variable_name is a variable name and 'indexn' is an integer
//	    expression identifying the array index.
//
//	    The function produces an operand list in postfix (Polish) format.
//	    The ::Reason() method can be used to obtain a string describing
//	    the reason for any failure.
//
//	    The function delegates the real work to the protected function
//	    ::_parse, which will generate a ParseException if the supplied
//	    string is not a variable reference. If a ParseException is
//	    thrown by a called routine this routine sets the _reason field to
//	    indicate the reason for the failure and returns 'false',
//	    otherwise it returns 'true'.
//
// @rdesc   Return status. 'true' if parsing successful, 'false' otherwise.
// @parm    String containing variable reference.
// @parm    Position in input string at which to start parsing.
//
bool TtVarRef::Parse(const string& inStr, const int inPos)
{
    try
    {
        _parse(inStr, inPos);
    }
    catch (ParseException& exc)
    {
        _reason = exc.Reason();
        return false;
    }
    return true;
} // End TtVarRef::Parse()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function parses a variable reference string. A variable
//	    reference string has the following format:
//
//	    <PRE>
//	    variable_name {[ index1 {,index2 {,indexn}} ]}
//	    </PRE>
//
//	    The function produces an operand list in postfix (Polish) format.
//
//	    If this routine, or any routine it calls, detetect that the
//	    string being parsed is not a variable reference (or has trailing
//	    characters following) a ParseException
//	    exception is thrown and the _reason field set up with an
//	    explanation of why.
//
// @parm    String containing variable reference.
// @parm    Pointer to position in input string at which to start parsing.
//
void TtVarRef::_parse(const string& inStr, const int pos)
{
    int	wrkpos = pos;	// Get copy of string position for local update
    bool isvar;		// Variable flag

    // See if we have a character or an integer variable.
    _clrExpr();		// Clear expression parsing data.
    isvar = _psChrVariable(inStr, &wrkpos);
    if (!isvar) isvar = _psIntVariable(inStr, &wrkpos);

    // Throw exception if NOT a variable.
    if (!isvar)
    {
        throw ParseException(
            "Invalid Variable Reference, \"" + _expandToEos (inStr, 0) + "\"");
    }

    // If there are extra characters following the variable reference, throw an exception.
    if (wrkpos != inStr.length())
    {
        throw ParseException(
            "Extra text following variable reference, \"" +
            _expandToEos (inStr, wrkpos) + "\"");
    }

    // Save the variable name - excluding indices.
    long namlen = wrkpos - pos;
    string::size_type ipos = inStr.find_first_of(ps_starray_indices + " \t", pos);
    if (ipos != string::npos)
    {
        namlen = ipos - pos;
    }
    _varnam = inStr.substr(pos, namlen);
    //
    // LINUX Port Changes
    //
#if defined(LINUX)||defined (HPUX)
    // Display resultant postfix format operand list.
    for (list<CECOperand>::iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#else
    // Display resultant postfix format operand list.
    for (list<CECOperand>::const_iterator it = _exprEncoded.begin();
            it != _exprEncoded.end();
            it++)
#endif
    {
        CECOperand opd = *it;
    }

    return;
} // End TtVarRef::_parse()

