/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//  Cvarlist implmentation
//
#ifdef WIN32
#pragma warning(disable:4786)
#endif
#ifndef CYRANO_TP_STANDALONE
#include "CECInterProcessVar.h"
#endif
#include "CECListVar.h"
#include "CECFileVar.h"
#include "CECRangeVar.h"
#include "CECArrayVar.h"
#include "CECSimpleVar.h"

#include "varlist.h"
#include "cxxutils.h"
#include <cassert>
// Type returned by map insert
typedef pair<VARLIST::iterator, bool> ir;
const string  THREAD_COUNT_VAR="THREADCOUNT";
//There is one varlist /thread so any default variables should be put here in the constructor
// sublist - does this list depend on another list, if so do not delete THREAD variables on completion
Cvarlist::Cvarlist():is_sub(false)
{
    //define TestManager Global Thread Count variable
    define_variable("",THREAD_COUNT_VAR,1,VTINT|VCGLOBAL);
}

// Destruction
Cvarlist::~Cvarlist()
{
    delete_all_variables();
}

// Define a new variable
CECVar* Cvarlist::define_variable(
    const string& Scriptname,//Whole file name of script being run in this context
    const string& Name,// Variable Name
    unsigned size,	// Variable Size (char type)
    unsigned options, // Options Bit Field
    const string& dimensions,  // array indecies
    const string& fid,       // File dataname
    const string& file,      // Filename
    const string& values,	//declaration
    unsigned long seed)//random seed

{

    string name(Name);
    string_toupper(name);
    // Create new variable object
    CECVar* vp;
    //Starting here, the general idea is to have all possible
    //methods for accessing variables to be in a single
    // base class from which all types of variables can be derived
    // This hopefully will eliminate the need for checks as to variable
    // type on each access with the variable class itself doing internal checks
    // Therefore the different types of variables should be created here
    // and never explicitly seen again

    //if thread wide variable check to see if already defined
    // if so check declaration is the same (check options)
    // if everything fine do not add to list and return pointer to
    // previously defined variable
    if( options & VCTHREAD)
    {
        VARLIST::iterator i=vars.find(name);
        if(i!=vars.end())//does already exist
        {
            //
            // LINUX Port Changes
            //
            // For linux we cannot use the arrow operator on this object
            if((*i).second->get_options() == options)//options mask is the same
            {
                return (*i).second;//break here and return previously defined variable
            }

            else//problem with variable return NULL
            {
                return NULL;
            }
        }
        //doesn't exist so carry on as normal
    }

#ifndef CYRANO_TP_STANDALONE
    if( options & VCSCRIPT || options & VCGLOBAL  )
    {
        string scriptname=filename_base(Scriptname);
        // Make new CORBA Global type variable
        if(options & VSRANGE )
        {
            vp=new CECInterRangeVar(name,options,scriptname,values,seed);
        }
        else if ( options & VSLIST )
        {
            vp=new CECInterListVar(name,size,options,scriptname,values,seed);
        }
        else if ( options &  VSFILE)
        {
            vp=new CECInterFileVar(name,size,options,fid,file,scriptname);
        }
        else if ( dimensions != "" )
        {
            vp=new CECInterArrayVar(name,size,options,scriptname,dimensions);
        }
        else
        {
            vp=new CECInterSimpleVar(name,size,options,scriptname);
        }

    }
    else
    {
#endif //#ifndef CYRANO_TP_STANDALONE
        if( options & VSLIST )vp=new CECListVar(name,size,options,values,seed);
        else if ( options &  VSFILE)vp=new CECFileVar(name,size,options,fid,file);
        else if(options & VSRANGE )vp=new CECRangeVar(name,options,values,seed);
        else if(dimensions!="")vp=new CECArrayVar(name,size,options,dimensions);
        else vp=new CECSimpleVar(name,size,options);
#ifndef CYRANO_TP_STANDALONE
    }
    assert(vp);
#endif //#ifndef CYRANO_TP_STANDALONE
    // Add variable object pointer to map, key is variable name
    ir s = vars.insert(VARLIST::value_type(name, vp));
    // If pair member "second" is true insert was ok
    if(s.second)
    {
        return vp;
    }
    // Insert failed, possible duplicate variable name
    delete vp;
    return NULL;
}

CECVar * Cvarlist::define_variable(const string& scriptname, const TRecord * rec, const string& dir, bool use_decl)
{
    // Ensure we have a name
    const TParam * par = rec->param("name");
    if(par == NULL)
        return NULL;
    string name = par->strVal();

    // Extract variable options mask
    unsigned options;
    par = rec->param("options");
    if(par == NULL)
        options = DEFAULT_VAR_OPTIONS;
    else
        options = par->intVal();

    // Extract variable size
    unsigned size;
    par = rec->param("size");
    if(par == NULL)
        size = DEFAULT_VAR_SIZE;
    else
        size = par->intVal();
    //See if we have any indecies
    string dimensions;
    par = rec->param("dimensions");
    if(par==NULL)
        dimensions= DEFAULT_ARRAY_SIZE;//Not an array
    else
        dimensions=par->strVal();

    // Extract file id (dataname)
    string fileid;
    par = rec->param("fident");
    if(par == NULL)
        fileid="";
    else
        fileid=par->strVal();

    // Extract filename
    string file;
    par = rec->param("file");
    if(par == NULL)
        file="";
    else
        file=par->strVal();

    string values;
    // Extract values text
    par = rec->param("lValues");
    if(par != NULL)
        values=par->strVal();
    else
        values="";
    // Extract seed for random values
    long seed;
    par = rec->param("seed");
    if(par == NULL)
        seed = 0;
    else
        seed = par->intVal();

    // Define the variable
    CECVar * vp = define_variable(scriptname,name, size, options,dimensions,fileid,file,values,seed);

    // Define failed, possibly duplicate name
    if(vp == NULL)
        return NULL;


    return vp;
}

// Get count of defined variables
int Cvarlist::get_variable_count(void)
{
    // Return count of entries in map
    return (int)vars.size();
}

// Return non zero if variable "name" is defined
int Cvarlist::is_variable_defined(string name)
{
    string_toupper(name);
    VARLIST::iterator it = vars.find(name);
    if(it == vars.end())
        return 0;
    return 1;
}

// Delete variable with "name"
int Cvarlist::delete_variable(string name)
{
    string_toupper(name);
    VARLIST::iterator it = vars.find(name);
    if(it == vars.end())
        return -1;

    delete (*it).second;
    vars.erase(it);
    return 0;
}

// Delete all variables in map
int Cvarlist::delete_all_variables(void)
{
    // Loop deleting each variable class instance
    for(VARLIST::iterator it = vars.begin() ; it != vars.end() ; it++)
    {

        //
        // LINUX Port Changes
        //
        // For linux we cannot use the arrow operator on this object
        if(is_sub && ((*it).second->get_options() & VCTHREAD))continue;//if sub list and is THREAD scope variable do not delete
        delete (*it).second;
    }
    // Clear out the map
    vars.erase(vars.begin(), vars.end());
    return 0;
}

// Delete all volatile variables. The volatile state is
// determined internally by the variable object and
// depends on the option bits set. This can be used for
// retaining variables which must exist across multiple
// tof files during edit or replay.
int Cvarlist::delete_volatile_variables(void)
{
    // Loop deleting each map entry if volatile
    for(VARLIST::iterator it = vars.begin() ; it != vars.end() ; )
    {
        // If a volatile variable
        if( ((*it).second)->is_volatile() )
        {
            // Delete variable object instance
            delete (*it).second;
            // Delete map entry, the returned iterator from erase()
            // should point to the next element beyond the one just
            // removed, this could be end().
            vars.erase(it++);
        }
        // Non volatile, just bump the iterator
        else
            it++;
    }
    return 0;
}

// Delete all temporary variables.
int Cvarlist::delete_temporary_variables(void)
{
    // Loop deleting each map entry if temp
    for(VARLIST::iterator it = vars.begin() ; it != vars.end() ; )
    {
        // If a temp variable
        if( ((*it).second)->is_temp() )
        {
            // Delete variable object instance
            delete (*it).second;
            // Delete map entry, the returned iterator from erase()
            // should point to the next element beyond the one just
            // removed, this could be end().
            vars.erase(it++);
        }
        // Non temp, just bump the iterator
        else
            it++;
    }
    return 0;
}

// Get pointer to variable class object with "name"
CECVar * Cvarlist::get_variable_ptr(string& name)
{
    string_toupper(name);
    VARLIST::iterator it = vars.find(name);

    if(it == vars.end())
        return NULL;
    return (*it).second;
}


// Get pointer to variable class object with offset "index"
CECVar * Cvarlist::get_variable_ptr(int index)
{
    // Initialize iterator to first map entry
    VARLIST::iterator it = vars.begin();
    // Loop through map entries until index reached
    for(int i = 0; it != vars.end() ; i++, it++)
    {
        if(i == index)
            return (*it).second;
    }
    // Index beyond end or empty map
    return NULL;
}

// Get pointer to variable class object with name
// as defined in TRecord
CECVar * Cvarlist::get_variable_ptr(const TRecord * rec)
{
    // Ensure we have a name
    const TParam * par = rec->param("name");
    if(par == NULL)
        return NULL;
    string name = par->strVal();
    return get_variable_ptr(name);
}

// Get variable value at "index" in value list.
// Return empty string on any failure (undefined variable
// or invalid index).
string Cvarlist::get_variable_value(string name, string index)
{
    CECVar * vp = get_variable_ptr(name);
    if(vp == NULL)
        return "";
    return vp->get_value(index);
}

// Set variable value at "index" in value list.
// Return -1 on failure (undefined variable or invalid index).
int Cvarlist::set_variable_value(string name, string value, string index)
{
    CECVar * vp = get_variable_ptr(name);
    if(vp == NULL)
        return -1;
    return vp->set_value(value,index);
}

// Clear value list for variable
void Cvarlist::clear_variable_values(string& name)
{
    CECVar * vp = get_variable_ptr(name);
    if(vp == NULL)
        return;
    vp->clear_values();
}

// Return true if variable defined in "rec" is volatile
// (Local, Script or Global) and is attempting to replace
// a non-volatile (Thread or Private) definition.
// This assumes that previous volatile variables have been
// deleted and this method is called while extracting
// declaration from a new script.
// NOTE: Will not detect non-volatile type conflicts

bool Cvarlist::volatile_conflict(const TRecord* rec)
{
    // Get pointer to any existing variable with same
    // name as that in TRecord name parameter.
    CECVar * vp = get_variable_ptr(rec);
    // If vp NULL it's not defined, so no conflict.
    if(rec == NULL)
        return false;

    // Extract variable options mask from record
    unsigned options;
    const TParam * par = rec->param("options");
    if(par == NULL)
        options = DEFAULT_VAR_OPTIONS;
    else
        options = par->intVal();

    // Attempting to replace non-volatile with volatile ?
    if((CECProcessVar::is_volatile(options)) && (!vp->is_volatile()))
        return true;
    // Assume no conflict
    return false;
}
//////////////////////////////////////////////////////
// @mfunc   return a variable list containing all the
//	    thread wide variables to pass to any scripts
//	    called
// @rdesc   list of thread wide variables
const VARLIST  Cvarlist::get_threadwide_vars()const
{
    VARLIST vlist;
    VARLIST::const_iterator i;
    for(i=vars.begin();i!=vars.end();i++)
    {

        //
        // LINUX Port Changes
        //
        // For linux we cannot use the arrow operator on this object
        if((*i).second->get_options() & VCTHREAD )
        {
            vlist.insert(*i);
        }
    }

    return vlist;
}
//////////////////////////////////////////////////////
// @mfunc   add variables to this list from another list
// @rdesc   number of variables successfully inserted
// @parm    list(map) of variables to insert
int Cvarlist::add_vars(const VARLIST& varlst)
{
    int iret=0;
    VARLIST::const_iterator i;
    for(i=varlst.begin();i!=varlst.end();i++)
    {
        vars.insert(*i);
        iret++;
    }
    is_sub=true;    //have had variables added from elsewhere so is
    //dependent on another list (currently only for thread scope)
    return iret;
}
