/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifdef TH_POOL_VER
#   include <Winsock2.h>
#endif

#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif

#ifdef TH_POOL_VER
#   include "AsyncVirtualUser.h"
#else
#   include "TEThread.h"
#endif

#include "TExecuterRcrd.h"
#include "TExecuterRcrd.inl"

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for Call Script Record
// ================================
///////////////////////////////////////////////////////////////////////////////
int TRcrdCscr::prepare(TScript* script, int objIndex, bool verbose)
{
    // Parameters must be parsed ready fo execute (where they can be evaluated)
    // errlab must exist
    // retvar should be declared (should be able to check this now as
    // all prepares are done after reading in all TRecords
    //parse parameters
    const TParam* prmP = param(TParam::stParameters);
    string l_parameters = prmP->strVal();

    _prmListP = new TtValList(script);
    if (!_prmListP->Parse(l_parameters , 0))
    {
        MSGERR ("Call Script: parameters: " + l_parameters +
                " Failed during PREPARE of parameters with: " + _prmListP->Reason());
        return -1;	// Return 'failure' status
    }

    _labelIndex=PrepareGoto(this,TParam::stErrlab,script,false);
    return 0;
}

#ifndef TH_POOL_VER

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Call Script Record
// ================================
///////////////////////////////////////////////////////////////////////////////
int TRcrdCscr::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_CALL_SCRIPT);
    //Cast context to be TE thread
    TEThread* TEctxt=static_cast<TEThread*>(&ctxt);
    //if not cast safely then this is definitely NOT a test Executer
    if(!TEctxt)
        throw mdbParseException("Invalid TestExecuter Execution Method Called");
    string realscriptname=strPrmVal(TParam::stName, ctxt);
    if(realscriptname=="")
        throw mdbParseException("NULL Script Name Called");
    if(logr) logr->AddTrace(DBGSTD,DETAILLEVEL,"Lock TtValList Mutex");
    TtValList *PList=NULL;
    {
        CTLockGuard<CTMutex> guard(m_mutex);
        PList=new TtValList(*_prmListP);//make local copy to give to thread
        if(logr) logr->AddTrace(DBGSTD,DETAILLEVEL,"Try TtValList Mutex Unlocked");
    }
    if(logr) logr->AddTrace(DBGSTD,DETAILLEVEL,"TtValList Mutex Unlocked");
    int retvar=TEctxt->CallScript(realscriptname,*PList);
    if(PList)delete PList;
    //After call script Thread is in state it was before
    //set the return variable to be the current return value of thread
    string return_var_name = strPrmVal(TParam::stRetvar, ctxt);
    if(return_var_name!="")//return values used
    {
        retvar=ctxt.set_variable(return_var_name, ctxt.GetCurrentReturnValue());
    }
    //if error goto error label if there is one
    int destination=0;
    if(retvar==-1)
    {
        destination=ExecuteGoto(this,ctxt,TParam::stErrlab,&_labelIndex,&m_mutex);//if no error label pass onto ON ERROR
        if(destination==0)//no jump
        {
            string msg=TEctxt->GetLastError();
            throw mdbParseException("Error in Called Script: "+msg);
        }
    }
    else
        destination=0;
    ctxt.set_phase_text(TP_PHASE_NONE);
    return destination;
}

#else

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Call Script Record
// ================================
//
// That object can be called once or twice:
// * once if the called script is pure CPU.
// * twice if the called script initiates asynchronous operations. Must then
//      set a new environment for the new script before starting it and set
//      back the calling script environment when over.
///////////////////////////////////////////////////////////////////////////////
int TRcrdCscr::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = static_cast<CAsyncVirtualUser&> (ctxt);
    int retVar = 0;

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            // Wait for the calling script's requests to complete.

            ctxt.set_phase_text(PHASE_CALL_SCRIPT);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdCscr: executing step 0 for TOF object " << VU.GetScriptPos() << " for script " << VU.GetCurrScriptName() << ".\r\n";
                ReleaseTraceStream();
            }

            // Must first make sure that the requests are over. Otherwise the results of requests from the calling
            // script may be stored in buffers owned by the called script.

            VU.SetTRecordCallStep(1);
            if(VU.SynchroniseAllHTTPClients() == FALSE)
            {
                return TOF_RES_ASYNC_STARTED;
            }
        }
        // No break since should continue if no need to synchronize.

    case 1:
        {
            // Disconnect the clients.

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdCscr: executing step 1 for TOF object " << VU.GetScriptPos() << " for script " << VU.GetCurrScriptName() << ".\r\n";
                ReleaseTraceStream();
            }

            VU.SetTRecordCallStep(2);

            // Break all the connections before calling new script.
            // Note the disconnection of SSL is asynchronous. Thus end-up in case 2.
            if(VU.DisconnectAllClients() == TOF_RES_ASYNC_STARTED)
                return TOF_RES_ASYNC_STARTED;
        }
        // No break since should continue if no need to synchronize.

    case 2:
        {
            // Execute the called script.

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdCscr: executing step 2 for TOF object " << VU.GetScriptPos() << " for script " << VU.GetCurrScriptName() << ".\r\n";
                ReleaseTraceStream();
            }

            CAsyncVirtualUser& VU = static_cast<CAsyncVirtualUser&> (ctxt);

            string realscriptname = strPrmVal(TParam::stName, ctxt);
            if(realscriptname.empty())
                throw mdbParseException("NULL Script Name Called");

            {
                CLock guard (m_pFastLock);
                VU.SetParamList(*_prmListP);    // Make local copy to give to thread.
            }


            VU.SetTRecordCallStep(3);

            retVar = VU.CallScript(realscriptname/*, PList*/);
            if(retVar != TOF_RES_OK)
            {
                // Either async op or over.
                return retVar;
            }

            // Called script is over (did not initiate any async. op.).

        }
        // No break since should continue if no need to synchronize.

    case 3:
        {
            // Complete the execution: set back the calling script variables.

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdCscr: executing step 3 for TOF object " << VU.GetScriptPos() << " for script " << VU.GetCurrScriptName() << ".\r\n";
                ReleaseTraceStream();
            }

            // Entered when the called script is over. Nothing to do except return TOF_RES_OK to

            // Set data values back to the calling script.
            //            VU.EndScript();

            // After call script Thread is in state it was before
            // set the return variable to be the current return value of thread
            string return_var_name = strPrmVal(TParam::stRetvar, ctxt);
            if(return_var_name.empty() == FALSE)    // Return values used
            {
                retVar = ctxt.set_variable(return_var_name, ctxt.GetCurrentReturnValue());
            }

            // If error goto error label if there is one
            ctxt.set_phase_text(TP_PHASE_NONE);
            VU.SetTRecordCallStep(0);   // Reset.
            if(retVar == -1)
            {
                int destination = ExecuteGoto(this, ctxt, TParam::stErrlab, &_labelIndex, m_pFastLock);//if no error label pass onto ON ERROR
                if(destination==0)//no jump
                {
                    //		std::cout << ctxt.GetThreadID() << " throw exception in callscript" << std::endl;
                    throw mdbParseException("Error in CallScript");
                }

                return destination;
            }
            else
            {
                return TOF_RES_OK;
            }
        }

    default :
        {
            assert(0);
            return TOF_RES_FAILED;
        }
    }
}

#endif

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for Entry Record
// ================================
///////////////////////////////////////////////////////////////////////////////
int TRcrdEntry::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stParameters);
    string l_parameters = prmP->strVal();

    _prmListP = new TtValList(script);
    if (!_prmListP->Parse(l_parameters , 0))
    {
        MSGERR ("Entry: parameters: " + l_parameters +
                " Failed during PREPARE of parameters with: " + _prmListP->Reason());
        return -1;	// Return 'failure' status
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Entry Record
// ================================
///////////////////////////////////////////////////////////////////////////////

#ifndef TH_POOL_VER

int TRcrdEntry::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ENTRY);
    //Cast context to be TE thread
    TEThread* TEctxt=static_cast<TEThread*>(&ctxt);
    //if not cast safely then this is NOT a test Executer
    if(!TEctxt)
        throw mdbParseException("Invalid TestExecuter Execution Method Called");
    {
        CTLockGuard<CTMutex> guard(m_mutex);
        TtValList PList(*_prmListP);//make local copy to give to thread
        TEctxt->Entry(PList);
    }//End of mutex guard
    ctxt.set_phase_text(TP_PHASE_NONE);
    return 0;
}

#else

int TRcrdEntry::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ENTRY);

    CAsyncVirtualUser& VU = static_cast<CAsyncVirtualUser&> (ctxt);

    {
        CLock guard (m_pFastLock);
        TtValList PList(*_prmListP); // Make local copy for use by the thread.
        VU.Entry(PList);
    }

    ctxt.set_phase_text(TP_PHASE_NONE);

    return 0;
}

#endif

