/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//========================================================================
//
// TEHttpRecords.h
// ---------------
//=========================================================================

#ifndef CYR_TEHTTPRECORDS_H
#define CYR_TEHTTPRECORDS_H

#ifndef TH_POOL_VER
#   include "HTTPconn.h"
#endif

#include <string>

#ifdef WIN32
#pragma warning(disable:4786)
using namespace std;
#endif

#include <TRecord.h>

#ifdef TH_POOL_VER

#   include "MemoryBufferHandler.hpp"
#   include "HTMLParser/PhParse.hpp"
#   include "PhParseSpecialised.tpp"
enum { HEADER_VAR=0, BODY_VAR=1 };
typedef map<std::string, TtChrExpr *> variableAddressMap;

#endif

///////////////////////////////////////////////////////////////////////////////
// @class   TRcrdConnect | This class is used to Connect in an HTTP environment
//
// @base    public | TRecord.
//
class TRcrdConnect : public TRecord
{

// @access Public Members.
public:
    // @cmember Set variable record ATVF key.
    static const AtvfKey KEY;

    // @cmember Constructor.
    TRcrdConnect (const TParamRule* headRuleP) : TRecord(headRuleP)
    , _exprP(NULL)
    {}
    // @cmember Destructor.
    ~TRcrdConnect ()
    {
        if(_exprP)
            delete _exprP;
    }

    // @cmember Execute record in current context and return status.
    virtual int execute (TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdConnect (*this);
    };

// @access Private Members
private:
    // @cmember
    static const bool _registeredCreator;
    // @cmember
    static const bool _RegisterCreator();

	TtChrExpr* _exprP;
#ifndef TH_POOL_VER
    static CTMutex m_mutex;
#endif
}
; // TRcrdConnect


/**
 * Definition of Request Record class
 **/
class TRcrdRequest : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdRequest(const TParamRule* headRuleP) : TRecord(headRuleP), 
                                        _uriP(NULL), _headerP(NULL), _withP(NULL), 
                                        _woutP(NULL), _bodyP(NULL), _certName(NULL), 
										_bSaveDocument(false), _iConid(0)
#ifdef TH_POOL_VER
										,_eResultsRequired(ALL_DATA) 
#endif
										{}

    // @cmember Destructor.
    ~TRcrdRequest ()
    {
        if (_uriP)  delete _uriP;
        if(_headerP)delete _headerP; 
        if(_withP)  delete _withP;
        if(_woutP)  delete _woutP; 
        if(_bodyP)  delete _bodyP;
		if(_certName) delete _certName;
    }

  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdRequest(*this);
    };
  // Mutators

#ifndef TH_POOL_VER
    std::string getRequestString(int eType) const;
#else
    typedef enum {HTTP_REQUEST_GET=0, HTTP_REQUEST_HEAD=1, HTTP_REQUEST_POST=2, /*HTTP_REQUEST_CONNECT=3, */HTTP_REQUEST_NONE=4} ERequestType;

    const std::string &getRequestString(ERequestType eType) const; 
    void SetRequiredResults(const ERequiredResults res) { _eResultsRequired = res; }
    ERequiredResults GetRequiredResults()               const { return _eResultsRequired; }

protected:
    //// Helpers.
    void BuildHeaderString(TContext& ctxt, std::string &szFormattedHeader) const;
    void ExtractHostPort(TContext& ctxt, const std::string &szHeader, const std::string &szUri, std::string &szRemoteServer, std::string &szPort) const;
    int	Ascii2Binary(const std::string &inBuf, std::string &outBuf) const;
#endif

private:
    static const bool _registeredCreator;
    static const bool _RegisterCreator();
	TtChrExpr* _uriP;
	TtValList* _headerP;
	TtValList* _withP;
	TtValList* _woutP;
	TtChrExpr* _bodyP;
	TtChrExpr* _certName;

#ifndef TH_POOL_VER
    static CTMutex evaluateMtx;
#endif

public:
    int _iConid;
    bool _bSaveDocument;
#ifdef TH_POOL_VER
    ERequiredResults    _eResultsRequired;
#endif
}
; // TRcrdRequest

/**
 * Definition of Disconnect Record class
 **/
class TRcrdDisconnect : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdDisconnect(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdDisconnect(*this);
    };
  // Mutators

private:
    static const bool _registeredCreator;
    static const bool _RegisterCreator();
}
; // TRcrdDisc

/**
 * Definition of HttpResponse Record class
 **/
class TRcrdHttpResponse : public TRecord
{
public:
	static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdHttpResponse(const TParamRule* headRuleP) : TRecord(headRuleP)
                    , _varP(NULL), _exprP(NULL), _identP(NULL), _varSize(0)
    {}
    // @cmember Destructor.
    ~TRcrdHttpResponse()
    {
        if(_varP) delete _varP;
        if(_exprP) delete _exprP; 
        if(_identP) delete _identP;
		for (AddressQueryList::iterator itr = _addressQueryList.begin(); itr != _addressQueryList.end(); itr ++)
			delete *itr;
		for (variableAddressMap::iterator itrMap = _varAddressMap.begin(); itrMap != _varAddressMap.end(); itrMap ++)
			delete itrMap->second;
    }

  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdHttpResponse(*this);
    };
  // Mutators

private:
    int _varSize;   // variable size

#ifdef TH_POOL_VER
    //// Helpers
    BOOL    ParseCookieValue(CDataBuffHandler &messageBuff, const string &szName,   BOOL bFoundContinueHeader, string &zCookieVal) const;
    void    ParseHeaderValue(CDataBuffHandler &messageBuff, const string &szIdent,  BOOL bFoundContinueHeader, string &zHeaderValue) const;
    void    DumpHeader      (CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const;
    void    DumpBody        (CDataBuffHandler &messageBuff, string &szVarVal, DWORD dwVarLength, BOOL bFoundContinueHeader) const;
#endif

    // @cmember Destination label index in script
    int	_labelIndex;
    // @cmember used to allow safe writes during run-phase
#ifndef TH_POOL_VER
	static CTMutex m_mutex;
#endif
    // @cmember Pointer to set variable 'name' object.
    TtVarRef* _varP;
    // @cmember Pointer to set variable 'expression' object.
    CECTofTxt* _exprP;
    // @cmember Pointer to set variable 'identity' object.
	TtChrExpr* _identP;
    // @cmember
    static const bool _registeredCreator;
    // @cmember
    static const bool _RegisterCreator();
public:

	int _iConid ;
	int _responseType ;				// header or body
	AddressQueryList _addressQueryList ; // List of compiled addresses 
	variableAddressMap	_varAddressMap ;	// Map of HTML address in ascii mode
									// These address cannot be converted in the
									// prepare phase because they are not constant.	
									// key = variable, value = expression
}
; // TRcrdHttpResponse


/**
 * Definition of Synchronize Record class
 **/
class TRcrdSynchronize : public TRecord
{
public:
    static const AtvfKey KEY;
  // Constructors & Destructor
    TRcrdSynchronize(const TParamRule* headRuleP) : TRecord(headRuleP)
    {}
  // Accessors
    virtual int execute(TContext&);
    // @cmember Prepare record for execution and return status.
    virtual int prepare (TScript* script, int objIndex, bool verbose=false);
    // @cmember Duplicate object.
    virtual TRecord* duplicate() const
    {
        return new TRcrdSynchronize(*this);
    };
  // Mutators

protected:
    int _labelIndex;    // Destination label index in script
private:
#ifndef TH_POOL_VER
	static CTMutex m_mutex;
#endif

    static const bool _registeredCreator;
    static const bool _RegisterCreator();
}
; // TRcrdDisc


#ifndef TH_POOL_VER
int server_specific_init(void);
int server_specific_cleanup(void);
#endif

#endif