/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyranoOTA
// @module  OTADetailQuery | This module contains the implementation of the
//          OTADetailQuery class.  This class allows the storage of detailed
//          queries in a file for storage.
//
///////////////////////////////////////////////////////////////////////////////

#include <OtaDetail.h>

///////////////////////////////////////////////////////////////////////////////
#include <fstream>
using std::ofstream;
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor assigns member variables for this query object
//          including the query string itself and the description.  It must
//          also calculate the header information for the query.
// @parm    The query string.
// @parm    The description for the query.
OtaDetailQuery::OtaDetailQuery( string query, string desc )
{
    // Assign member variables
    m_query = query;
    m_desc = desc;
    m_headers = "";
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor does nothing.
OtaDetailQuery::~OtaDetailQuery()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a result to the list of current results.  The
//          string is already formatted appropriately for output to a file.
// @parm    The output string.

void OtaDetailQuery::AddResult( string result )
{
    m_results.push_back( result );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the string used in the query.
// @rdesc   The query string.
string OtaDetailQuery::GetQuery()
{
    return m_query;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method sets the header information for this query.  This
//          information will be stored in the final results file.
// @parm    The header information.
void OtaDetailQuery::SetHeaders( string headers )
{
    m_headers = headers;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description for the query.
// @rdesc   The description string.
string OtaDetailQuery::GetDescription()
{
    return m_desc;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the header string for the query.  Each entry
//          represents a column from the query.  The entries are tab separated
//          so that they can be written directly to a file.
// @rdesc   The header string.
string OtaDetailQuery::GetHeaders()
{
    return m_headers;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method outputs all the results to a named file.
// @parm    The fully-qualified name of the file to write to.
void OtaDetailQuery::WriteToFile( const char* filename )
{
    // Open file for writing
    ofstream* f = new ofstream( filename, std::ios::binary|std::ios::out);
    if ( f->good() )
    {
        // Write header information
        f->write( "0\t", 2 );
        f->write( m_desc.c_str(), m_desc.length() );
        f->write( "\r\n", 2 );

        f->write( "1\t", 2 );
        f->write( m_headers.c_str(), m_headers.length() );
        f->write( "\r\n", 2 );

        // Write out results.  They are already in a format suitable for this task.
        for( size_t i = 0; i < m_results.size(); i++ )
        {
            string entry = m_results[i];
            f->write( entry.c_str(), entry.length() );
        }
    }

    // Close file and flush contents
    delete f;
}

