/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_FASTLOCK_H_
#define _OPENSTA_FASTLOCK_H_

#include <assert.h>

////////////////////////////////////////////////////////////////////////////////
// Fast lock definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

// This class implements a spinlock. Cross-process locking is not supported.
// The worst-case locking involves an event kernel object, which by default is
// not pre-created.
class CFastLock
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    // Constructor. The spin count specified is only taken into account for a multiprocessor
    // machine; otherwise it is forced to 0.
    CFastLock(DWORD dwSpinCount = 4000, BOOL bPreCreateEvent = FALSE)
        : m_lLockRequests(0), m_dwLockDepth(0), m_dwOwningThreadId(0),
          m_hEvent(NULL)
    {
        assert(s_bSingleProcessorMachine != (BOOL)-1);
        m_dwSpinCount = (s_bSingleProcessorMachine ? 0 : dwSpinCount);
        if(bPreCreateEvent)
        {
            // Create the event kernel object as auto-reset.
            m_hEvent = CreateEventW(NULL, FALSE, FALSE, NULL);
        }
    }
    ~CFastLock()
    {
        if(m_hEvent)
            CloseHandle(m_hEvent);
    }

    //// --------------------------------------------------------------------------------
    //// Initialization.
    //// --------------------------------------------------------------------------------
    static void Init()
    {
        SYSTEM_INFO si;
        GetSystemInfo(&si);
        s_bSingleProcessorMachine = (si.dwNumberOfProcessors == 1);
    }

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------
    void Lock();
    void Unlock();

private:
    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    static BOOL s_bSingleProcessorMachine;  // TRUE if the host is a single-processor machine.

    DWORD   m_dwSpinCount;          // The spin count. Always 0 for single-processor machines.
    LONG    m_lLockRequests;        // The number of threads waiting for the lock.
    DWORD   m_dwLockDepth;          // The recurse count of the lock.
    DWORD   m_dwOwningThreadId;     // The owning thread. 0 = no owning thread.
    HANDLE  m_hEvent;               // Handle of the associated event kernel object. NULL = no event.
};

// To use a fast lock, create an instance of the following object on the stack:
class CLock
{
public:
    CLock(CFastLock* pSl) : m_pSl(pSl)
    {
        m_pSl->Lock();
    }
    CLock(CFastLock& sl) : m_pSl(&sl)
    {
        m_pSl->Lock();
    }
    ~CLock()
    {
        m_pSl->Unlock();
    }

private:
    CFastLock *m_pSl;

    // The following prototypes make the class stack-only:
    void* operator new(size_t);
    void* operator new[](size_t);
};

#endif  // _OPENSTA_FASTLOCK_H_
