/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "Tools.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

// Debault values.
enum { DEF_MAX_DATA_BUFF_COUNT=64000 }; // Multiplied by the page size to compute
                                        // the MAXIMUM global buffer length (256 Mb).
                                        // If devided by the max number of users,
                                        // gives the average number of buffers 
                                        // available to each user.

enum { DEF_GROWING_BUFF_COUNT=1000 };   // Number of buffers that are committed at once.
                                        // Each commit reserves memory for 
                                        // DEF_GROWING_BUFF_COUNT/WSABUF_COUNT requests slot.
                                        // Ideally, should be WSABUF_COUNT*(max number of
                                        // concurrent requests) = WSABUF_COUNT*4*(max nb VU).

/* #########################################################################

    CDataBuff

   ######################################################################### */

// Move dwInc buffers further in the list.
WILL_INLINE CDataBuff *CDataBuff::operator+(DWORD dwInc)
{
    CDataBuff *cursor = this;

    for(DWORD dw=0 ; dw<dwInc && cursor!=NULL ; dw++)
    {
        cursor = cursor->GetNextBuff();
    }

    return cursor;
}

WILL_INLINE std::ostream &operator << (std::ostream& os, const CDataBuff &dataBuff)
{
    os.write(dataBuff.GetDataPtr(), dataBuff.GetUsedDataSize());

    return os;
}


/* #########################################################################

    CGlobalDataBuffHandler

   ######################################################################### */

WILL_INLINE CGlobalDataBuffHandler::CGlobalDataBuffHandler()
    : m_dwCommittedMem(0), m_freeDataBuffsList(NULL), m_lastFreeDataBuffer(NULL), m_dwFreeBuffsCount(0), m_dwAllocatedBuffsCount(0)
{
}

WILL_INLINE CGlobalDataBuffHandler::~CGlobalDataBuffHandler()
{
    assert(m_dwAllocatedBuffsCount == m_dwFreeBuffsCount);  // Make sure all the buffers were released.
}

// Initialise static data in CGlobalDataBuffHandler.
// Includes enquiring for system page size and allocating the virtual
// memory buffer.
// Return FALSE if the virtual memory buffer could not be allocated.
WILL_INLINE BOOL CGlobalDataBuffHandler::Init(DWORD dwMaxDataBuff, DWORD dwGrowingBuffCount)
{
    assert(m_pFl == NULL);
    if(m_pFl == NULL)
        m_pFl = new CFastLock();

    if(m_pFl == NULL)   return FALSE;

    if(dwMaxDataBuff == 0)      s_dwMaxBuffCount    = DEF_MAX_DATA_BUFF_COUNT;

    if(dwGrowingBuffCount == 0) s_dwGrowingBuffCount= DEF_GROWING_BUFF_COUNT;

#ifdef USE_PERF_DATA
    s_pRecvBuffAllocatedMem    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_RECV_BUFF_ALLOCATED_MEM);
    (*s_pRecvBuffAllocatedMem) = 0;
    s_pRecvBuffUsedMem         = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_RECV_BUFF_USED_MEM);
    (*s_pRecvBuffUsedMem)      = 0;
    s_pDataStoredCounterRef    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_DATA_STORED);
    (*s_pDataStoredCounterRef) = 0;
#endif

    SYSTEM_INFO sysInfo;

    GetSystemInfo(&sysInfo);

    s_dwMaxPhysBuffLength       = sysInfo.dwPageSize;
    s_dwDataBuffExtraInfoSize   = sizeof(CDataBuff *) + sizeof(DWORD);
    s_dwDataBuffUsableSize      = s_dwMaxPhysBuffLength-s_dwDataBuffExtraInfoSize;

    s_dwBytesGrownBy = s_dwMaxPhysBuffLength * s_dwGrowingBuffCount;

    return AllocMemoryBuffer();
}

WILL_INLINE void CGlobalDataBuffHandler::UnInit()
{
    if(s_pvVirtualMemBuff != NULL)
    {
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
        InterlockedExchange(CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem, 0);
#   else
        (*CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem) = 0;
#   endif
#endif
        if(VirtualFree(s_pvVirtualMemBuff, 0, MEM_RELEASE) == FALSE)
        {
            DWORD dwErr = GetLastError();
            assert(0);
        }
    }
}

// ----------------------------------------------------------------------------

// Returns a list of dwBuffCount free buffers allocated from the global list.
// Returns FALSE if all the buffers could not be allocated.

// [IN]     dwBuffCount:    number of buffers requested.
// [OUT]    pNewList:       pointer to the first item in the allocated list.
// [OUT]    pLastItem:      pointer to the last item in the allocated list.

// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::RequestBuffs(DWORD dwBuffCount, CDataBuff *&pNewList, CDataBuff *&pLastItem)
{
    assert(dwBuffCount != 0);

    CLock lock(m_pFl);

    if(m_dwFreeBuffsCount <= dwBuffCount)
    {
        // Commit enough memory to hold the required buffers rounded up to the growing size.
        if(CommitMemoryBuffer((((dwBuffCount-m_dwFreeBuffsCount)/s_dwGrowingBuffCount)+1)*s_dwGrowingBuffCount) == FALSE)
            return FALSE;
    }

    // The free buffers are taken from the begining of the free list.
    pNewList    = m_freeDataBuffsList;
    pLastItem   = (*m_freeDataBuffsList)+(dwBuffCount-1);

    m_freeDataBuffsList = (*pLastItem)+1;
    m_dwFreeBuffsCount -= dwBuffCount;

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pRecvBuffUsedMem, dwBuffCount*s_dwMaxPhysBuffLength);
#   else
    (*CGlobalDataBuffHandler::s_pRecvBuffUsedMem) += dwBuffCount*s_dwMaxPhysBuffLength;
#   endif
#endif

    pLastItem->SetNextBuff(NULL);

    return TRUE;
}


// ----------------------------------------------------------------------------
//
//  ReturnBuffs
//
//  Return buffers to the global free buffers list.
//
// ----------------------------------------------------------------------------
// TODO : recevoir pointeur sur dernier + nombre (pour counter)?

WILL_INLINE void CGlobalDataBuffHandler::ReturnBuffs(CDataBuff *pDataBuffs)
{
    assert(pDataBuffs != NULL);

    CLock lock(m_pFl);

    m_lastFreeDataBuffer->SetNextBuff(pDataBuffs);

    while(m_lastFreeDataBuffer->GetNextBuff() != NULL)
    {
#ifdef USE_PERF_DATA
        // TODO: should do it only once.
#   ifdef LOCK_COUNTERS
        InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pRecvBuffUsedMem, -(LONG)s_dwMaxPhysBuffLength);
#   else
        (*CGlobalDataBuffHandler::s_pRecvBuffUsedMem) -= (LONG)s_dwMaxPhysBuffLength;
#   endif
#endif
        m_lastFreeDataBuffer = m_lastFreeDataBuffer->GetNextBuff();
        m_dwFreeBuffsCount++;
    }
}


// ----------------------------------------------------------------------------
//
// Create the global memory buffer as a Virtual Memory area.
//
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::AllocMemoryBuffer()
{
    s_pvVirtualMemBuff = VirtualAlloc(NULL, s_dwMaxBuffCount*s_dwMaxPhysBuffLength, MEM_RESERVE, PAGE_READWRITE);
    return s_pvVirtualMemBuff != NULL;
}


// ----------------------------------------------------------------------------
//
// Reserve several pages in the virtual memory pointed to by s_pvVirtualMemBuff.
//
// [IN] dwNbBuffRequired: Number of memory pages to allocate.
//
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::CommitMemoryBuffer(DWORD dwNbBuffRequired)
{
    dwNbBuffRequired = __min(dwNbBuffRequired, s_dwMaxBuffCount-m_dwAllocatedBuffsCount);

    DWORD dwToCommitSize = dwNbBuffRequired*s_dwMaxPhysBuffLength;
    LPVOID pNewBuff = VirtualAlloc((LPVOID)((CHAR*)s_pvVirtualMemBuff + m_dwCommittedMem), dwToCommitSize, MEM_COMMIT, PAGE_READWRITE);

    if(pNewBuff == NULL)
    {
        return FALSE;
    }

    m_dwAllocatedBuffsCount += dwNbBuffRequired;
    m_dwFreeBuffsCount      += dwNbBuffRequired;
    m_dwCommittedMem        += dwToCommitSize;

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem, dwToCommitSize);
#   else
    (*CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem) += dwToCommitSize;
#   endif
#endif

    // Add these free buffers to m_freeDataBuffsList.
    if(m_freeDataBuffsList == NULL)
    {
        // We are allocating data buffers for the first time.
        m_freeDataBuffsList = (CDataBuff*)s_pvVirtualMemBuff;
    }
    else
    {
        assert(m_lastFreeDataBuffer != NULL);

        m_lastFreeDataBuffer->SetNextBuff((CDataBuff*)pNewBuff);
    }

    // Map m_lastFreeDataBuffer to the last buffer allocated (end of the committed memory minus the size of a CDataBuff.
    m_lastFreeDataBuffer = (CDataBuff*)((CHAR*)pNewBuff + dwToCommitSize - s_dwMaxPhysBuffLength);

    // Init the buffers.
    CHAR *pBuff = (CHAR *)pNewBuff;
    for(DWORD dw=0 ; dw<dwNbBuffRequired-1 ; dw++)
    {
        ((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*dw))->Init((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*(dw+1)));
    }

    ((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*dw))->Init(NULL);

    return TRUE;
}


/* #########################################################################

    CDataBuffList

   ######################################################################### */

WILL_INLINE CDataBuffList::CDataBuffList()
    : m_pDataBuffList(NULL), m_pLastItem(NULL)
{}


// ----------------------------------------------------------------------------
//
// Add dwNbItems data-buffers, allocated from the global buffers list.
//
// [IN] dwNbItems: Number of data-buffers to allocate.
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CDataBuffList::AllocateBuffs(DWORD dwNbItems)
{
    CDataBuff *pLastBuff = m_pLastItem;
    CDataBuff *pNewBuffs;
    
    BOOL bRes = m_globalDataBuffHandler.RequestBuffs(dwNbItems, pNewBuffs, m_pLastItem);
    assert(bRes == TRUE);

    if(pLastBuff == NULL)
    {
        assert(m_pDataBuffList == NULL);
        m_pDataBuffList = pNewBuffs;
    }
    else
        pLastBuff->SetNextBuff(pNewBuffs);

    return bRes;
}


// ----------------------------------------------------------------------------
//
// Release the whole data buff list.
// Send back the items to the global buffer list to make them available again.
//
// ----------------------------------------------------------------------------

WILL_INLINE void CDataBuffList::ReleaseBuffList()
{
    if(m_pDataBuffList != NULL)
    {
        m_globalDataBuffHandler.ReturnBuffs(m_pDataBuffList);
    }

    m_pDataBuffList = m_pLastItem = NULL;
}

// ----------------------------------------------------------------------------
//
// Append a buffer list to the end of the current list.
//
// ----------------------------------------------------------------------------

WILL_INLINE const CDataBuffList &CDataBuffList::operator+=(const CDataBuffList &buffList)
{
    assert(m_pLastItem != NULL);

    m_pLastItem->SetNextBuff(buffList.m_pDataBuffList);
    m_pLastItem = buffList.m_pLastItem;

    return *this;
}


// ----------------------------------------------------------------------------
//
// Stores the first dwNbItemsToAdd buffers to the current list, and return a 
// list containing the remaining buffers. buffList should not be used anymore.
// Argument passed as buffList may refer to the same object as the one passed 
// as unusedBuffList.
//
// ----------------------------------------------------------------------------

WILL_INLINE void CDataBuffList::AddBuffList(const CDataBuffList &buffList, DWORD dwNbItemsToAdd, CDataBuffList &unusedBuffList)
{
    if(dwNbItemsToAdd != 0)
    {
        if(m_pDataBuffList == NULL)
        {
            // We are an empty list.
            m_pDataBuffList = buffList.m_pDataBuffList;
        }
        else
        {
            m_pLastItem->SetNextBuff(buffList.m_pDataBuffList);
        }

        m_pLastItem = (*buffList.m_pDataBuffList) + (dwNbItemsToAdd - 1);

        assert(m_pLastItem != NULL);

        unusedBuffList.m_pDataBuffList = (*m_pLastItem) + 1;

        if(unusedBuffList.m_pDataBuffList == NULL)
        {
            // All the list was added. Return an empty one.
            unusedBuffList.m_pLastItem = NULL;
        }
        else
        {
            unusedBuffList.m_pLastItem = buffList.m_pLastItem;
        }

        m_pLastItem->SetNextBuff(NULL);
    }
    else
    {
        // Returning an empty list.
        unusedBuffList.m_pDataBuffList  = buffList.m_pDataBuffList;
        unusedBuffList.m_pLastItem      = buffList.m_pLastItem;
    }
}

WILL_INLINE void CDataBuffList::AddBuffList(const CDataBuffList &buffList, CDataBuff *pLastBuff, CDataBuffList &unusedBuffList)
{
    if(pLastBuff != NULL)
    {
        if(m_pDataBuffList == NULL)
        {
            // We are an empty list.
            m_pDataBuffList = buffList.m_pDataBuffList;
        }
        else
        {
            m_pLastItem->SetNextBuff(buffList.m_pDataBuffList);
        }

        m_pLastItem = pLastBuff;

        unusedBuffList.m_pDataBuffList = pLastBuff->GetNextBuff();

        if(unusedBuffList.m_pDataBuffList == NULL)
        {
            // All the list was added. Return an empty one.
            unusedBuffList.m_pLastItem = NULL;
        }
        else
        {
            unusedBuffList.m_pLastItem = buffList.m_pLastItem;
        }

        m_pLastItem->SetNextBuff(NULL);
    }
    else
    {
        // Returning an empty list.
        unusedBuffList.m_pDataBuffList  = buffList.m_pDataBuffList;
        unusedBuffList.m_pLastItem      = buffList.m_pLastItem;
    }
}

WILL_INLINE std::ostream &operator << (std::ostream& os, const CDataBuffList &dataBuffList)
{
    const CDataBuff *cursor = dataBuffList.GetBuffList();

    while(cursor != NULL)
    {
        os << *cursor;
        cursor = cursor->GetNextBuff();
    }

    return os;
}


/* #########################################################################

CDataBuffListChunkAccessor

  Accessor that specializes in parsing chunk data.
  The iterator must first be moved right after the header since the chunk 
  accessor expects the first element to be a chunk size string.

  The chunk accessor should not be used to parse data 'on the fly'.

  Define the specialisation of the methods in CDataBuffListBaseAccessor
  that behave differently when used with chunks.

######################################################################### */

template <>
WILL_INLINE CDataBuffListBaseAccessor<CdBLCA>::CDataBuffListBaseAccessor<CdBLCA>(CDataBuffList *pDataBuffList)
    : m_dwCurrentChunkSizeLeft(0), m_dwPosInCurrDataBuff(0), m_pDataBuffList(pDataBuffList)
{
    if(pDataBuffList == NULL)
    {
        assert(0);
        throw;
    }

    m_pCurrDataBuff = m_pDataBuffList->GetBuffList();

    // Adjust chunk information.
    // TODO: deal with return values.
    BOOL bRes = ckReadChunkSize();
    assert(bRes == TRUE);
}


// ----------------------------------------------------------------------------
//
// SkipChunkCRLF
//
// Local helper function that reads the CRLF following the chunk size string.
// Assumes the first char found is either CR or LF.
//
// WARNING: parameters will change even when returning FALSE.
//
// Note that because we are a tolerant HTTP application, we consider 
// any of the \r\n and \n sequences as CRLF .
//
// ----------------------------------------------------------------------------

WILL_INLINE BOOL SkipChunkCRLF(DWORD &dwPosInBuff, CDataBuff *&pDataBuffCurs)
{
    CHAR curChar;

    for(DWORD dw=0 ; dw<2 ; dw++)
    {
        // Make sure we point at the right char.
        if(dwPosInBuff == pDataBuffCurs->GetUsedDataSize())
        {
            // Move to the next buffer.
            if(pDataBuffCurs->GetNextBuff() == NULL)
                return FALSE;

            pDataBuffCurs = pDataBuffCurs->GetNextBuff();
            dwPosInBuff = 0;
        }

        curChar = pDataBuffCurs->GetDataPtr()[dwPosInBuff];
        if(IsEOL(curChar))
            dwPosInBuff++;
        // else do not move the cursor the second iteration in case server returned a single-char EOL.
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// ckSkipChunkCRLF()
//
// Reads the CRLF following the chunk size string.
// Assumes the first char found is either CR or LF.
//
// Does not move the cursor if could not skip the CRLF entirely.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckSkipChunkCRLF()
{
    DWORD dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;
    CDataBuff *pCurrDataBuff = m_pCurrDataBuff;

    if(SkipChunkCRLF(dwPosInCurrDataBuff, pCurrDataBuff) == FALSE)
        return FALSE;

    m_pCurrDataBuff = pCurrDataBuff;
    m_dwPosInCurrDataBuff = dwPosInCurrDataBuff;

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// Try moving the iterator dwItems items further. Returns the number of items 
// it skipped actually.
// The three cases are: move in the current buffer; must move to another buffer; 
// must move to the next buffer, but not received yet.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLCA>::Skip(DWORD dwItems)
{
    DWORD dwToSkip = dwItems;
    DWORD dwCurBuffUsedSize, dwCurrBuffSizeLeft;

    while(dwToSkip > 0L)
    {
        // Skip either dwToSkip characters, or till the end of the current chunk, 
        // or till the end of current buffer, or any combination of these.
        dwCurBuffUsedSize   = m_pCurrDataBuff->GetUsedDataSize();
        dwCurrBuffSizeLeft  = dwCurBuffUsedSize - m_dwPosInCurrDataBuff;

        if(dwToSkip > dwCurrBuffSizeLeft)
        {
            // Current buffer not wide enough. Must skip till one of the next buffers.
            if(dwCurrBuffSizeLeft <= m_dwCurrentChunkSizeLeft)
            {
                // Meets end of current chunk before or at the same time as the end of the current buffer.
                dwToSkip                    -= m_dwCurrentChunkSizeLeft;
                m_dwPosInCurrDataBuff       += m_dwCurrentChunkSizeLeft;

                m_dwCurrentChunkSizeLeft    =  0L;
            }
            else
            {
                // Meets the end of buffer before end of chunk.
                dwToSkip                    -= dwCurrBuffSizeLeft;
                m_dwCurrentChunkSizeLeft    -= dwCurrBuffSizeLeft;

                m_dwPosInCurrDataBuff       = dwCurBuffUsedSize;
            }
        }
        else
        {
            // Move cursor in the current buffer.
            if(dwToSkip >= m_dwCurrentChunkSizeLeft)
            {
                // Meets end of current chunk. Skip it.
                dwToSkip                    -= m_dwCurrentChunkSizeLeft;
                m_dwPosInCurrDataBuff       += m_dwCurrentChunkSizeLeft;

                m_dwCurrentChunkSizeLeft    = 0L;
            }
            else
            {
                // Move in current chunk.
                m_dwPosInCurrDataBuff       += dwToSkip;
                m_dwCurrentChunkSizeLeft    -= dwToSkip;

                dwToSkip                    = 0L;
            }
        }

        // Check if reached the end of the current buffer.
        if(m_dwPosInCurrDataBuff == dwCurBuffUsedSize)
        {
            // The current buffer is either empty or we must use the next buffer.
            if(m_pCurrDataBuff->GetNextBuff() != NULL)
            {
                m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                m_dwPosInCurrDataBuff = 0L;

                continue;
            }
            else
            {
                // No other buffers available. Cannot skip all the items.
                return dwItems - dwToSkip;
            }
        }

        // Check if reached the end of the current chunk.
        if(m_dwCurrentChunkSizeLeft == 0)
        {
            if(ckReadChunkSize() == FALSE || m_dwCurrentChunkSizeLeft == 0)
            {
                // Reached the end of the data.
                return dwItems - dwToSkip;
            }
        }
    }

    return dwItems;
}

// ----------------------------------------------------------------------------
//
// Compute the number of items 'first' should be moved forward in order to find 
// 'second'. Returns ULONG_MAX if could not find 'second' or if an error 
// occurred.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLCA>::ComputeDistance(const CDataBuffListBaseAccessor<CdBLCA> &first, const CDataBuffListBaseAccessor<CdBLCA> &second) const
{
    if(((CDataBuffListChunkAccessor&)first).m_pDataBuffList == NULL || ((CDataBuffListChunkAccessor&)second).m_pDataBuffList == NULL)
    {
        return ULONG_MAX;   // No data buff list associated.
    }

    CDataBuffListChunkAccessor iter = (CDataBuffListChunkAccessor&)first;
    CDataBuffListChunkAccessor end  = (CDataBuffListChunkAccessor&)second;

    DWORD dwDist = 0;

    // Move to the right buffer, skipping the chunk data.
    while(iter.m_pDataBuffList->GetBuffList() != end.m_pDataBuffList->GetBuffList())
    {
        dwDist += iter.m_dwCurrentChunkSizeLeft;
        if(iter.ckMoveNextChunk() != TRUE)
            return ULONG_MAX;
    }

    // Move to the right chunk in the buffer pointed to by 'end'.
    while(iter.m_dwPosInCurrDataBuff + iter.m_dwCurrentChunkSizeLeft < end.m_dwPosInCurrDataBuff)
    {
        assert(iter.m_dwPosInCurrDataBuff + iter.m_dwCurrentChunkSizeLeft < iter.m_pCurrDataBuff->GetUsedDataSize());

        dwDist += iter.m_dwCurrentChunkSizeLeft;
        iter.m_dwPosInCurrDataBuff += iter.m_dwCurrentChunkSizeLeft;
        iter.ckReadChunkSize();
    }

    // Move to the right position in the found chunk.
    dwDist += end.m_dwPosInCurrDataBuff - iter.m_dwPosInCurrDataBuff;

    return dwDist;
}


// ----------------------------------------------------------------------------
//
// Move the cursor to the next chunk.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckMoveNextChunk()
{
    // Search for the right buffer (a chunk may occupy several buffers).

    DWORD dwCurrBuffSizeLeft = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;

    while(m_dwCurrentChunkSizeLeft >= dwCurrBuffSizeLeft)
    {
        // Meets end of buffer before end of chunk. Move to the end of the current buffer.
        m_dwCurrentChunkSizeLeft -= dwCurrBuffSizeLeft;

        if(m_pCurrDataBuff->GetNextBuff() == FALSE)
            return FALSE;

        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
        m_dwPosInCurrDataBuff = 0;

        dwCurrBuffSizeLeft = m_pCurrDataBuff->GetUsedDataSize();
    }

    // Move to the end of the chunk.
    m_dwPosInCurrDataBuff += m_dwCurrentChunkSizeLeft;

    return ckReadChunkSize();
}


// ----------------------------------------------------------------------------
//
// Copy dwCharCount chars into destStr starting from current pos.
// Removes chunk information.
// Return FALSE if couldn't find enough characters.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::BuildString(std::string &destStr, DWORD dwCharCount) const
{
    destStr.erase(); destStr.reserve(dwCharCount);

    CDataBuff * pDataBuffCurs = m_pCurrDataBuff;
    CHAR *      pAvDataBuff;
    DWORD dwCopyBytes;
    DWORD dwCurrBuffSizeLeft= m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;
    DWORD dwChunkSizeLeft   = m_dwCurrentChunkSizeLeft;

    while(TRUE)
    {
        // Copy till the end of dwCharCount or end of buffer or end of chunk, whatever comes first.
        dwCopyBytes = __min(__min(dwCharCount, dwCurrBuffSizeLeft), dwChunkSizeLeft);
        pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[pDataBuffCurs->GetUsedDataSize() - dwCurrBuffSizeLeft]);
        destStr.append(pAvDataBuff, dwCopyBytes);

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0)
            return TRUE;
        else
        {
            dwChunkSizeLeft -= dwCopyBytes;

            // Ajust values.
            if(dwChunkSizeLeft == 0)
            {
                // Move to the next chunk.
                dwCurrBuffSizeLeft -= dwCopyBytes;

                // Read new chunk size.
                char curChar;
                while(TRUE)
                {
                    if(dwCurrBuffSizeLeft == 0)
                    {
                        // Move to the next buffer.
                        pDataBuffCurs = pDataBuffCurs->GetNextBuff();

                        if(pDataBuffCurs == NULL)
                            return FALSE;

                        dwCurrBuffSizeLeft = pDataBuffCurs->GetUsedDataSize();
                    }
                    curChar = pDataBuffCurs->GetDataPtr()[pDataBuffCurs->GetUsedDataSize() - dwCurrBuffSizeLeft];

                    if(IsDigit(curChar) == FALSE)
                    {
                        // End of the chunk description.
                        assert(curChar == '\r' || curChar == '\n');
                        break;
                    }

                    dwChunkSizeLeft *= 10;
                    dwChunkSizeLeft += curChar - '0';
                    dwCurrBuffSizeLeft --;
                }

                DWORD dwPosInBuff = pDataBuffCurs->GetUsedDataSize() - dwCurrBuffSizeLeft;
                if(SkipChunkCRLF(dwPosInBuff, pDataBuffCurs) == FALSE)
                    return FALSE;
                dwCurrBuffSizeLeft = pDataBuffCurs->GetUsedDataSize() - dwPosInBuff;
            }
            else
            {
                // Move to the next buffer.
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

                if(pDataBuffCurs == NULL)
                    return FALSE;

                dwCurrBuffSizeLeft = pDataBuffCurs->GetUsedDataSize();
            }
        }
    }
}



// ----------------------------------------------------------------------------
//
// Read the next char in buffs without moving the iterator.
// Return FALSE if no value could be read.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::NextChar(CHAR &nextChar) const
{
    CDataBuff * pCurrDataBuff           = m_pCurrDataBuff;
    DWORD       dwPosInCurrDataBuff     = m_dwPosInCurrDataBuff;
    DWORD       dwCurrentChunkSizeLeft  = m_dwCurrentChunkSizeLeft;

    for(DWORD dw=0 ; dw<2 ; dw++)
    {
        // Adjust values. First time to prepare moving to the next buffer; second time to make sure the current char is accessible.
            // Eventually move to the next data buffer.
        if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
        {
            // No data left to read in the current buffer. Try to move to the next.
            if(pCurrDataBuff->GetNextBuff() != NULL)
            {
                pCurrDataBuff = pCurrDataBuff->GetNextBuff();
                dwPosInCurrDataBuff = 0;
            }
            else
            {
                return FALSE;
            }
        }

        // Eventually move to the next chunk.
        if(dwCurrentChunkSizeLeft == 0)
        {
            // TODO: merger ce code.
            // Moving to the next chunk.
            // Read the chunk size.
            char curChar;
            while(IsDigit(curChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff]) == TRUE)
            {
                dwCurrentChunkSizeLeft *= 10;
                dwCurrentChunkSizeLeft += curChar - '0';
                dwPosInCurrDataBuff++;

                // Eventually move to the next data buffer.
                if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
                {
                    // No data left to read in the current buffer. Try to move to the next.
                    if(pCurrDataBuff->GetNextBuff() != NULL)
                    {
                        pCurrDataBuff = pCurrDataBuff->GetNextBuff();
                        dwPosInCurrDataBuff = 0;
                    }
                    else
                    {
                        return FALSE;
                    }
                }
            }

            if(SkipChunkCRLF(dwPosInCurrDataBuff, pCurrDataBuff) == FALSE)
                return FALSE;
        }

        if(dw == 0)
        {
            // Only in the first loop.
                // Move to the next char.
            dwPosInCurrDataBuff++;
            dwCurrentChunkSizeLeft--;
        }
    }

    // Read the following char.
    nextChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

    return TRUE;
}

// ----------------------------------------------------------------------------
//
// Move to the next char or return FALSE if reached the end of the buffers list.
// m_dwPosInCurrDataBuff may then point to a position exceeding the current buff size
// and m_dwCurrentChunkSizeLeft may evaluate to zero, but this will be handled in 
// functions retrieving the current char. The reason for this is that the next buffer
// may not be available yet if parsing during reception.
// OK, not very usefull for the chunk accessor yet, that may be in the future.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::MoveNextChar()
{
    m_dwPosInCurrDataBuff++;
    m_dwCurrentChunkSizeLeft--;

    if(m_dwPosInCurrDataBuff == m_pCurrDataBuff->GetUsedDataSize())
    {
        // Move to the next buffer.
        if(m_pCurrDataBuff->GetNextBuff() == NULL)
            return FALSE;

        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
        m_dwPosInCurrDataBuff = 0;
    }

    if(m_dwCurrentChunkSizeLeft == 0)
    {
        // Read the next chunk size.
        if(ckReadChunkSize() == FALSE)
            return FALSE;
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// Depending on 'bCaseSensitive', performs a case-sensitive or not-sensitive 
// search of the 'dwStrLen'-size string 'str' in buffers delimitted by 'this' 
// and 'IterStopAt'.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ContainString(const CHAR *str, DWORD dwStrLen, const CDataBuffListChunkAccessor &iterStopAt, BOOL bCaseSensitive)   const
{
    CDataBuffListChunkAccessor baseIter = *this;     // The one used for come-backs.
    CDataBuffListChunkAccessor movingIter = *this;   // The one that iterates the buffs.
    DWORD dwMatchedChars = 0;
    DWORD dwToMatchChars;
    int iCmpRes = 1;

    while(dwMatchedChars < dwStrLen)
    {
        // Compare up to dwMatchedChars or end of current buff or end of current chunk, whichever comes first.
        dwToMatchChars = __min(__min(dwStrLen, movingIter.m_dwPosInCurrDataBuff-movingIter.m_pCurrDataBuff->GetUsedDataSize()), movingIter.m_dwCurrentChunkSizeLeft);

        // Check if we reach iterStopAt.
        if(movingIter.m_pCurrDataBuff == iterStopAt.m_pCurrDataBuff)
        {
            if(movingIter.m_dwPosInCurrDataBuff + dwToMatchChars > iterStopAt.m_dwPosInCurrDataBuff)
            {
                return FALSE;
            }
        }

        if(bCaseSensitive)
            iCmpRes = memcmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);
        else
            iCmpRes = _memicmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                return TRUE;
            }
            else
            {
                // Must look in the next buffer and/or the next chunk for the remaining of the string.
                movingIter.m_dwCurrentChunkSizeLeft -= dwToMatchChars;
                movingIter.m_dwPosInCurrDataBuff    += dwToMatchChars;

                if(movingIter.m_dwCurrentChunkSizeLeft == 0)
                {
                    if(movingIter.ckReadChunkSize() == FALSE)
                        return FALSE;
                }
                if(movingIter.m_dwPosInCurrDataBuff == movingIter.m_pCurrDataBuff->GetUsedDataSize())
                {
                    if(movingIter.m_pCurrDataBuff->GetNextBuff() == NULL)
                        return FALSE;
                    else
                    {
                        movingIter.m_pCurrDataBuff = movingIter.m_pCurrDataBuff->GetNextBuff();
                        movingIter.m_dwPosInCurrDataBuff = 0;
                    }
                }
            }
        }
        else
        {
            // Found a difference. Try again starting with the next base-character.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            movingIter = baseIter;
            dwMatchedChars = 0;
        }
    }

    return dwMatchedChars == dwStrLen;
}

// ----------------------------------------------------------------------------
//
// Finds or skips the str string.
// If not found, returns FALSE and set *this so that it is possible to continue the search if buffers are
// added. That way, if the begining of the searched string is found but chars are missing, the next 
// search will start right at that position.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::FindSkipStr(LPCSTR str, DWORD dwStrLen, EFindSkip findSkip, BOOL bCaseSensitive)
{
    CDataBuffListChunkAccessor baseIter = *this;     // The one used for come-backs.
    DWORD dwMatchedChars = 0;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        dwToMatchChars = __min(__min(dwStrLen, m_dwPosInCurrDataBuff-m_pCurrDataBuff->GetUsedDataSize()), m_dwCurrentChunkSizeLeft);

        if(bCaseSensitive)
            iCmpRes = memcmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);
        else
            iCmpRes = _memicmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                if(findSkip == FIND)
                {
                    // Were searching for the begining of the string.
                    operator=(baseIter);
                    return TRUE;
                }
                else
                {
                    // Must skip the found string.
                    m_dwCurrentChunkSizeLeft -= dwToMatchChars;
                    m_dwPosInCurrDataBuff    += dwToMatchChars;

                    return TRUE;
                }
            }
            else
            {
                // Must look in the next buffer and/or the next chunk for the remaining of the string.
                if(m_dwCurrentChunkSizeLeft == 0)
                {
                    if(ckReadChunkSize() == FALSE)
                    {
                        operator=(baseIter);    // Move back.
                        return FALSE;
                    }
                }
                if(m_dwPosInCurrDataBuff == m_pCurrDataBuff->GetUsedDataSize())
                {
                    if(m_pCurrDataBuff->GetNextBuff() == NULL)
                    {
                        operator=(baseIter);    // Move back.
                        return FALSE;
                    }
                    else
                    {
                        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                        m_dwPosInCurrDataBuff = 0;
                    }
                }
            }
        }
        else
        {
            // Found a difference. Try again starting with the next base-character.
            if(baseIter.MoveNextChar() == FALSE)
            {
                operator=(baseIter);    // Move back.
                return FALSE;
            }
            operator=(baseIter);
            dwMatchedChars = 0;
        }
    }
}

// ----------------------------------------------------------------------------
//
// Read and skip the chunk size *this is pointing to.
// If not enough data available, then return FALSE and reset iterator to 
// current pos.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckReadChunkSize()
{
    // Adjust the pointers.
    assert(m_pCurrDataBuff);

    return ckLocalReadChunkSize();
}

// ----------------------------------------------------------------------------
//
// Local version of ckReadChunkSize, which assumes that the iterator was set 
// correctly.
// Typically used in methods like ckAdjustIter() to avoid circular calls.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckLocalReadChunkSize()
{
    if(ckReadChunkDWORD(m_dwCurrentChunkSizeLeft) == FALSE)
    {
        // Miss some chars. Must wait for the next buffer receival to get the missing info.
        return FALSE;
    }

    // Skip the CRLF sequence.
    if(ckSkipChunkCRLF() == FALSE)
    {
        // Miss some chars. Must wait for the next buffer receival to get the missing info.
        return FALSE;
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// ckReadChunkDWORD
//
// Local version of ReadDWORD used to avoid re-entering calls.
// Reads a sequence of numbers till finds any other char.
//// Does not move the cursor through the buffer in case could not read an entire
//// DWORD.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckReadChunkDWORD(DWORD &res)
{
    CHAR curChar;

    res = 0;

    while(TRUE)
    {
        CurrentChar(curChar);

        if(IsDigit(curChar) == FALSE)
        {
            // Reached the end of the value string. The value was successfully parsed.
            assert(curChar == '\r' || curChar == '\n');
            return TRUE;
        }

        res *= 10L;
        res += curChar - '0';

        m_dwPosInCurrDataBuff++;

        if(m_dwPosInCurrDataBuff == m_pCurrDataBuff->GetUsedDataSize())
        {
            if(m_pCurrDataBuff->GetNextBuff() != NULL)
            {
                m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                m_dwPosInCurrDataBuff = 0;
            }
            else
            {
                return FALSE;
            }
        }
    }

    // Reached the end of the buffers, but not the end of the value string.
    return FALSE;
}


/* #########################################################################

CDataBuffListAccessor

  Define the specialisation of the methods in CDataBuffListBaseAccessor
  for normal buffers.

######################################################################### */


template <>
WILL_INLINE CDataBuffListBaseAccessor<CdBLA>::CDataBuffListBaseAccessor<CdBLA>(CDataBuffList *pDataBuffList)
        : m_dwPosInCurrDataBuff(0), m_pDataBuffList(pDataBuffList)
{
    if(pDataBuffList != NULL)
    {
        m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
    }
    // Otherwise, will have to call AdjustPointers() when the data is available.
}

// Reset the iterator.
template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLA>::Rewind()
{
    assert(m_pDataBuffList != NULL);

    m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
    m_dwPosInCurrDataBuff = 0;
}


// ----------------------------------------------------------------------------
//
// Adjust the iterator position.
// Set the iterator to point at the data contained in the CDataBuffList.
//
// Note that we should not continue using that iterator until AdjustIterator 
// returns TRUE.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::AdjustIterator()
{
    assert(m_pDataBuffList != NULL);

    if(m_pCurrDataBuff == NULL)
    {
        // The iterator was created before we received the first data buffers.
        m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
        m_dwPosInCurrDataBuff = 0;
    }

    if(m_pCurrDataBuff == NULL)
        return FALSE;

    if(m_dwPosInCurrDataBuff == m_pCurrDataBuff->GetUsedDataSize())
    {
        // Iterator was left beyond the end of the current buffer because there was no next buff at that time.
        // Try to go to the next buffer.
        if(m_pCurrDataBuff->GetNextBuff() != NULL)
        {
            m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
            m_dwPosInCurrDataBuff = 0L;
        }
        else
        {
            return FALSE;
        }
    }
    return TRUE;
}


// Try to move the iterator dwItems items further. Returns the number of items it skipped actually
// (may be less than dwItems if not enough data is available).
// The three cases are: move in the current buffer; must move to another buffer; must move to the 
// next buffer, but not received yet.

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLA>::Skip(DWORD dwItems)
{
    assert(m_pDataBuffList != NULL);

    DWORD dwToSkip = dwItems;
    DWORD dwCurBuffUsedSize;

    while(dwToSkip > 0)
    {
        dwCurBuffUsedSize = m_pCurrDataBuff->GetUsedDataSize();

        if(dwToSkip >= dwCurBuffUsedSize - m_dwPosInCurrDataBuff)
        {
            // Current buffer not wide enough. Must skip till one of the next buffers.
            dwToSkip -= dwCurBuffUsedSize - m_dwPosInCurrDataBuff;

            // Reached the end of the current buffer. Move the cursor to the next buffer.
            // The current buffer is either empty or we must use the next buffer.
            if(m_pCurrDataBuff->GetNextBuff() != NULL)
            {
                m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                m_dwPosInCurrDataBuff = 0L;
            }
            else
            {
                // No other buffers available.
                m_dwPosInCurrDataBuff = dwCurBuffUsedSize;
                return dwItems - dwToSkip;
            }
        }
        else
        {
            // Move cursor in the current buffer.
            m_dwPosInCurrDataBuff += dwToSkip;
            dwToSkip = 0;
        }
    }

    return dwItems;
}


// Compute the number of items 'first' should be moved forwardin order to find 'second'.
// Returns ULONG_MAX if could not find 'second' or if an error occurred.

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLA>::ComputeDistance(const CDataBuffListBaseAccessor<CdBLA> &first, const CDataBuffListBaseAccessor<CdBLA> &second) const
{
    CDataBuffListAccessor iter =    (CDataBuffListAccessor&)first;
    CDataBuffListAccessor end =     (CDataBuffListAccessor&)second;

    const CDataBuff *pFirstDataBuff = iter.m_pCurrDataBuff;
    const CDataBuff *pSecondDataBuff = end.m_pCurrDataBuff;

    if(pFirstDataBuff == pSecondDataBuff)
    {
        // Pointers to the same buffers.
        if(end.m_dwPosInCurrDataBuff < iter.m_dwPosInCurrDataBuff)
            return ULONG_MAX;
        else
            return end.m_dwPosInCurrDataBuff-iter.m_dwPosInCurrDataBuff;
    }
    else
    {
        // Search in the buffers following *this.

        DWORD dwRes = iter.m_pCurrDataBuff->GetUsedDataSize()-iter.m_dwPosInCurrDataBuff;

        pFirstDataBuff = pFirstDataBuff->GetNextBuff();
        if(pFirstDataBuff == NULL)
            return ULONG_MAX;

        while(pFirstDataBuff != end.m_pCurrDataBuff)
        {
            dwRes += pFirstDataBuff->GetUsedDataSize();
            pFirstDataBuff = pFirstDataBuff->GetNextBuff();

            if(pFirstDataBuff == NULL)
                return ULONG_MAX;
        }

        return dwRes + end.m_dwPosInCurrDataBuff;
    }
}


// Copy dwCharCount chars into destStr starting from current pos.
// Return FALSE if couldn't find enough characters.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::BuildString(std::string &destStr, DWORD dwCharCount) const
{
    destStr.erase(); destStr.reserve(dwCharCount);

    CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    CHAR *  pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[m_dwPosInCurrDataBuff]);
    DWORD dwCopyBytes;
    DWORD dwDataAvailable = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;

    while(TRUE)
    {
        dwCopyBytes = __min(dwCharCount, dwDataAvailable);
        destStr.append(pAvDataBuff, dwCopyBytes);

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0)
            return TRUE;
        else
        {
            pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;

            pAvDataBuff     = pDataBuffCurs->GetDataPtr();
            assert(pAvDataBuff != NULL);
            dwDataAvailable = pDataBuffCurs->GetUsedDataSize();
        }
    }
}


// Copy dwCharCount chars into destStr starting from current pos.
// Return FALSE if couldn't find enough characters.
// Won't add a \0 to the end of string.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::BuildString(LPSTR destStr, DWORD dwCharCount)   const
{
    CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    CHAR *  pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[m_dwPosInCurrDataBuff]);
    DWORD   dwCopyBytes;
    DWORD   dwDataAvailable = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;
    LPSTR   pCurrStrPos = destStr;

    while(TRUE)
    {
        dwCopyBytes = __min(dwCharCount, dwDataAvailable);
        CopyMemory(pCurrStrPos, pAvDataBuff, dwCopyBytes);
        pCurrStrPos += dwCopyBytes;

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0)
            return TRUE;
        else
        {
            pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;

            pAvDataBuff     = pDataBuffCurs->GetDataPtr();
            assert(pAvDataBuff != NULL);
            dwDataAvailable = pDataBuffCurs->GetUsedDataSize();
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ReadNextWord(std::string &destStr) const
{
    destStr.erase();

    CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    DWORD   dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    CHAR cCurChar;

    while(TRUE)
    {
        cCurChar = pDataBuffCurs->GetDataPtr()[dwPosInCurrDataBuff];
        if(IsWS(cCurChar) == FALSE && IsEOL(cCurChar) == FALSE)
        {
            destStr += cCurChar;

            dwPosInCurrDataBuff++;
            if(dwPosInCurrDataBuff == pDataBuffCurs->GetUsedDataSize())
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;
        }
        else
        {
            return (dwPosInCurrDataBuff != m_dwPosInCurrDataBuff);
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ReadNextWord(DWORD &dwDest) const
{
    dwDest = 0;

    CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    DWORD   dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    CHAR cCurChar;

    while(TRUE)
    {
        cCurChar = pDataBuffCurs->GetDataPtr()[dwPosInCurrDataBuff];
        if(IsDigit(cCurChar) == TRUE)
        {
            dwDest *= 10;
            dwDest += cCurChar - '0';

            dwPosInCurrDataBuff++;
            if(dwPosInCurrDataBuff == pDataBuffCurs->GetUsedDataSize())
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;
        }
        else
        {
            return (dwPosInCurrDataBuff != m_dwPosInCurrDataBuff);
        }
    }
}


// Read the next char in buffs.
// Return FALSE if no value could be read.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::NextChar(CHAR &nextChar) const
{
    assert(m_pCurrDataBuff);

    CDataBuff * pCurrDataBuff = m_pCurrDataBuff;
    DWORD       dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    dwPosInCurrDataBuff++;

    // Make sure the following char is available.
    if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
    {
        // No data left to read in the current buffer. Try to move to the next.
        if(pCurrDataBuff->GetNextBuff() != NULL)
        {
            pCurrDataBuff = pCurrDataBuff->GetNextBuff();
            dwPosInCurrDataBuff = 0;
        }
        else
        {
            return FALSE;
        }
    }

    // Read the following char.
    nextChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

    return TRUE;
}


// Move to the next char or does not move and return FALSE if reached the end 
// of the buffers list.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::MoveNextChar()
{
    assert(m_pCurrDataBuff != NULL);

    m_dwPosInCurrDataBuff++;

    if(m_dwPosInCurrDataBuff == m_pCurrDataBuff->GetUsedDataSize())
    {
        // No data left to read in the current buffer. Try to move to the next.
        if(m_pCurrDataBuff->GetNextBuff() != NULL)
        {
            m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
            m_dwPosInCurrDataBuff = 0;
        }
        else
        {
            return FALSE;
        }
    }

    return TRUE;
}


// Depending on bCaseSensitive, performs a case-sensitive or insensitive search of str in buffers delimitted by this and IterStopAt.
// TODO: merge with FindSkipStr ?
template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ContainString(const CHAR *str, DWORD dwStrLen, const CDataBuffListAccessor &iterStopAt, BOOL bCaseSensitive) const
{
    CDataBuffListAccessor baseIter = *this;     // The one used for come-backs.
    CDataBuffListAccessor movingIter = *this;   // The one that iterates the buffs.
    DWORD dwMatchedChars = 0;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        dwToMatchChars = __min(dwStrLen, movingIter.m_dwPosInCurrDataBuff-movingIter.m_pCurrDataBuff->GetUsedDataSize());

        // Check if we reach iterStopAt.
        if(movingIter.m_pCurrDataBuff == iterStopAt.m_pCurrDataBuff)
        {
            if(movingIter.m_dwPosInCurrDataBuff + dwToMatchChars > iterStopAt.m_dwPosInCurrDataBuff)
            {
                return FALSE;
            }
        }

        if(bCaseSensitive)
            iCmpRes = memcmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);
        else
            iCmpRes = _memicmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                return TRUE;
            }
            else
            {
                // Must look in the next buffer for the remaining of the string.
                if(movingIter.MoveNextBuff() == FALSE)
                    return FALSE;
            }
        }
        else
        {
            // Found a difference.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            movingIter = baseIter;
            dwMatchedChars = 0;
        }
    }
}

// Finds or skips the str string.
// If not found, returns FALSE and set *this so that it is possible to continue the search if buffers are
// added. That way, if the begining of the searched string is found but chars are missing, the next 
// search will start right at that position.
template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::FindSkipStr(LPCSTR str, DWORD dwStrLen, EFindSkip findSkip, BOOL bCaseSensitive)
{
    CDataBuffListAccessor baseIter = *this;     // The one used for come-backs.
    DWORD dwMatchedChars = 0;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        dwToMatchChars = __min(dwStrLen-dwMatchedChars, m_dwPosInCurrDataBuff-m_pCurrDataBuff->GetUsedDataSize());

        if(bCaseSensitive)
            iCmpRes = memcmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);
        else
            iCmpRes = _memicmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                if(findSkip == FIND)
                {
                    // Move back to the begining of the found string.
                    operator=(baseIter);
                    return TRUE;
                }
                else
                {
                    // Must skip the found string.
                    m_dwPosInCurrDataBuff += dwStrLen;
                    return TRUE;
                }
            }
            else
            {
                // Must look in the next buffer for the remaining of the string.
                if(MoveNextBuff() == FALSE)
                    return FALSE;
            }
        }
        else
        {
            // Found a difference. Try with the next character.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            operator=(baseIter);
            dwMatchedChars = 0;
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::MoveNextBuff()
{
    if(m_pCurrDataBuff->GetNextBuff() == NULL)
    {
        // Caller must wait for more data before continuing.
        return FALSE;
    }
    else
    {
        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
        m_dwPosInCurrDataBuff = 0;

        return TRUE;
    }
}

// ----------------------------------------------------------------------------
//
// Defined only for standard CDataBuffListAccessor.
// Move cursor past the end of the data buffers.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLA>::SkipEnd()
{
    while(m_pCurrDataBuff->GetNextBuff() != NULL)
        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();

    m_dwPosInCurrDataBuff = m_pCurrDataBuff->GetUsedDataSize();
}


/* #########################################################################

    CDataBuffHandler

   ######################################################################### */

WILL_INLINE WSABUF* CDataBuffHandler::AllocWsaBuffList()
{
    // Reserve a list of data buffers.
    if(m_wsaMappedBuffList.AllocateBuffs(WSABUF_COUNT-m_dwRemainingBuffsInMappedList) == FALSE)
        return NULL;

    // Map the data buffers to the WSA list.
    CDataBuffListAccessor dbListAcc(&m_wsaMappedBuffList);
    for(DWORD dw=0 ; dw<WSABUF_COUNT ; dw++)
    {
        m_wsaBuffArray[dw].buf = dbListAcc.GetCurrDataBuff()->GetDataPtr();
        m_wsaBuffArray[dw].len = GetDataBuffUsableSize();
        dbListAcc.MoveNextBuff();   // Nb: last call to MoveNextBuff will return FALSE since reached the end of the list.
    }

    return m_wsaBuffArray;
}

// Store the buffers where data was stored and keep a list of the empty ones 
// for use by the next AllocWsaBuffList. Remaining buffs are stored in 
// m_wsaMappedBuffList.
WILL_INLINE void CDataBuffHandler::ReturnWsaBuffList(DWORD dwBytesStored)
{
    // Update the size of the WSA buffers into the matching data buffers.
    DWORD dwBuffLen;
    DWORD dwRelevantBuffCount = 0;

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
        InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pDataStoredCounterRef, dwBytesStored);
#   else
        (*CGlobalDataBuffHandler::s_pDataStoredCounterRef) += dwBytesStored;
#   endif
#endif

    CDataBuff *pLastUsedBuff = NULL;

    // Update the used memory in each CDataBuff involved.
    CDataBuffListAccessor dbListAcc(&m_wsaMappedBuffList);
    while(dwBytesStored != 0)
    {
        dwBuffLen = __min(dwBytesStored, GetDataBuffUsableSize());

        dbListAcc.GetCurrDataBuff()->SetUsedDataSize(dwBuffLen);
        dwBytesStored -= dwBuffLen;

        if(dwBytesStored == 0)
            pLastUsedBuff = dbListAcc.GetCurrDataBuff();

        dbListAcc.MoveNextBuff();
        dwRelevantBuffCount++;
    }

    // Store the buffers containing data, and keep the empty buffers for the next AllocWsaBuffList.
//    m_dataBuffList.AddBuffList(m_wsaMappedBuffList, dwRelevantBuffCount, m_wsaMappedBuffList);
    m_dataBuffList.AddBuffList(m_wsaMappedBuffList, pLastUsedBuff, m_wsaMappedBuffList);

    assert(WSABUF_COUNT >= dwRelevantBuffCount);

    m_dwRemainingBuffsInMappedList = WSABUF_COUNT-dwRelevantBuffCount;
}

WILL_INLINE void CDataBuffHandler::ClearBuffer()
{
#ifdef USE_PERF_DATA
    CDataBuffListAccessor dbListAcc(&m_dataBuffList);
    while(dbListAcc.GetCurrDataBuff() != NULL)
    {
#   ifdef LOCK_COUNTERS
        InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pDataStoredCounterRef, -(LONG)dbListAcc.GetCurrDataBuff()->GetUsedDataSize());
#   else
        (*CGlobalDataBuffHandler::s_pDataStoredCounterRef) -= (LONG)dbListAcc.GetCurrDataBuff()->GetUsedDataSize();
#   endif
        if(dbListAcc.MoveNextBuff() == FALSE)
            break;
    }
#endif

    m_dataBuffList.ReleaseBuffList();
    m_wsaMappedBuffList.ReleaseBuffList();
    m_dwRemainingBuffsInMappedList = 0;
}

WILL_INLINE std::ostream &operator <<(std::ostream &os, CDataBuffHandler &dbHdl)
{
    os << dbHdl.GetWholeBuffList();
    return os;
}
