/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_THREADPOOLCONTROLLER_H_
#define _OPENSTA_THREADPOOLCONTROLLER_H_

////////////////////////////////////////////////////////////////////////////////
// Thread pool controller definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Iocp.hpp"
#include "ThreadPool.hpp"
#include "Trace.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif


// This class implements a high level container for a thread pool and IOCP.
// Typically, derived classes will contain or manager asynchronous request
// managers.
// THREAD SAFE, except Init(), Run() and Stop().
class CThreadPoolController : public IocpEvent, public IocpEventDispatcher
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CThreadPoolController(CTrace* pTrace = NULL)
        : m_pIocp(NULL), m_pThreadPool(NULL),
          m_pTrace(pTrace)
//          ,m_sl(4000)
    {}
    virtual ~CThreadPoolController()
    {
        if(m_pThreadPool)
            delete m_pThreadPool;
        if(m_pIocp)
            delete m_pIocp;
    }

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Initialization. Create the IOCP, then the thread pool.
    // The derived overrides must call this base implementation.
    virtual DWORD Init(DWORD dwThreadPoolSize, DWORD dwNumberOfConcurrentThreadsInIocp = 0)
    {
        DWORD dwError = 0;

#ifdef USE_PERF_DATA
        *g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_THREADS) = dwThreadPoolSize;
#endif

        m_pIocp = new CIOCP(dwNumberOfConcurrentThreadsInIocp, m_pTrace);
        if(m_pIocp)
        {
            m_pIocp->Init();
            m_pThreadPool = new CThreadPool(*m_pIocp, dwThreadPoolSize, m_pTrace);
            if(!m_pThreadPool)
            {
                delete m_pIocp;
                m_pIocp = NULL;
                dwError = ERROR_NOT_ENOUGH_MEMORY;
            }
        }
        else
            dwError = ERROR_NOT_ENOUGH_MEMORY;

        return dwError;
    }

    // Run the thread pool.
    DWORD Run(DWORD m_dwTimeoutMilliseconds = INFINITE)
    {
        return m_pThreadPool->Run(this, m_dwTimeoutMilliseconds);
    }

    // Stop the thread pool.
    DWORD Stop()
    {
        return m_pThreadPool->Stop();
    }

    // Get peak count of runnable threads in the thread pool.
    DWORD GetPeakRunnableThreadCount() const {return m_pThreadPool->GetPeakRunnableThreadCount();}

    // Get peak number of oustanding I/Os in the I/O completion queue.
    DWORD GetPeakOustandingIos() const {return m_pIocp->GetPeakOustandingIos();}

    // Get the total number of threads initially created in the pool.
    DWORD GetPoolSize() const {return m_pThreadPool->GetPoolSize();}

    // Return the IOCP.
    CIOCP& GetIocp() {return *m_pIocp;}

    //// --------------------------------------------------------------------------------
    //// IocpEvent methods.
    //// --------------------------------------------------------------------------------
    virtual void OnEvent      (DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/) {}
    virtual void OnEventFailed(DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/, DWORD /*dwError*/) {}

    //// --------------------------------------------------------------------------------
    //// IocpEventDispatcher methods.
    //// --------------------------------------------------------------------------------
    IocpEvent* GetIocpEventTarget() {return this;}

    //// --------------------------------------------------------------------------------
    //// Callbacks from CAsyncRequestManager. This base implementation does nothing.
    //// --------------------------------------------------------------------------------
    virtual void OnAllRequestsFinished(DWORD /*dwManagerNumber*/)  {}
    virtual void OnAllRequestsCancelled(DWORD /*dwManagerNumber*/) {}

protected:
//    CFastLock           m_sl;           // Method-level protection against concurrent accesses.
    CIOCP*              m_pIocp;
    CThreadPool*        m_pThreadPool;
    CTrace             *m_pTrace;       // The ostream trace support object.
};

#endif  // _OPENSTA_THREADPOOLCONTROLLER_H_
